/* aenc.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <pthread.h>
#include <errno.h>

#include "acodec.h"
#include "aframe.h"
#include "aframe_mgr.h"
#include "aread.h"
#include "aenc.h"


/* Thread function */

static void aenc_error(aenc_t *aenc, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "aenc: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);

  if (!aenc) return;

  pthread_mutex_lock(&aenc->mutex);
  aenc->state = AENC_STATE_ERROR;
  pthread_mutex_unlock(&aenc->mutex);
  if (aenc->error_cb)
    aenc->error_cb(aenc->error_cb_data);
//  exit(1);
}

static inline int
get_state(aenc_t *aenc)
{
  int state;
  pthread_mutex_lock(&aenc->mutex);
  state = aenc->state;
  pthread_mutex_unlock(&aenc->mutex);
  return state;
}

static void
state_change_done(aenc_t *aenc)
{
  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_INTERFACE_WAIT) {
    aenc->state &= ~AENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aenc->interface_wait_cond);
  }
  pthread_mutex_unlock(&aenc->mutex);
}

static void
set_state(aenc_t *aenc, uint32_t state)
{
  pthread_mutex_lock(&aenc->mutex);
  aenc->state &= ~(AENC_STATE_START|AENC_STATE_PAUSE|AENC_STATE_STOP|AENC_STATE_QUIT);
  aenc->state |= state;
  if (aenc->state & AENC_STATE_INTERFACE_WAIT) {
    aenc->state &= ~AENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aenc->interface_wait_cond);
  }
  pthread_mutex_unlock(&aenc->mutex);
}

static int
add_aframe_chain(aenc_t *aenc, aframe_t *afrm)
{
  int ret;
  pthread_mutex_lock(&aenc->mutex);
  ret = aframe_chain_add(&aenc->aframe_chain, afrm);
  pthread_mutex_unlock(&aenc->mutex);
  if (aenc->aframe_ready_cb)
    aenc->aframe_ready_cb(aenc->aframe_ready_cb_data);
  return ret;
}

static int
put_enc_aframe(aenc_t *aenc)
{
  int stream_length;
  int ret = AENC_OK;

#ifndef NDEBUG
  if (!aenc->src_aframe->bufp) {
    aenc_error(aenc, "put_enc_aframe: error EXT_AFRAME, flag 0x%x.\n",aenc->src_aframe->flag);
    return AENC_FAIL;
  } else if (aenc->src_aframe->samples <= 0) {
    aenc_error(aenc, "put_enc_aframe: error samples <= 0, %d.\n",aenc->src_aframe->samples);
    return AENC_FAIL;
  }
#endif /* NDEBUG */

  if (!aenc->dst_aframe) {
    while((aenc->dst_aframe=aframe_mgr_get_aframe(aenc->aframe_mgr)) == NULL) {
#ifndef NDEBUG
      fprintf(stderr, "put_enc_aframe: dafrm == NULL\n");
#endif /* NDEBUG */
      pthread_mutex_lock(&aenc->mutex);
      aenc->state |= AENC_STATE_THREAD_WAIT;
      pthread_cond_wait(&aenc->thread_wait_cond, &aenc->mutex);
      if (aenc->state & AENC_STATE_ERROR) {
        pthread_mutex_unlock(&aenc->mutex);
        return AENC_FAIL;
      }
      pthread_mutex_unlock(&aenc->mutex);
    }
  }

  stream_length = acodec_encode(
      (void**)aenc->src_aframe->sample,
      &aenc->src_aframe->samples,
      aenc->dst_aframe->stream,
      aenc->dst_aframe->stream_buf_size);

  if (stream_length == 0 || stream_length == ACODEC_NEED_MORE_DATA) {
    stream_length = 0;
  }
  else if (stream_length < 0) {
    aenc_error(aenc, "put_enc_aframe: acodec_encode() error %d.\n",stream_length);
    ret = AENC_FAIL;
  }

  aenc->dst_aframe->stream_length = stream_length;
  aenc->dst_aframe->samples = aenc->src_aframe->samples;
  aenc->dst_aframe->flag = aenc->src_aframe->flag;

  if (stream_length > 0)
    add_aframe_chain(aenc, aenc->dst_aframe);
  else
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->dst_aframe);

  aenc->dst_aframe = aenc->src_aframe;
  aenc->src_aframe = NULL;

  return ret;
}

static int
put_flush_aframe(aenc_t *aenc)
{
  int stream_length;
  int ret;

  do {
    if (!aenc->dst_aframe) {
      while((aenc->dst_aframe=aframe_mgr_get_aframe(aenc->aframe_mgr))==NULL) {
#ifndef NDEBUG
        fprintf(stderr, "put_enc_aframe: dafrm == NULL\n");
#endif /* NDEBUG */
        pthread_mutex_lock(&aenc->mutex);
        aenc->state |= AENC_STATE_THREAD_WAIT;
        pthread_cond_wait(&aenc->thread_wait_cond, &aenc->mutex);
        if (aenc->state & AENC_STATE_ERROR) {
          pthread_mutex_unlock(&aenc->mutex);
          return AENC_FAIL;
        }
        pthread_mutex_unlock(&aenc->mutex);
      }
    }

    stream_length = acodec_encode_flush(
      &aenc->dst_aframe->samples,
      aenc->dst_aframe->stream,
      aenc->dst_aframe->stream_buf_size);

    if (stream_length < 0) {
      aenc_error(aenc, "put_enc_aframe: vcodec_encode() error %d.\n",stream_length);
      stream_length = 0;
      ret = AENC_FAIL;
    }
    else if (stream_length > 0) {
      aenc->dst_aframe->stream_length = stream_length;
      aenc->dst_aframe->flag = AFRAME_FLAG_NORMAL;

      add_aframe_chain(aenc, aenc->dst_aframe);
      aenc->dst_aframe = NULL;
      ret = AENC_OK;
    } else {
      ret = AENC_OK;
    }
  } while (stream_length > 0);

  return ret;
}

/* Thread */

static void*
aenc_thread (void *data)
{
  uint32_t state;
  int rec_start_flag = 0;
  int quit_flag = 0;
  int drop_frame_num = 0;
  aenc_t *aenc = (aenc_t*)data;

  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_INTERFACE_WAIT) {
    pthread_cond_signal(&aenc->interface_wait_cond);
  }
  aenc->state = AENC_STATE_STOP;
  pthread_mutex_unlock(&aenc->mutex);

  while (1) {
    state = get_state(aenc);

    if (quit_flag || state & AENC_STATE_ERROR) {
      break;
    }

    pthread_mutex_lock(&aenc->mutex);
    while ((aenc->src_aframe = aread_get_aframe(aenc->aread)) == NULL) {
      aenc->state |= AENC_STATE_THREAD_WAIT;
      pthread_cond_wait(&aenc->thread_wait_cond, &aenc->mutex);
      state = aenc->state;
      if (state & AENC_STATE_ERROR)
	break;
    }
    pthread_mutex_unlock(&aenc->mutex);

    if (state & AENC_STATE_ERROR) {
#ifndef NDEBUG
      fprintf(stderr,"aenc_thread: state AENC_STATE_ERROR.\n");
#endif /* NDEBUG */
      break;
    }

    if (aenc->src_aframe->flag & AFRAME_FLAG_NORMAL) {
      if (put_enc_aframe(aenc) == AENC_FAIL) {
        aenc_error(aenc, "aenc_thread: put_enc_aframe() failed.\n");
	break;
      }
    } else {
      if (aenc->src_aframe->flag & AFRAME_FLAG_START) {
        rec_start_flag = 1;
        set_state(aenc, AENC_STATE_START);
        drop_frame_num = 0;
	state_change_done(aenc);
      }
      else if (aenc->src_aframe->flag & AFRAME_FLAG_END) {
        if (rec_start_flag) {
          if (put_flush_aframe(aenc) != AENC_OK) {
            aenc_error(aenc, "aenc_thread: vcodec_reset() error.\n");
            break;
          }
        }
        rec_start_flag = 0;
        set_state(aenc, AENC_STATE_STOP);
	state_change_done(aenc);
      }
      else if (aenc->src_aframe->flag & AFRAME_FLAG_QUIT) {
        rec_start_flag = 0;
        set_state(aenc, AENC_STATE_QUIT);
        quit_flag = 1;
      }
      else if (aenc->src_aframe->flag & AFRAME_FLAG_ERROR) {
        rec_start_flag = 0;
        set_state(aenc, AENC_STATE_ERROR);
        quit_flag = 1;
      }
      else if (aenc->src_aframe->flag & AFRAME_FLAG_DROP) {
        drop_frame_num++;
      }
      add_aframe_chain(aenc, aenc->src_aframe);
      aenc->src_aframe = NULL;
    }
  }

  if (aenc->dst_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->dst_aframe);
    aenc->dst_aframe = NULL;
  }
  if (aenc->src_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->src_aframe);
    aenc->src_aframe = NULL;
    aenc_error(aenc, "src_aframe remains.\n");
  }

  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_ERROR)
    fprintf(stderr, "aenc_thread: STATE_ERROR break\n");
  if (aenc->state & AENC_STATE_INTERFACE_WAIT) {
    aenc->state &= ~AENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aenc->interface_wait_cond);
  }
  aenc->state &= (AENC_STATE_QUIT|AENC_STATE_ERROR);
  aenc->state |= AENC_STATE_THREAD_FINISH;
  pthread_mutex_unlock(&aenc->mutex);

#ifndef NDEBUG
  fprintf(stderr, "aenc_thread: done\n");
#endif /* NDEBUG */
  return aenc;
}

/* interface functions */

static int
aenc_src_aframe_ready_cb(void *cb_data)
{
  aenc_t *aenc = (aenc_t*)cb_data;

  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_THREAD_WAIT) {
    aenc->state &= ~AENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&aenc->thread_wait_cond);
  }
  pthread_mutex_unlock(&aenc->mutex);
  return AENC_OK;
}

int
aenc_set_aframe_ready_cb(aenc_t *aenc, aframe_ready_cb_t aframe_ready_cb, void *aframe_ready_cb_data)
{
  pthread_mutex_lock(&aenc->mutex);
  aenc->aframe_ready_cb = aframe_ready_cb;
  aenc->aframe_ready_cb_data = aframe_ready_cb_data;
  pthread_mutex_unlock(&aenc->mutex);
  return AENC_OK;
}

int
aenc_set_aframe_mgr(aenc_t *aenc, aframe_mgr_t *afrm_mgr)
{
  pthread_mutex_lock(&aenc->mutex);
  aenc->aframe_mgr = afrm_mgr;
  pthread_mutex_unlock(&aenc->mutex);
  aenc->dst_aframe = aframe_mgr_get_aframe(afrm_mgr);
  return AENC_OK;
}

int
aenc_set_error_cb(aenc_t *aenc, error_cb_t error_cb, void *error_cb_data)
{
  pthread_mutex_lock(&aenc->mutex);
  aenc->error_cb = error_cb;
  aenc->error_cb_data = error_cb_data;
  pthread_mutex_unlock(&aenc->mutex);
  return AENC_OK;
}

aframe_t*
aenc_get_aframe(aenc_t *aenc)
{
  aframe_t *afrm;
  pthread_mutex_lock(&aenc->mutex);
  afrm = aframe_chain_remove(&aenc->aframe_chain);
  pthread_mutex_unlock(&aenc->mutex);
  return afrm;
}

int
aenc_get_frame_sample_size(aenc_t *aenc)
{
  int size;
  pthread_mutex_lock(&aenc->mutex);
  size = aenc->frame_sample_size;
  pthread_mutex_unlock(&aenc->mutex);
  return size;
}

int
aenc_get_out_rate(aenc_t *aenc)
{
  int rate;
  pthread_mutex_lock(&aenc->mutex);
  rate = aenc->out_rate;
  pthread_mutex_unlock(&aenc->mutex);
  return rate;
}

int
aenc_state_error(aenc_t *aenc)
{
  int ret;
  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_THREAD_WAIT) {
    aenc->state &= ~AENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&aenc->thread_wait_cond);
  }
  if (aenc->state & (AENC_STATE_ERROR|AENC_STATE_THREAD_FINISH)) {
    ret = AENC_OK;
  } else {
    aenc->state &= ~(AENC_STATE_START|AENC_STATE_PAUSE|AENC_STATE_STOP);
    aenc->state |= AENC_STATE_ERROR;
#if 0
    aenc->state |= AENC_STATE_INTERFACE_WAIT;
    while (aenc->state & AENC_STATE_INTERFACE_WAIT)
      pthread_cond_wait(&aenc->interface_wait_cond, &aenc->mutex);
#endif
    ret = AENC_OK;
  }
  pthread_mutex_unlock(&aenc->mutex);
  return ret;
}

static int
aenc_state_change_wait(aenc_t *aenc, uint32_t state)
{
  int ret;
  pthread_mutex_lock(&aenc->mutex);
  if (aenc->state & AENC_STATE_THREAD_WAIT) {
    aenc->state &= ~AENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&aenc->thread_wait_cond);
  }
  if (aenc->state & (state|AENC_STATE_QUIT|AENC_STATE_ERROR|AENC_STATE_THREAD_FINISH)) {
    ret = AENC_OK;
  } else {
    while (!(aenc->state & state)) {
      aenc->state |= AENC_STATE_INTERFACE_WAIT;
      pthread_cond_wait(&aenc->interface_wait_cond, &aenc->mutex);
    }
    ret = AENC_OK;
  }
  pthread_mutex_unlock(&aenc->mutex);
  return ret;
}

int
aenc_state_start(aenc_t *aenc)
{
  return aenc_state_change_wait(aenc, AENC_STATE_START);
}

int
aenc_state_stop(aenc_t *aenc)
{
  return aenc_state_change_wait(aenc, AENC_STATE_STOP);
}

int
aenc_state_quit(aenc_t *aenc)
{
  return aenc_state_change_wait(aenc, AENC_STATE_QUIT);
}

int
aenc_state_toggle_pause(aenc_t *aenc)
{
  if (aenc) return AENC_FAIL;
  return AENC_OK;
}

int
aenc_error_quit (aenc_t *aenc)
{
  int retcode, state;
  void* retval;
  aframe_t *afrm;

  aenc_state_error(aenc);

  state = get_state(aenc);
  if (!(state & AENC_STATE_THREAD_FINISH) &&
      !(state & AENC_STATE_NOT_READY)) {
    retcode = pthread_join(aenc->aenc_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "aenc_error_quit: thread join error %d\n", retcode);
  }

  if (aenc->dst_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->dst_aframe);
    aenc->dst_aframe = NULL;
  }
  if (aenc->src_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->src_aframe);
    aenc->src_aframe = NULL;
  }

  pthread_mutex_lock(&aenc->mutex);
  while ((afrm = aframe_chain_remove(&aenc->aframe_chain)) != NULL)
    aframe_mgr_add_free_frame(aenc->aframe_mgr, afrm);
  pthread_mutex_unlock(&aenc->mutex);

  free(aenc);

  return AENC_OK;
}

int
aenc_quit (aenc_t *aenc)
{
  int retcode, state;
  void* retval;

  aenc_state_quit(aenc);

  state = get_state(aenc);
  if (!(state & AENC_STATE_THREAD_FINISH) &&
      !(state & AENC_STATE_NOT_READY)) {
    retcode = pthread_join(aenc->aenc_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "aenc_quit: thread join error %d\n", retcode);
  }

  if (aenc->dst_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->dst_aframe);
    aenc->dst_aframe = NULL;
  }
  if (aenc->src_aframe) {
    aframe_mgr_add_free_frame(aenc->aframe_mgr, aenc->src_aframe);
    aenc->src_aframe = NULL;
  }

  acodec_encode_quit();

  free(aenc);

  return AENC_OK;
}

aenc_t*
aenc_init (int rate, int channels, int bits_per_sample, uint32_t acodec, aread_t *aread)
{
  int retcode;
  int frame_sample_size;
  aenc_t *aenc = NULL;

  if (acodec_encode_init(acodec, rate, channels, bits_per_sample) != ACODEC_OK){
    aenc_error(aenc, "acodec_encode_init failed.\n");
    return NULL;
  }

  frame_sample_size = acodec_encode_get_frame_size();

  aenc = (aenc_t*)malloc(sizeof(aenc_t));
  if (!aenc) {
    aenc_error(aenc, "memory error.\n");
    return NULL;
  }

  aenc->state = AENC_STATE_NOT_READY;
  aenc->src_aframe = NULL;
  aenc->dst_aframe = NULL;
  aenc->aframe_mgr = NULL;
  aenc->aframe_ready_cb = NULL;
  aenc->aframe_ready_cb_data = NULL;
  aframe_chain_init(&aenc->aframe_chain);

  aenc->error_cb = NULL;
  aenc->error_cb_data = NULL;

  aenc->out_rate = acodec_encode_get_out_rate();
  aenc->rate = rate;
  aenc->channels = channels;
  aenc->bits_per_sample = bits_per_sample;
  aenc->acodec = acodec;
  aenc->frame_sample_size = frame_sample_size;
  aenc->aread = aread;

  aread_set_aframe_ready_cb(aread, aenc_src_aframe_ready_cb, aenc);

  pthread_mutex_init (&aenc->mutex, NULL);
  pthread_cond_init (&aenc->interface_wait_cond, NULL);
  pthread_cond_init (&aenc->thread_wait_cond, NULL);

  retcode = pthread_create (&aenc->aenc_th, NULL, aenc_thread, aenc);
  if (retcode != 0) {
    aenc_error(aenc, "aenc_init() thread create failed %d\n", retcode);
    aenc_error_quit(aenc);
    return NULL;
  }

  pthread_mutex_lock(&aenc->mutex);
  while (aenc->state & AENC_STATE_NOT_READY) {
    aenc->state |= AENC_STATE_INTERFACE_WAIT;
    pthread_cond_wait(&aenc->interface_wait_cond, &aenc->mutex);
  }
  pthread_mutex_unlock(&aenc->mutex);

  return aenc;
}

#if 0
void
aenc_print_param(void)
{
  AENC_CONFIG *conf = &aenc_conf;
//  SEL_COMPO *sel;
//  int i;

  fprintf (stderr, "VIDEO PARAM\n");
#if 0
  sel = codecs;
  for (i=0; i < sizeof(codecs)/sizeof(SEL_COMPO); i++, sel++) {
    if (sel->id == conf->fourcc) {
      fprintf (stderr, "video codec:            %s\n", sel->name);
      break;
    }
  }
#endif
  fprintf (stderr, "fps:                    %f\n", conf->fps);
  fprintf (stderr, "dimmention:             %dx%d\n",conf->width,conf->height);
}
#endif

