/* aframe_mgr.c */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <malloc.h>

#include "aframe.h"
#include "aframe_mgr.h"

#define AFRAME_NUM_MAX 512
#define AFRAME_NUM_MIN 5
#define AFRAME_MAX_LAST_FRAME_ALLOC_COUNT (64)
#define AFRAME_MAX_EXT_FRAME_NUM 256

static void error(const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "aframe_mgr: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);
  exit(1);
}

static aframe_t*
alloc_aframe(aframe_mgr_t *afrm_mgr, int buf_alloc_flag)
{
  int i;
  aframe_t *afrm;

  if (buf_alloc_flag) {
    if (afrm_mgr->alloc_frame_num >= afrm_mgr->max_frame_num)
      return NULL;
  } else if (afrm_mgr->alloc_ext_frame_num >= afrm_mgr->max_ext_frame_num)
    return NULL;

  afrm = (aframe_t*)malloc(sizeof(aframe_t));
  if (!afrm) {
    pthread_mutex_unlock(&afrm_mgr->mutex);
    error("memory allocate error.\n");
  }

  if (buf_alloc_flag) {
#ifdef HAVE_MEMALIGN
    afrm->bufp =
      (uint8_t*)memalign(16, afrm_mgr->aframe_buf_byte_size*afrm_mgr->channels);
    afrm->stream = afrm->bufp;
#else
    afrm->bufp =
      (uint8_t*)malloc(afrm_mgr->aframe_buf_byte_size*afrm_mgr->channels+15);
    afrm->stream = (uint8_t*)((((uint32_t)(afrm->bufp)) + 15) & ~0xf);
#endif
    if (!afrm->bufp) {
      free(afrm);
      pthread_mutex_unlock(&afrm_mgr->mutex);
      error("memory allocate error.\n");
    }

    afrm->sample[0] = afrm->stream;
    for (i = 1; i < afrm_mgr->channels; i++)
      afrm->sample[i] = afrm->sample[i-1] + afrm_mgr->aframe_buf_byte_size;

    afrm->rate = afrm_mgr->rate;
    afrm->channels = afrm_mgr->channels;
    afrm->bits_per_sample = afrm_mgr->bits_per_sample;
    afrm->samples = 0;
    afrm->sample_buf_size = afrm_mgr->aframe_buf_sample_size;
    afrm->stream_length = 0;
    afrm->stream_buf_size = afrm_mgr->aframe_buf_byte_size;
    afrm->flag = AFRAME_FLAG_NORMAL;

    afrm_mgr->alloc_frame_num++;
    afrm_mgr->last_frame_alloc_count = 0;
  } else {
    afrm->bufp = afrm->stream = NULL;
    for (i = 0; i < afrm_mgr->channels; i++)
      afrm->sample[i] = NULL;
    afrm->rate = 0;
    afrm->channels = 0;
    afrm->bits_per_sample = 0;
    afrm->samples = 0;
    afrm->sample_buf_size = 0;
    afrm->stream_length = 0;
    afrm->stream_buf_size = 0;
    afrm->flag = AFRAME_FLAG_EXT;
    afrm_mgr->alloc_ext_frame_num++;
//    afrm_mgr->last_frame_alloc_count = 0;
  }

  afrm->prev = afrm->next = NULL;

  return afrm;
}

static void
free_aframe(aframe_mgr_t *afrm_mgr, aframe_t *afrm)
{
#ifndef NDEBUG
  if (afrm->prev || afrm->next) {
    error("free_aframe: frame doesn't removed chain.\n");
  }
#endif /* NDEBUG */
  if (afrm->bufp) {
    free (afrm->bufp);
    afrm_mgr->alloc_frame_num--;
    afrm_mgr->last_frame_alloc_count = 0;
#ifndef NDEBUG
    if (afrm_mgr->alloc_frame_num < 0)
      error("free_aframe: alloc_frame_num %d\n", afrm_mgr->alloc_frame_num);
#endif /* NDEBUG */
  } else {
    afrm_mgr->alloc_ext_frame_num--;
//    afrm_mgr->last_frame_alloc_count = 0;
#ifndef NDEBUG
    if (afrm_mgr->alloc_ext_frame_num < 0)
      error("free_aframe: alloc_ext_frame_num %d\n", afrm_mgr->alloc_ext_frame_num);
#endif /* NDEBUG */
  }
  free(afrm);
}

int
aframe_mgr_add_free_frame(aframe_mgr_t *afrm_mgr, aframe_t *afrm)
{
  int ret;

  pthread_mutex_lock(&afrm_mgr->mutex);
  if (!afrm->bufp) {
    free_aframe(afrm_mgr, afrm);
    pthread_mutex_unlock(&afrm_mgr->mutex);
    return 0;
  }

  if (afrm_mgr->alloc_frame_num > AFRAME_NUM_MIN &&
      afrm_mgr->last_frame_alloc_count > afrm_mgr->max_last_frame_alloc_count &&
      afrm_mgr->free_frame_num > 0) {
    free_aframe(afrm_mgr, afrm);
    afrm_mgr->last_frame_alloc_count = 0;
    ret = 1;
  } else {
    afrm->flag = AFRAME_FLAG_NORMAL;
    afrm->samples = afrm->stream_length = 0;
//    afrm->exflag = 0;
    aframe_chain_add(&afrm_mgr->free_chain, afrm);
    afrm_mgr->free_frame_num++;
    afrm_mgr->last_frame_alloc_count++;
    ret = 0;
  }
  afrm_mgr->used_frame_num--;
  pthread_mutex_unlock(&afrm_mgr->mutex);

  return ret;
}

aframe_t*
aframe_mgr_get_ext_aframe(aframe_mgr_t *afrm_mgr)
{
  aframe_t *afrm = NULL;
  pthread_mutex_lock(&afrm_mgr->mutex);
  afrm = alloc_aframe(afrm_mgr, 0);
  pthread_mutex_unlock(&afrm_mgr->mutex);
  return afrm;
}

aframe_t*
aframe_mgr_get_aframe(aframe_mgr_t *afrm_mgr)
{
  aframe_t *afrm = NULL;

  pthread_mutex_lock(&afrm_mgr->mutex);
//  if (afrm_mgr->used_frame_num < afrm_mgr->max_frame_num-1) {
    if (afrm_mgr->free_chain.aframes) {
      afrm = aframe_chain_remove(&(afrm_mgr->free_chain));
      afrm_mgr->free_frame_num--;
      afrm_mgr->used_frame_num++;
    } else {
      afrm = alloc_aframe(afrm_mgr, 1);
      if (afrm)
        afrm_mgr->used_frame_num++;
    }
//  }
#ifndef NDEBUG
  if (!afrm) {
    fprintf(stderr, "aframe_mgr_get_aframe: frame_num %d, free_frame_num %d\n", afrm_mgr->alloc_frame_num, afrm_mgr->free_frame_num);
  }
#endif /* NDEBUG */
  pthread_mutex_unlock(&afrm_mgr->mutex);

  return afrm;
}

int
aframe_mgr_quit(aframe_mgr_t *afrm_mgr)
{
  aframe_t *afrm;

  pthread_mutex_lock(&afrm_mgr->mutex);
  while (afrm_mgr->free_chain.aframes) {
    afrm = aframe_chain_remove(&afrm_mgr->free_chain);
    free_aframe(afrm_mgr, afrm);
  }
  pthread_mutex_unlock(&afrm_mgr->mutex);
//  pthread_mutex_destroy(&afrm_mgr->mutex);
  free(afrm_mgr);
  return AFRAME_MGR_OK;
}

aframe_mgr_t*
aframe_mgr_init(int rate, int channels, int bits_per_sample, int frame_sample_size, double buffer_time)
{
  int i;
  int max_frame_num;
  int bytes_per_sample;
  aframe_t *afrm;
  aframe_mgr_t *afrm_mgr;

  if (frame_sample_size & 0xf) {
    error("aframe_mgr_init: frame_sample_size isn't multiple of 16, %d\n", frame_sample_size);
  }

  afrm_mgr = (aframe_mgr_t*)malloc(sizeof(aframe_mgr_t));
  if (!afrm_mgr)
    error("memory allocate error.\n");

  max_frame_num =
    (int)((double)rate*buffer_time+frame_sample_size-1) / frame_sample_size;
  if (max_frame_num < AFRAME_NUM_MIN)
    max_frame_num = AFRAME_NUM_MIN;
  else if (max_frame_num > AFRAME_NUM_MAX)
    max_frame_num = AFRAME_NUM_MAX;

  bytes_per_sample = ((bits_per_sample+7)/8) * channels;

  afrm_mgr->rate = rate;
  afrm_mgr->channels = channels;
  afrm_mgr->bits_per_sample = bits_per_sample;
  afrm_mgr->aframe_buf_sample_size = frame_sample_size;
  afrm_mgr->bytes_per_sample = bytes_per_sample;
  afrm_mgr->aframe_buf_byte_size = frame_sample_size * bytes_per_sample;

  afrm_mgr->last_frame_alloc_count = 0;
  afrm_mgr->max_last_frame_alloc_count = AFRAME_MAX_LAST_FRAME_ALLOC_COUNT;
  afrm_mgr->max_frame_num = max_frame_num;
  afrm_mgr->max_ext_frame_num = AFRAME_MAX_EXT_FRAME_NUM;
  afrm_mgr->alloc_frame_num = 0;
  afrm_mgr->alloc_ext_frame_num = 0;
  afrm_mgr->free_frame_num = 0;
  afrm_mgr->used_frame_num = 0;

  aframe_chain_init(&afrm_mgr->free_chain);

  pthread_mutex_init(&afrm_mgr->mutex, NULL);

  for (i = 0; i < AFRAME_NUM_MIN; i++) {
    afrm = alloc_aframe(afrm_mgr, 1);
    aframe_chain_add(&afrm_mgr->free_chain, afrm);
    afrm_mgr->free_frame_num++;
  }

  return afrm_mgr;
}

