/* aread.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
//#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include "get_cur_u_time.h"
#include "oss.h"
#include "aframe_mgr.h"
#include "aread.h"
#include "error_mgr.h"

#define BUFFER_TIME 2.0
#define CHECK_TIME    10.0

//#define INTERP_THRESHOLD_MIN -32
#define INTERP_THRESHOLD_MIN -0
//#define INTERP_THRESHOLD_MAX  32 
#define INTERP_THRESHOLD_MAX  0 

#define READ_FRAME_COUNT_MAX  600
#define READ_FRAME_MAX  1024
#define READ_FRAME_MIN  2

/* Thread functions */

static void aread_error(aread_t *aread, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "aread: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);

  if (!aread) return;

  pthread_mutex_lock(&aread->mutex);
  aread->state = AREAD_STATE_ERROR;
  pthread_mutex_unlock(&aread->mutex);
  if (aread->error_cb)
    aread->error_cb(aread->error_cb_data);
//  exit(1);
}

static int
sample_read(aread_t *aread)
{
  int idx;
  int r;

  if (aread->record_flag &&
     (aread->read_frame_count-aread->read_frame_cur) >= aread->read_frame_num) {

    fprintf(stderr, "sample_read: read_frame overrun. count %d, cur %d, num %d\n", aread->read_frame_count, aread->read_frame_cur, aread->read_frame_num);

    aread->read_frame_cur = aread->read_frame_count - aread->read_frame_num + 1;
    aread->read_frame_cur_offset = 0;
  }

  idx = aread->read_frame_count % aread->read_frame_num;
  r = oss_read_data(aread->oss,aread->read_frame[idx],aread->read_frame_byte_size);
  if (r != aread->read_frame_byte_size) {
    aread_error(aread, "oss_read_data failed, %d, %d\n", r, aread->read_frame_byte_size);
    return AREAD_FAIL;
  }
  aread->read_frame_time[idx] = get_cur_u_time();

  aread->read_frame_count++;

  return AREAD_OK;
}

static int
get_pos(aread_t *aread, u_time_t time, int *pos_ret, int *offset_ret)
{
  int i, idx;
  int32_t pos, offset;
  int32_t count, frame_num;
  u_time_t *frame_time;
  double offset_time, rate;

  count = aread->read_frame_count;
  if (count <= 0) {
    do {
      sample_read(aread);
      count = aread->read_frame_count;
    } while (count <= 0);
  }
  frame_time = aread->read_frame_time;
  frame_num = aread->read_frame_num;
  frame_num = (count < frame_num) ? count : frame_num;

  idx = (count - 1) % frame_num;
  if (frame_time[idx] <= time) {
    *pos_ret = count;
    *offset_ret = 0;
    return AREAD_FAIL;
  }

  for (i = 1; i < frame_num; i++) {
    idx = (count - i) % frame_num;
    if (frame_time[idx] <= time) {
      i--;
      break;
    }
  }
  pos = count - i;

  idx = pos % frame_num;
  /* must be frame_time[idx] >= time, becouse u_time_t is unsigned. */
  offset_time = frame_time[idx] - time;
  rate = aread->rate;
  offset = (int)(rate * offset_time / ((double)U_TIME_SECOND));
  offset = aread->read_frame_sample_size - offset;
  if (offset < 0) {
#ifndef NDEBUG
    fprintf(stderr, "get_pos: offset value is negative, %d.\n", offset);
#endif
    offset = 0;
  }

  *pos_ret = pos;
  *offset_ret = offset;

  return AREAD_OK;
}

static int
add_frame_chain(aread_t *aread, aframe_t *afrm)
{
  int ret;
  pthread_mutex_lock(&aread->mutex);
  ret = aframe_chain_add(&aread->aframe_chain, afrm);
  pthread_mutex_unlock(&aread->mutex);
  if (aread->aframe_ready_cb)
    aread->aframe_ready_cb(aread->aframe_ready_cb_data);
  return ret;
}

static int
sample_cpy(aread_t *aread)
{
  int rd_len, af_len;
  int num_samples;

  rd_len = aread->read_frame_sample_size - aread->read_frame_cur_offset;
  af_len = aread->aframe->sample_buf_size - aread->aframe->samples;
  num_samples = (rd_len < af_len) ? rd_len : af_len;

  if (aread->bits_per_sample == 16) {
    int c, channels = aread->channels;
    int i;
    short *src, *sp, *dp;

    src = (short*)
      aread->read_frame[aread->read_frame_cur % aread->read_frame_num];
    src += aread->read_frame_cur_offset * channels;

    for (c = 0; c < channels; c++) {
      dp = (short*)aread->aframe->sample[c];
      dp += aread->aframe->samples;
      sp = src + c;
      for (i = 0; i < num_samples; i++) {
	*dp = *sp;
	sp += channels;
	dp++;
      }
    }
  } else if (aread->bits_per_sample == 8) {
    int c, channels = aread->channels;
    int i;
    char *src, *sp, *dp;

    src = (char*)
      aread->read_frame[aread->read_frame_cur % aread->read_frame_num];
    src += aread->read_frame_cur_offset * channels;

    for (c = 0; c < channels; c++) {
      dp = (char*)aread->aframe->sample[c];
      dp += aread->aframe->samples;
      sp = src + c;
      for (i = 0; i < num_samples; i++) {
	*dp = *sp;
	sp += channels;
	dp++;
      }
    }
  } else {
    aread_error(aread, "sample_cpy: unsupported bits per sample.\n");
    return AREAD_FAIL;
  }

  aread->read_frame_cur_offset += num_samples;
  aread->aframe->samples += num_samples;

  if ((rd_len - num_samples) <= 0) {
    aread->read_frame_cur++;
    aread->read_frame_cur_offset = 0;
  }
  if ((af_len - num_samples) <= 0) {
    add_frame_chain(aread, aread->aframe);
    aread->aframe = NULL;
  }

  return num_samples;
}

static int
sample_put_start(aread_t *aread, u_time_t start_time)
{
  int pos, offset;

  while (get_pos(aread, start_time, &pos, &offset) == AREAD_FAIL) {
    sample_read(aread);
  }

  offset += aread->gap;
  while (offset >= aread->read_frame_sample_size) {
    pos++;
    offset -= aread->read_frame_sample_size;
  }
  while (offset < 0) {
    pos--;
    offset += aread->read_frame_sample_size;
  }
  if (pos < 0) {
    pos = 0;
    offset = 0;
  }

  while (pos >= aread->read_frame_count) {
    sample_read(aread);
  }

  aread->gap = 0;

  aread->read_frame_cur = pos;
  aread->read_frame_cur_offset = offset;
  aread->record_flag = 1;

  return AREAD_OK;
}

static int
sample_put(aread_t *aread)
{
  int spls;

  while (aread->read_frame_cur < (aread->read_frame_count-1)) {
    if (!aread->aframe) {
      aread->aframe = aframe_mgr_get_aframe(aread->aframe_mgr);
      if (!aread->aframe) {
        fprintf(stderr, "aread: aframe_mgr_get_aframe failed.\n");
	return AREAD_FAIL;
      }
      aread->aframe->flag = AFRAME_FLAG_NORMAL;
      aread->aframe->samples = 0;
    }

    spls = sample_cpy(aread);
    aread->record_samples += spls;
  }

  return AREAD_OK;
}

static int
sample_put_gap(aread_t *aread)
{
  int spls;

//fprintf(stderr, "aread: sample gap %d\n", aread->gap);

  while (aread->read_frame_cur < (aread->read_frame_count-1)) {
    if (!aread->aframe) {
      aread->aframe = aframe_mgr_get_aframe(aread->aframe_mgr);
      if (!aread->aframe) {
        fprintf(stderr, "aread: aframe_mgr_get_aframe failed.\n");
	return AREAD_FAIL;
      }
      aread->aframe->flag = AFRAME_FLAG_NORMAL;
      aread->aframe->samples = 0;
    }

    if (aread->read_frame_cur >= aread->next_interp_frame_count &&
	  aread->read_frame_cur_offset == 0) {
      if (aread->gap < INTERP_THRESHOLD_MIN) {
        if (aread->bits_per_sample == 16) {
          int c, channels = aread->channels;
          short *sp, *src;

          src = (short*)
            aread->read_frame[aread->read_frame_cur % aread->read_frame_num];
          for (c = 0; c < channels; c++) {
            sp = src + c;
	    sp[channels] = (sp[0] + sp[channels] + 1) >> 1;
          }
        } else if (aread->bits_per_sample == 8) {
          int c, channels = aread->channels;
          char *sp, *src;
          src = (char*)
            aread->read_frame[aread->read_frame_cur % aread->read_frame_num];

          for (c = 0; c < channels; c++) {
            sp = src + c;
	    sp[channels] = (sp[0] + sp[channels] + 1) >> 1;
          }
        } else {
          aread_error(aread, "sample_pu_gap: unsupported bits per sample.\n");
	  return AREAD_FAIL;
        }

	aread->read_frame_cur_offset++;
	aread->gap++;
      } else if (aread->gap > INTERP_THRESHOLD_MAX) {
        if (aread->bits_per_sample == 16) {
          int c, channels = aread->channels;
          short *sp, *dp, *src;

          src = (short*)
            aread->read_frame[aread->read_frame_cur % aread->read_frame_num];
          for (c = 0; c < channels; c++) {
            dp = (short*)aread->aframe->sample[c];
            dp += aread->aframe->samples;
            *dp = src[c];
          }
          for (c = 0; c < channels; c++) {
            sp = src + c;
	    sp[0] = (sp[0] + sp[channels]) >> 1;
          }
        } else if (aread->bits_per_sample == 8) {
          int c, channels = aread->channels;
          char *sp, *dp, *src;

          src = (char*)
            aread->read_frame[aread->read_frame_cur % aread->read_frame_num];
          for (c = 0; c < channels; c++) {
            dp = (char*)aread->aframe->sample[c];
            dp += aread->aframe->samples;
            *dp = src[c];
          }
          for (c = 0; c < channels; c++) {
            sp = src + c;
	    sp[0] = (sp[0] + sp[channels] + 1) >> 1;
          }
        } else {
          aread_error(aread, "sample_pu_gap: unsupported bits per sample.\n");
	  return AREAD_FAIL;
        }

	aread->aframe->samples++;
	aread->gap--;
      }
    }

    if (aread->aframe->samples < aread->aframe->sample_buf_size) {
      spls = sample_cpy(aread);
      aread->record_samples += spls;
    }
    else {
      add_frame_chain(aread, aread->aframe);
      aread->aframe = NULL;
    }
  }

  return AREAD_OK;
}

static int
sample_put_end(aread_t *aread, u_time_t end_time)
{
  int pos, offset;
  int spls;
  int end_gap;

  while (get_pos(aread, end_time, &pos, &offset) == AREAD_FAIL) {
    sample_put(aread);
    sample_read(aread);
  }

  offset += aread->gap;
  while (offset >= aread->read_frame_sample_size) {
    pos++;
    offset -= aread->read_frame_sample_size;
  }

//  while (aread->aframe || aread->read_frame_cur < pos ||
//      (aread->read_frame_cur==pos && aread->read_frame_cur_offset<offset)) {
  while (1) {
    while (!aread->aframe) {
      aread->aframe = aframe_mgr_get_aframe(aread->aframe_mgr);
      if (!aread->aframe) {
        fprintf(stderr, "aread: aframe_mgr_get_aframe failed.\n");
        sample_read(aread);
      }
      aread->aframe->flag = AFRAME_FLAG_NORMAL;
      aread->aframe->samples = 0;
    }

    spls = sample_cpy(aread);
    aread->record_samples += spls;

    if (!aread->aframe) {
      if (aread->read_frame_cur > pos ||
         (aread->read_frame_cur == pos && aread->read_frame_cur_offset > offset)) {
	break;
      }
    }
  }

  end_gap = (aread->read_frame_cur - pos) * aread->read_frame_sample_size;
  end_gap += aread->read_frame_cur_offset - offset;
  aread->gap = end_gap;

  return AREAD_OK;
}

static int
sample_put_flush(aread_t *aread)
{
  int num_samples;
  int c, channels = aread->channels;

  if (aread->aframe && aread->aframe->samples > 0) {
    num_samples = aread->aframe->sample_buf_size - aread->aframe->samples;
    if (aread->bits_per_sample == 16) {
      short *dp;

      for (c = 0; c < channels; c++) {
        dp = (short*)aread->aframe->sample[c];
        dp += aread->aframe->samples;
	memset(dp, 0, num_samples * sizeof(short));
      }
    } else if (aread->bits_per_sample == 8) {
      char *dp;

      for (c = 0; c < channels; c++) {
        dp = (char*)aread->aframe->sample[c];
        dp += aread->aframe->samples;
	memset(dp, 128, num_samples * sizeof(char));
      }
    } else {
      aread_error(aread, "sample_cpy: unsupported bits per sample.\n");
      return AREAD_FAIL;
    }
    aread->aframe->samples += num_samples;
    add_frame_chain(aread, aread->aframe);
    aread->aframe = NULL;
  }

  return AREAD_OK;
}

static void
calc_gap(aread_t *aread)
{
  int count, idx;
  int interp_frame_interval;
  double rate;
  double read_time;
  double time_samples;
  double read_samples;
  double gap;

  count = aread->read_frame_count;
  idx = (count-1) % aread->read_frame_num;
  read_time = aread->read_frame_time[idx] - aread->read_start_time;
  read_time = read_time / ((double)U_TIME_SECOND);
  read_samples =
    (count - aread->read_start_count) * aread->read_frame_sample_size;

  rate = aread->rate;
  time_samples = rate * read_time;
  gap = time_samples - read_samples;
  aread->gap += (gap - aread->read_gap);
  aread->read_gap = gap;

  if (aread->gap) {
    interp_frame_interval = (aread->check_count_interval / ((aread->gap<0) ? -aread->gap : aread->gap)) - 1;
    if (interp_frame_interval < 1) {
      fprintf(stderr, "aread: interp_frame_interval < 1, %d.\n",
         	interp_frame_interval);
      interp_frame_interval = 1;
    }
  } else {
    interp_frame_interval = aread->check_count_interval;
  }
  aread->interp_frame_interval = interp_frame_interval;
  aread->next_interp_frame_count =
    aread->read_frame_cur + interp_frame_interval;
}

static int
put_ext_aframe(aread_t *aread, uint32_t flag)
{
  aframe_t *afrm;

#ifndef NDEBUG
  if (aread->aframe) {
    aread_error(aread, "put_ext_aframe: aframe remains.\n");
  }
#endif /* NDEBUG */
  afrm = aframe_mgr_get_ext_aframe(aread->aframe_mgr);
  if (!afrm) {
    aread_error(aread, "aframe_mgr_get_ext_aframe failed.\n");
    return AREAD_FAIL;
  }
  afrm->flag = flag | AFRAME_FLAG_EXT;
  afrm->samples = 0;
  afrm->stream_length = 0;
  add_frame_chain(aread, afrm);
  return AREAD_OK;
}

static inline int
get_state(aread_t *aread)
{
  int state;
  pthread_mutex_lock(&aread->mutex);
  state = aread->state;
  pthread_mutex_unlock(&aread->mutex);
  return state;
}

static void
set_state(aread_t *aread, uint32_t state)
{
  pthread_mutex_lock(&aread->mutex);
  aread->state &= ~(AREAD_STATE_START|AREAD_STATE_PAUSE|AREAD_STATE_STOP|AREAD_STATE_QUIT);
  aread->state |= state;
  if (aread->state & AREAD_STATE_INTERFACE_WAIT) {
    aread->state &= ~AREAD_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aread->interface_wait_cond);
  }
  pthread_mutex_unlock(&aread->mutex);
}

static void
state_change_done(aread_t *aread)
{
  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_INTERFACE_WAIT) {
    aread->state &= ~AREAD_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aread->interface_wait_cond);
  }
  pthread_mutex_unlock(&aread->mutex);
}

/* Thread */

static void*
aread_thread(void *data)
{
  aread_t *aread = (aread_t*)data;
  int state;
  int rec_start_flag = 0;

  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_INTERFACE_WAIT) {
    pthread_cond_signal(&aread->interface_wait_cond);
  }
  aread->state = AREAD_STATE_STOP;
  pthread_mutex_unlock(&aread->mutex);

  while (1) {
    state = get_state(aread);

    if (state & AREAD_STATE_START) {
      if (!rec_start_flag) {
        if (put_ext_aframe(aread, AFRAME_FLAG_START) != AREAD_OK) {
          aread_error(aread, "aread_thread: put_ext_aframe() failed.\n");
          break;
        } else {
          rec_start_flag = 1;
	  aread->record_time = 0;
	  aread->record_samples = 0;

	  aread->read_start_time = 0;
	  aread->read_start_count = 0;
	  aread->next_check_count = 0;
	  aread->gap = 0;
        }
      }

      sample_put_start(aread, aread->state_time);

      if (aread->sync_system_time_flag) {
	int idx;
        aread->read_start_count = aread->read_frame_count;
        idx = (aread->read_frame_count-1) % aread->read_frame_num;
        aread->read_start_time = aread->read_frame_time[idx];
	aread->read_gap = 0;
	aread->interp_frame_interval = aread->check_count_interval;
	aread->next_interp_frame_count =
	  aread->read_frame_count + aread->check_count_interval;
      }

      aread->next_check_count =
       	aread->read_frame_count + aread->check_count_interval;
      aread->record_start_time = aread->state_time;
      aread->record_flag = 1;
      state_change_done(aread);

      while (state & AREAD_STATE_START) {
	if (aread->gap)
          sample_put_gap(aread);
	else
          sample_put(aread);

	sample_read(aread);

	if (aread->read_frame_count >= aread->next_check_count) {
	  if (aread->sync_system_time_flag)
	    calc_gap(aread);
	  else
	    aread->gap = 0;

	  if (aread->read_frame_count >= READ_FRAME_COUNT_MAX) {
	    int sub;
	    sub = aread->read_frame_cur % aread->read_frame_num;
	    sub = aread->read_frame_cur - sub;

	    aread->read_frame_count -= sub;
	    aread->read_frame_cur   -= sub;

	    if (aread->sync_system_time_flag) {
	      int idx;
	      aread->read_start_count = aread->read_frame_count;
	      idx = (aread->read_frame_count-1) % aread->read_frame_num;
	      aread->read_start_time = aread->read_frame_time[idx];
	      aread->read_gap = 0;

	      aread->next_interp_frame_count -= sub;
	    }
	  }

	  aread->next_check_count =
	    aread->read_frame_count + aread->check_count_interval;
	}

        state = get_state(aread);
      }

      sample_put_end(aread, aread->state_time);
      aread->record_flag = 0;
      aread->record_time += aread->state_time - aread->record_start_time;
    }
    else if (state & AREAD_STATE_PAUSE) {
      state_change_done(aread);
      while (state & AREAD_STATE_PAUSE) {
	sample_read(aread);
        state = get_state(aread);
      }
    }
    else if (state & AREAD_STATE_STOP) {
      if (rec_start_flag) {
	sample_put_flush(aread);
        if (put_ext_aframe(aread, AFRAME_FLAG_END) == AREAD_FAIL) {
          aread_error(aread, "aread_thread: put_ext_aframe() failed.\n");
          break;
        }
        rec_start_flag = 0;
      }
      state_change_done(aread);

      while (state & AREAD_STATE_STOP) {
	sample_read(aread);
        state = get_state(aread);
      }
    }
    else if (state & AREAD_STATE_QUIT) {
      if (rec_start_flag) {
	sample_put_flush(aread);
        if (put_ext_aframe(aread, AFRAME_FLAG_END) == AREAD_FAIL) {
          aread_error(aread, "aread_thread: put_ext_aframe() failed.\n");
          break;
        }
        rec_start_flag = 0;
      }
      if (put_ext_aframe(aread, AFRAME_FLAG_QUIT) == AREAD_FAIL) {
        aread_error(aread, "aread_thread: put_ext_aframe() failed.\n");
        break;
      }
      break;
    }
    else if (state & AREAD_STATE_ERROR) {
      if (put_ext_aframe(aread, AFRAME_FLAG_END) == AREAD_FAIL) {
        aread_error(aread, "aread_thread: put_ext_aframe() failed.\n");
        break;
      }
      break;
    }
  }

  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_ERROR)
    fprintf(stderr, "aread_thread: STATE_ERROR break\n");
  if (aread->state & AREAD_STATE_INTERFACE_WAIT) {
    aread->state &= ~AREAD_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&aread->interface_wait_cond);
  }
  aread->state &= (AREAD_STATE_QUIT|AREAD_STATE_ERROR);
  aread->state |= AREAD_STATE_THREAD_FINISH;
  pthread_mutex_unlock(&aread->mutex);
//  state_change_done(aread);

#ifndef NDEBUG
  fprintf(stderr, "aread_thread: done\n");
#endif
  return aread;
}

/* Interface functions */

int
aread_get_rate(aread_t *aread)
{
  return aread->rate;
}

int
aread_get_channels(aread_t *aread)
{
  return aread->channels;
}

int
aread_get_bits_per_sample(aread_t *aread)
{
  return aread->bits_per_sample;
}

aframe_t*
aread_get_aframe(aread_t *aread)
{
  aframe_t *afrm;
  pthread_mutex_lock(&aread->mutex);
  afrm = aframe_chain_remove(&aread->aframe_chain);
  pthread_mutex_unlock(&aread->mutex);
  return afrm;
}

int
aread_set_aframe_ready_cb(aread_t *aread, aframe_ready_cb_t aframe_ready_cb, void *aframe_ready_cb_data)
{
  pthread_mutex_lock(&aread->mutex);
  aread->aframe_ready_cb = aframe_ready_cb;
  aread->aframe_ready_cb_data = aframe_ready_cb_data;
  pthread_mutex_unlock(&aread->mutex);
  return AREAD_OK;
}

int
aread_set_aframe_mgr(aread_t *aread, aframe_mgr_t *afrm_mgr)
{
  pthread_mutex_lock(&aread->mutex);
  aread->aframe_mgr = afrm_mgr;
  pthread_mutex_unlock(&aread->mutex);
  return AREAD_OK;
}

int
aread_set_error_cb(aread_t *aread, error_cb_t error_cb, void *error_cb_data)
{
  pthread_mutex_lock(&aread->mutex);
  aread->error_cb = error_cb;
  aread->error_cb_data = error_cb_data;
  pthread_mutex_unlock(&aread->mutex);
  return AREAD_OK;
}

static int
aread_state_change(aread_t *aread, u_time_t *state_time, int state)
{
  int ret = AREAD_FAIL;

  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_THREAD_WAIT) {
    aread->state &= ~AREAD_STATE_THREAD_WAIT;
    pthread_cond_signal(&aread->thread_wait_cond);
  }
  if (aread->state & (state|AREAD_STATE_QUIT|AREAD_STATE_ERROR|AREAD_STATE_THREAD_FINISH)) {
    ret = AREAD_OK;
  } else {
    aread->state &= ~(AREAD_STATE_START|AREAD_STATE_PAUSE|AREAD_STATE_STOP);
    aread->state |= state;
    aread->state |= AREAD_STATE_INTERFACE_WAIT;
    aread->state_time = *state_time;
    while (aread->state & AREAD_STATE_INTERFACE_WAIT)
      pthread_cond_wait(&aread->interface_wait_cond, &aread->mutex);
    *state_time = aread->state_time;
    ret = AREAD_OK;
  }
  pthread_mutex_unlock(&aread->mutex);
  return ret;
}

int
aread_state_start(aread_t *aread, u_time_t *state_time)
{
  return aread_state_change(aread, state_time, AREAD_STATE_START);
}

int
aread_state_stop(aread_t *aread, u_time_t *state_time)
{
  return aread_state_change(aread, state_time, AREAD_STATE_STOP);
}

int
aread_state_toggle_pause(aread_t *aread, u_time_t *state_time)
{
  int state;

  state = get_state(aread);
  if (state & AREAD_STATE_START)
    return aread_state_change(aread, state_time, AREAD_STATE_PAUSE);
  else if (state & AREAD_STATE_PAUSE)
    return aread_state_change(aread, state_time, AREAD_STATE_START);
  return AREAD_OK;
}

int
aread_state_quit(aread_t *aread, u_time_t *state_time)
{
  return aread_state_change(aread, state_time, AREAD_STATE_QUIT);
}

int
aread_state_error(aread_t *aread, u_time_t *state_time)
{
  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_THREAD_WAIT) {
    aread->state &= ~AREAD_STATE_THREAD_WAIT;
    pthread_cond_signal(&aread->thread_wait_cond);
  }
  if (aread->state & (AREAD_STATE_ERROR|AREAD_STATE_THREAD_FINISH)) {
    pthread_mutex_unlock(&aread->mutex);
    return AREAD_OK;
  }
  aread->state = AREAD_STATE_ERROR;
  *state_time = *state_time;
  pthread_mutex_unlock(&aread->mutex);
  return AREAD_OK;
}

int
aread_error_quit (aread_t *aread, u_time_t *quit_time)
{
  int retcode, state;
  void* retval;
  aframe_t *afrm;

  aread_state_error(aread, quit_time);

  state = get_state(aread);
  if (!(state & AREAD_STATE_THREAD_FINISH) &&
      !(state & AREAD_STATE_NOT_READY)) {
    retcode = pthread_join(aread->aread_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "aread_error_quit: thread join error %d\n", retcode);
  }

  if (aread->read_frame) {
    if (aread->read_frame[0])
      free(aread->read_frame[0]);
    free(aread->read_frame);
  }
  if (aread->read_frame_time)
    free(aread->read_frame_time);

  if (aread->oss)
    oss_quit(aread->oss);

  if (aread->aframe)
    aframe_mgr_add_free_frame(aread->aframe_mgr, aread->aframe);

  pthread_mutex_lock(&aread->mutex);
  while ((afrm = aframe_chain_remove(&aread->aframe_chain)) != NULL)
    aframe_mgr_add_free_frame(aread->aframe_mgr, afrm);
  pthread_mutex_unlock(&aread->mutex);

  free(aread);

  return AREAD_OK;
}


int
aread_quit (aread_t *aread, u_time_t *quit_time)
{
  int retcode;
  void* retval;

  aread_state_quit(aread, quit_time);

  if (aread->state != AREAD_STATE_NOT_READY) {
    retcode = pthread_join(aread->aread_th, &retval);
    if (retcode != 0)
      fprintf (stderr, "aread_quit() thread join error %d\n", retcode);
//    pthread_mutex_destroy(&aread->mutex);
  }
  
  if (aread->read_frame) {
    if (aread->read_frame[0])
      free(aread->read_frame[0]);
    free(aread->read_frame);
  }
  if (aread->read_frame_time)
    free(aread->read_frame_time);

  if (aread->oss)
    oss_quit(aread->oss);

  if (aread->aframe)
    aframe_mgr_add_free_frame(aread->aframe_mgr, aread->aframe);

  free(aread);

  return AREAD_OK;
}

aread_t*
aread_init (int rate, int channels, int bits_per_sample, int sync_system_time, const char *device_name, const char *mixer_name, const char *source_name)
{
  int retcode;
  int fragsize;
  int fragstotal;
  int bytespspl;
  double buffer_time = BUFFER_TIME;
  double check_time = CHECK_TIME;
  int buffer_size;
  int frame_num;
  int i;
  u_time_t dummy_time = 0;
  aread_t *aread;

  aread = (aread_t*)malloc(sizeof(aread_t));
  if (!aread) {
    aread_error(aread, "memory allocate error.\n");
    return NULL;
  }

  aread->state = AREAD_STATE_NOT_READY;
  aread->read_frame = NULL;
  aread->read_frame_time = NULL;
  aread->aframe = NULL;
  aread->aframe_mgr = NULL;
  aread->aframe_ready_cb = NULL;
  aread->aframe_ready_cb_data = NULL;
  aframe_chain_init(&aread->aframe_chain);

  aread->error_cb = NULL;
  aread->error_cb_data = NULL;

  aread->state_time = 0;

  aread->oss = oss_init(device_name, rate, channels,bits_per_sample);
  if (!aread->oss) {
    aread_error(aread, "oss_init failed.\n");
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  if (source_name) {
    int id;
    if (oss_mixer_open(aread->oss, mixer_name) == OSS_FAIL) {
      if (mixer_name)
        aread_error(aread,"aread_init: mixer open failed, \"%s\"\n",mixer_name);
      else
        aread_error(aread, "aread_init: mixer open failed, \"/dev/mixer\"\n");
      aread_quit(aread, &dummy_time);
      return NULL;
    }
    if ((id = oss_mixer_dev_id(aread->oss, source_name)) == OSS_FAIL) {
      aread_error(aread, "aread_init: capture source open failed, \"%s\"\n", source_name);
      aread_quit(aread, &dummy_time);
      return NULL;
    }
    if (oss_mixer_record_enable(aread->oss, id) == OSS_FAIL) {
      aread_error(aread, "aread_init: \"%s\" cannot capture.\n", source_name);
      aread_quit(aread, &dummy_time);
      return NULL;
    }
    oss_mixer_close(aread->oss);
  }

  if (rate != oss_get_rate(aread->oss)) {
    aread_error(aread, "aread_init: rate error, rate %d, %d\n", rate, oss_get_rate(aread->oss));
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  if (channels != oss_get_channels(aread->oss)) {
    aread_error(aread, "aread_init: channels error, channels %d, %d\n", channels, oss_get_channels(aread->oss));
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  if (bits_per_sample != oss_get_bitspspl(aread->oss)) {
    aread_error(aread, "aread_init: bits_per_sample error, %d, %d\n", bits_per_sample, oss_get_bitspspl(aread->oss));
    aread_quit(aread, &dummy_time);
    return NULL;
  }

  fragsize = oss_get_fragsize(aread->oss);
  fragstotal = oss_get_fragstotal(aread->oss);

  aread->sync_system_time_flag = sync_system_time;
  aread->record_flag = 0;

  aread->rate = rate;
  aread->channels = channels;
  aread->bits_per_sample = bits_per_sample;
  bytespspl = channels * ((bits_per_sample+7)/8);
  aread->bytes_per_sample = bytespspl;
//  aread->fragsize = fragsize;
//  aread->fragstotal = fragstotal;

  buffer_size = (int)((double)(rate * bytespspl) * buffer_time);
  buffer_size = buffer_size + (fragsize * fragstotal) - 1;
  buffer_size = buffer_size / (fragsize * fragstotal);
  buffer_size = buffer_size * (fragsize * fragstotal);
  frame_num = buffer_size / fragsize;
  if (frame_num > READ_FRAME_MAX)
    frame_num = READ_FRAME_MAX;
  else if (frame_num < READ_FRAME_MIN)
    frame_num = READ_FRAME_MIN;

  aread->read_frame_num = frame_num;
  aread->read_frame_byte_size = fragsize;
  aread->read_frame_sample_size = fragsize / bytespspl;
  aread->read_frame_time = (u_time_t*)malloc(sizeof(u_time_t)*frame_num);
  if (!aread->read_frame_time) {
    aread_error(aread, "memory allocate error.\n");
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  aread->read_frame =
    (uint8_t**)malloc(sizeof(uint8_t*) * frame_num);
  if (!aread->read_frame) {
    aread_error(aread, "memory allocate error.\n");
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  aread->read_frame[0] = (uint8_t*)malloc(buffer_size);
  if (!aread->read_frame[0]) {
    aread_error(aread, "memory allocate error.\n");
    aread_quit(aread, &dummy_time);
    return NULL;
  }
  for (i = 1; i < frame_num; i++) {
    aread->read_frame[i] = aread->read_frame[i-1] + aread->read_frame_byte_size;
    aread->read_frame_time[i] = 0;
  }
  aread->read_frame_count = 0;
  aread->read_frame_cur = 0;
  aread->read_frame_cur_offset = 0;

  aread->record_samples = 0;
  aread->record_time = 0;
  aread->record_start_time = 0;

  aread->read_start_time = 0;
  aread->read_start_count = 0;
  aread->check_count_interval =
    (check_time * (double)rate) / aread->read_frame_sample_size;
  aread->next_check_count = 0;
  aread->gap = 0;
  aread->read_gap = 0;
  aread->interp_frame_interval = 0;
  aread->next_interp_frame_count = 0;

  pthread_mutex_init (&aread->mutex, NULL);
  pthread_cond_init (&aread->interface_wait_cond, NULL);
  pthread_cond_init (&aread->thread_wait_cond, NULL);

  aread->state = AREAD_STATE_NOT_READY;
  retcode = pthread_create (&aread->aread_th, NULL, aread_thread, aread);
  if (retcode != 0) {
    aread_error(aread, "thread_crate failed.\n");
    aread_quit(aread, &dummy_time);
    return NULL;
  }

  pthread_mutex_lock(&aread->mutex);
  if (aread->state & AREAD_STATE_NOT_READY) {
    aread->state |= AREAD_STATE_INTERFACE_WAIT;
    pthread_cond_wait(&aread->interface_wait_cond, &aread->mutex);
  }
  pthread_mutex_unlock(&aread->mutex);

  return aread;
}

