/*
 * csp_conv.c
 */

#include <stdio.h>
#include <inttypes.h>

#include "csp.h"
#include "csp_conv.h"
#include "csp_conv_yuv_to_rgb.h"
#include "csp_conv_rgb_to_rgb.h"
#include "csp_conv_yuv_to_yuv.h"
#include "csp_conv_yv12_to_rgb.h"

typedef void (*csp_conv_init_func_t)(void);

typedef struct {
  csp_conv_func_t func;
  csp_conv_init_func_t init_func;
  int dst_csp;
} conv_func_table_t;

static conv_func_table_t yuv420_func[] = {
  { csp_yuv420_to_argb32, csp_yuv_to_argb32_init, CSP_ARGB32 },
  { csp_yuv420_to_rgb24, csp_yuv_to_rgb24_init, CSP_RGB24 },
  { csp_yuv420_to_rgb565, csp_yuv_to_rgb565_init, CSP_RGB565 },
  { csp_yuv420_to_rgb555, csp_yuv_to_rgb555_init, CSP_RGB555 },
  { csp_yuv420_to_bgra32, csp_yuv_to_bgra32_init, CSP_BGRA32 },
  { csp_yuv420_to_bgr24, csp_yuv_to_bgr24_init, CSP_BGR24 },
  { csp_yuv420_to_bgr565, csp_yuv_to_bgr565_init, CSP_BGR565 },
  { csp_yuv420_to_bgr555, csp_yuv_to_bgr555_init, CSP_BGR555 },
  { csp_yuv420_to_yuv420, NULL, CSP_YUV420 },
  { csp_yuv420_to_i420, NULL, CSP_I420 },
  { csp_yuv420_to_yv12, NULL, CSP_YV12},
};

static conv_func_table_t yuv422_func[] = {
  { csp_yuv422_to_argb32, csp_yuv_to_argb32_init, CSP_ARGB32 },
  { csp_yuv422_to_rgb24, csp_yuv_to_rgb24_init, CSP_RGB24 },
  { csp_yuv422_to_rgb565, csp_yuv_to_rgb565_init, CSP_RGB565 },
  { csp_yuv422_to_rgb555, csp_yuv_to_rgb555_init, CSP_RGB555 },
  { csp_yuv422_to_bgra32, csp_yuv_to_bgra32_init, CSP_BGRA32 },
  { csp_yuv422_to_bgr24, csp_yuv_to_bgr24_init, CSP_BGR24 },
  { csp_yuv422_to_bgr565, csp_yuv_to_bgr565_init, CSP_BGR565 },
  { csp_yuv422_to_bgr555, csp_yuv_to_bgr555_init, CSP_BGR555 },
  { csp_yuv422_to_yuv422, NULL, CSP_YUV422 },
  { csp_yuv422_to_yuy2, NULL, CSP_YUY2 },
  { csp_yuv422_to_uyvy, NULL, CSP_UYVY },
};

static conv_func_table_t yuv444_func[] = {
  { csp_yuv444_to_argb32, csp_yuv_to_argb32_init, CSP_ARGB32 },
  { csp_yuv444_to_rgb24, csp_yuv_to_rgb24_init, CSP_RGB24 },
  { csp_yuv444_to_rgb565, csp_yuv_to_rgb565_init, CSP_RGB565 },
  { csp_yuv444_to_rgb555, csp_yuv_to_rgb555_init, CSP_RGB555 },
  { csp_yuv444_to_bgra32, csp_yuv_to_bgra32_init, CSP_BGRA32 },
  { csp_yuv444_to_bgr24, csp_yuv_to_bgr24_init, CSP_BGR24 },
  { csp_yuv444_to_bgr565, csp_yuv_to_bgr565_init, CSP_BGR565 },
  { csp_yuv444_to_bgr555, csp_yuv_to_bgr555_init, CSP_BGR555 },
  { csp_yuv444_to_yuv444, NULL, CSP_YUV444 },
};

static conv_func_table_t yv12_func[] = {
  { csp_yv12_to_argb32, csp_yuv_to_argb32_init, CSP_ARGB32 },
  { csp_yv12_to_rgb24, csp_yuv_to_rgb24_init, CSP_RGB24 },
  { csp_yv12_to_rgb565, csp_yuv_to_rgb565_init, CSP_RGB565 },
  { csp_yv12_to_rgb555, csp_yuv_to_rgb555_init, CSP_RGB555 },
  { csp_yv12_to_bgra32, csp_yuv_to_bgra32_init, CSP_BGRA32 },
  { csp_yv12_to_bgr24, csp_yuv_to_bgr24_init, CSP_BGR24 },
  { csp_yv12_to_bgr565, csp_yuv_to_bgr565_init, CSP_BGR565 },
  { csp_yv12_to_bgr555, csp_yuv_to_bgr555_init, CSP_BGR555 },
  { csp_yv12_to_yuv420, NULL, CSP_YUV420 },
  { csp_yv12_to_i420, NULL, CSP_I420 },
  { csp_yv12_to_yv12, NULL, CSP_YV12},
};

static conv_func_table_t yuy2_func[] = {
  { csp_yuy2_to_yuv422, NULL, CSP_YUV422 },
  { csp_yuy2_to_yuy2, NULL, CSP_YUY2 },
  { csp_yuy2_to_uyvy, NULL, CSP_UYVY },
};

static conv_func_table_t uyvy_func[] = {
  { csp_uyvy_to_yuv422, NULL, CSP_YUV422 },
  { csp_uyvy_to_yuy2, NULL, CSP_YUY2 },
  { csp_uyvy_to_uyvy, NULL, CSP_UYVY },
};



#ifdef ARCH_X86

static conv_func_table_t yuv420_func_mmxext[] = {
  { csp_yuv420_to_argb32_mmxext, NULL, CSP_ARGB32 },
//  { csp_yuv420_to_rgb24_mmxext, NULL, CSP_RGB24 },
  { csp_yuv420_to_rgb565_mmxext, NULL, CSP_RGB565 },
  { csp_yuv420_to_rgb555_mmxext, NULL, CSP_RGB555 },
  { csp_yuv420_to_bgra32_mmxext, NULL, CSP_BGRA32 },
//  { csp_yuv420_to_bgr24_mmxext, NULL, CSP_BGR24 },
  { csp_yuv420_to_bgr565_mmxext, NULL, CSP_BGR565 },
  { csp_yuv420_to_bgr555_mmxext, NULL, CSP_BGR555 },
  { csp_yuv420_to_yuv420_mmxext, NULL, CSP_YUV420 },
  { csp_yuv420_to_i420_mmxext, NULL, CSP_I420 },
  { csp_yuv420_to_yv12_mmxext, NULL, CSP_YV12},
};

static conv_func_table_t yuv422_func_mmxext[] = {
  { csp_yuv422_to_argb32_mmxext, NULL, CSP_ARGB32 },
//  { csp_yuv422_to_rgb24_mmxext, NULL, CSP_RGB24 },
  { csp_yuv422_to_rgb565_mmxext, NULL, CSP_RGB565 },
  { csp_yuv422_to_rgb555_mmxext, NULL, CSP_RGB555 },
  { csp_yuv422_to_bgra32_mmxext, NULL, CSP_BGRA32 },
//  { csp_yuv422_to_bgr24_mmxext, NULL, CSP_BGR24 },
  { csp_yuv422_to_bgr565_mmxext, NULL, CSP_BGR565 },
  { csp_yuv422_to_bgr555_mmxext, NULL, CSP_BGR555 },
  { csp_yuv422_to_yuv422_mmxext, NULL, CSP_YUV422 },
  { csp_yuv422_to_yuy2_mmxext, NULL, CSP_YUY2 },
  { csp_yuv422_to_uyvy_mmxext, NULL, CSP_UYVY },
};

static conv_func_table_t yuv444_func_mmxext[] = {
  { csp_yuv444_to_argb32_mmxext, NULL, CSP_ARGB32 },
//  { csp_yuv444_to_rgb24_mmxext, NULL, CSP_RGB24 },
  { csp_yuv444_to_rgb565_mmxext, NULL, CSP_RGB565 },
  { csp_yuv444_to_rgb555_mmxext, NULL, CSP_RGB555 },
  { csp_yuv444_to_bgra32_mmxext, NULL, CSP_BGRA32 },
//  { csp_yuv444_to_bgr24_mmxext, NULL, CSP_BGR24 },
  { csp_yuv444_to_bgr565_mmxext, NULL, CSP_BGR565 },
  { csp_yuv444_to_bgr555_mmxext, NULL, CSP_BGR555 },
  { csp_yuv444_to_yuv444_mmxext, NULL, CSP_YUV444 },
};

static conv_func_table_t yv12_func_mmxext[] = {
  { csp_yv12_to_argb32_mmxext, NULL, CSP_ARGB32 },
//  { csp_yv12_to_rgb24_mmxext, NULL, CSP_RGB24 },
  { csp_yv12_to_rgb565_mmxext, NULL, CSP_RGB565 },
  { csp_yv12_to_rgb555_mmxext, NULL, CSP_RGB555 },
  { csp_yv12_to_bgra32_mmxext, NULL, CSP_BGRA32 },
//  { csp_yv12_to_bgr24_mmxext, NULL, CSP_BGR24 },
  { csp_yv12_to_bgr565_mmxext, NULL, CSP_BGR565 },
  { csp_yv12_to_bgr555_mmxext, NULL, CSP_BGR555 },
  { csp_yv12_to_yuv420_mmxext, NULL, CSP_YUV420 },
  { csp_yv12_to_i420_mmxext, NULL, CSP_I420 },
  { csp_yv12_to_yv12_mmxext, NULL, CSP_YV12},
};

static conv_func_table_t yuy2_func_mmxext[] = {
  { csp_yuy2_to_yuv422_mmxext, NULL, CSP_YUV422 },
  { csp_yuy2_to_yuy2_mmxext, NULL, CSP_YUY2 },
  { csp_yuy2_to_uyvy_mmxext, NULL, CSP_UYVY },
};

static conv_func_table_t uyvy_func_mmxext[] = {
  { csp_uyvy_to_yuv422_mmxext, NULL, CSP_YUV422 },
  { csp_uyvy_to_yuy2_mmxext, NULL, CSP_YUY2 },
  { csp_uyvy_to_uyvy_mmxext, NULL, CSP_UYVY },
};

static conv_func_table_t yuv420_func_mmx[] = {
  { csp_yuv420_to_argb32_mmx, NULL, CSP_ARGB32 },
//  { csp_yuv420_to_rgb24_mmx, NULL, CSP_RGB24 },
  { csp_yuv420_to_rgb565_mmx, NULL, CSP_RGB565 },
  { csp_yuv420_to_rgb555_mmx, NULL, CSP_RGB555 },
  { csp_yuv420_to_bgra32_mmx, NULL, CSP_BGRA32 },
//  { csp_yuv420_to_bgr24_mmx, NULL, CSP_BGR24 },
  { csp_yuv420_to_bgr565_mmx, NULL, CSP_BGR565 },
  { csp_yuv420_to_bgr555_mmx, NULL, CSP_BGR555 },
};

static conv_func_table_t yuv422_func_mmx[] = {
  { csp_yuv422_to_argb32_mmx, NULL, CSP_ARGB32 },
//  { csp_yuv422_to_rgb24_mmx, NULL, CSP_RGB24 },
  { csp_yuv422_to_rgb565_mmx, NULL, CSP_RGB565 },
  { csp_yuv422_to_rgb555_mmx, NULL, CSP_RGB555 },
  { csp_yuv422_to_bgra32_mmx, NULL, CSP_BGRA32 },
//  { csp_yuv422_to_bgr24_mmx, NULL, CSP_BGR24 },
  { csp_yuv422_to_bgr565_mmx, NULL, CSP_BGR565 },
  { csp_yuv422_to_bgr555_mmx, NULL, CSP_BGR555 },
};

static conv_func_table_t yuv444_func_mmx[] = {
  { csp_yuv444_to_argb32_mmx, NULL, CSP_ARGB32 },
//  { csp_yuv444_to_rgb24_mmx, NULL, CSP_RGB24 },
  { csp_yuv444_to_rgb565_mmx, NULL, CSP_RGB565 },
  { csp_yuv444_to_rgb555_mmx, NULL, CSP_RGB555 },
  { csp_yuv444_to_bgra32_mmx, NULL, CSP_BGRA32 },
//  { csp_yuv444_to_bgr24_mmx, NULL, CSP_BGR24 },
  { csp_yuv444_to_bgr565_mmx, NULL, CSP_BGR565 },
  { csp_yuv444_to_bgr555_mmx, NULL, CSP_BGR555 },
};

static conv_func_table_t yv12_func_mmx[] = {
  { csp_yv12_to_argb32_mmx, NULL, CSP_ARGB32 },
//  { csp_yv12_to_rgb24_mmx, NULL, CSP_RGB24 },
  { csp_yv12_to_rgb565_mmx, NULL, CSP_RGB565 },
  { csp_yv12_to_rgb555_mmx, NULL, CSP_RGB555 },
  { csp_yv12_to_bgra32_mmx, NULL, CSP_BGRA32 },
//  { csp_yv12_to_bgr24_mmx, NULL, CSP_BGR24 },
  { csp_yv12_to_bgr565_mmx, NULL, CSP_BGR565 },
  { csp_yv12_to_bgr555_mmx, NULL, CSP_BGR555 },
};

static conv_func_table_t argb32_func_mmxext[] = {
  { csp_argb32_to_argb32_mmxext, NULL, CSP_ARGB32 },
};

static conv_func_table_t bgra32_func_mmxext[] = {
  { csp_bgra32_to_bgra32_mmxext, NULL, CSP_BGRA32 },
};

#endif

static conv_func_table_t rgb24_func[] = {
  { csp_rgb24_to_rgb24, NULL, CSP_RGB24 },
  { csp_rgb24_to_bgr24, NULL, CSP_BGR24 },
  { csp_rgb24_to_argb32, NULL, CSP_ARGB32 },
  { csp_rgb24_to_bgra32, NULL, CSP_BGRA32 },
  { csp_rgb24_to_rgb565, NULL, CSP_RGB565 },
  { csp_rgb24_to_rgb555, NULL, CSP_RGB555 },
};

static conv_func_table_t bgr24_func[] = {
  { csp_bgr24_to_bgr24, NULL, CSP_BGR24 },
  { csp_bgr24_to_rgb24, NULL, CSP_RGB24 },
  { csp_bgr24_to_argb32, NULL, CSP_ARGB32 },
  { csp_bgr24_to_bgra32, NULL, CSP_BGRA32 },
};

static conv_func_table_t argb32_func[] = {
  { csp_argb32_to_argb32, NULL, CSP_ARGB32 },
  { csp_argb32_to_bgra32, NULL, CSP_BGRA32 },
};

static conv_func_table_t bgra32_func[] = {
  { csp_bgra32_to_argb32, NULL, CSP_ARGB32 },
  { csp_bgra32_to_bgra32, NULL, CSP_BGRA32 },
};



typedef struct {
  int csp;
  conv_func_table_t *table;
  int table_num;
} src_table_t;

static src_table_t src_table[] = {
  { CSP_YUV420, yuv420_func, sizeof(yuv420_func)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv422_func, sizeof(yuv422_func)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv444_func, sizeof(yuv444_func)/sizeof(conv_func_table_t) },
  { CSP_RGB24, rgb24_func, sizeof(rgb24_func)/sizeof(conv_func_table_t) },
  { CSP_BGR24, bgr24_func, sizeof(bgr24_func)/sizeof(conv_func_table_t) },
  { CSP_ARGB32, argb32_func, sizeof(argb32_func)/sizeof(conv_func_table_t) },
  { CSP_BGRA32, bgra32_func, sizeof(bgra32_func)/sizeof(conv_func_table_t) },
  { CSP_I420, yuv420_func, sizeof(yuv420_func)/sizeof(conv_func_table_t) },
  { CSP_YV12, yv12_func, sizeof(yv12_func)/sizeof(conv_func_table_t) },
  { CSP_YUY2, yuy2_func, sizeof(yuy2_func)/sizeof(conv_func_table_t) },
  { CSP_UYVY, uyvy_func, sizeof(uyvy_func)/sizeof(conv_func_table_t) },
  { CSP_UNKNOWN, NULL, 0 },
};
static int src_table_num = sizeof(src_table)/sizeof(src_table_t);

#ifdef ARCH_X86

static src_table_t src_table_mmxext[] = {
  { CSP_YUV420, yuv420_func_mmxext, sizeof(yuv420_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv422_func_mmxext, sizeof(yuv422_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv444_func_mmxext, sizeof(yuv444_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_ARGB32, argb32_func_mmxext, sizeof(argb32_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_BGRA32, bgra32_func_mmxext, sizeof(bgra32_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_I420, yuv420_func_mmxext, sizeof(yuv420_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_YV12, yv12_func_mmxext, sizeof(yv12_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_YUY2, yuy2_func_mmxext, sizeof(yuy2_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_UYVY, uyvy_func_mmxext, sizeof(uyvy_func_mmxext)/sizeof(conv_func_table_t) },
  { CSP_UNKNOWN, NULL, 0 },
};
static int src_table_num_mmxext = sizeof(src_table_mmxext)/sizeof(src_table_t);

static src_table_t src_table_mmx[] = {
  { CSP_YUV420, yuv420_func_mmx, sizeof(yuv420_func_mmx)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv422_func_mmx, sizeof(yuv422_func_mmx)/sizeof(conv_func_table_t) },
  { CSP_YUV422, yuv444_func_mmx, sizeof(yuv444_func_mmx)/sizeof(conv_func_table_t) },
  { CSP_I420, yuv420_func_mmx, sizeof(yuv420_func_mmx)/sizeof(conv_func_table_t) },
  { CSP_YV12, yv12_func_mmx, sizeof(yv12_func_mmx)/sizeof(conv_func_table_t) },
  { CSP_UNKNOWN, NULL, 0 },
};
static int src_table_num_mmx = sizeof(src_table_mmx)/sizeof(src_table_t);

#endif /* ARCH_X86 */

#define CPU_ACCEL_UNKNOWN     0
#define CPU_ACCEL_NONE        (1UL<<0)
#define CPU_ACCEL_MMX         (1UL<<1)
#define CPU_ACCEL_MMXEXT      (1UL<<2)

#include <mm_accel.h>

static unsigned int cpu_accel = CPU_ACCEL_UNKNOWN;

int
csp_conv_init(int accel)
{
  uint32_t caps;

  if (!accel) {
    cpu_accel = CPU_ACCEL_NONE;
    return 0;
  }

  if (cpu_accel != CPU_ACCEL_UNKNOWN)
    return 0;

  caps = mm_accel();
  if (caps & MM_ACCEL_X86_MMXEXT)
    cpu_accel |= CPU_ACCEL_MMXEXT;
  if (caps & MM_ACCEL_X86_MMX)
    cpu_accel |= CPU_ACCEL_MMX;

  if (cpu_accel == CPU_ACCEL_UNKNOWN)
    cpu_accel = CPU_ACCEL_NONE;

  return 0;
}

static src_table_t*
find_table_entry(src_table_t *table, int table_num, int src_csp)
{
  int i;
  for (i = 0; i < table_num; i++, table++) {
    if (table->csp == src_csp)
      return table;
  }
  return NULL;
}

static conv_func_table_t*
find_func_entry(conv_func_table_t *table, int func_num, int dst_csp)
{
  int i;
  for (i = 0; i < func_num; i++, table++) {
    if (table->dst_csp == dst_csp)
      return table;
  }
  return NULL;
}

csp_conv_func_t
csp_conv_get_convert_func(int src_csp, int dst_csp)
{
  conv_func_table_t *table;
  src_table_t *s_table;

  if (cpu_accel == CPU_ACCEL_UNKNOWN)
    csp_conv_init(1);

  table = NULL;
#ifdef ARCH_X86
  if (cpu_accel != CPU_ACCEL_NONE) {
    s_table = find_table_entry(src_table_mmxext, src_table_num_mmxext, src_csp);
    if (s_table)
      table = find_func_entry(s_table->table, s_table->table_num, dst_csp);
  }
  if (!table && (cpu_accel != CPU_ACCEL_NONE)) {
    s_table = find_table_entry(src_table_mmx, src_table_num_mmx, src_csp);
    if (s_table)
      table = find_func_entry(s_table->table, s_table->table_num, dst_csp);
  }
#endif /* ARCH_X86 */

  if (!table) {
    s_table = find_table_entry(src_table, src_table_num, src_csp);
    if (s_table)
      table = find_func_entry(s_table->table, s_table->table_num, dst_csp);
  }

  if (!table)
    return NULL;

  if (table->init_func)
    table->init_func();

  return table->func;
}

void
csp_conv_quit(void)
{
  csp_yuv_to_argb32_quit();
  csp_yuv_to_rgb24_quit();
  csp_yuv_to_rgb565_quit();
  csp_yuv_to_rgb555_quit();
  csp_yuv_to_bgra32_quit();
  csp_yuv_to_bgr24_quit();
  csp_yuv_to_bgr565_quit();
  csp_yuv_to_bgr555_quit();
}

