/*
 * rgb_to_yuv.c
 *
 * code from mjpegtools
 * Copyright: see below
 *
 */

/*
 * colorspace.c:  Routines to perform colorspace conversions.
 *
 *
 *  Copyright (C) 2001 Matthew J. Marjanovic <maddog@mir.com>
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "rgb_to_yuv.h"


static uint32_t matrix_coefficients = 6;

static double yuv_coef[7][3] = {
  {0.2125,0.7154,0.0721}, /* ITU-R Rec. 709 (1990) */
  {0.299, 0.587, 0.114},  /* unspecified */
  {0.299, 0.587, 0.114},  /* reserved */
  {0.30,  0.59,  0.11},   /* FCC */
  {0.299, 0.587, 0.114},  /* ITU-R Rec. 624-4 System B, G */
  {0.299, 0.587, 0.114},  /* SMPTE 170M */
  {0.212, 0.701, 0.087}}; /* SMPTE 240M (1987) */

/* precomputed tables */

int *rgb_to_yuv_Y_R = NULL;
int *rgb_to_yuv_Y_G = NULL;
int *rgb_to_yuv_Y_B = NULL;
int *rgb_to_yuv_U_R = NULL;
int *rgb_to_yuv_U_G = NULL;
int *rgb_to_yuv_U_B = NULL;
int *rgb_to_yuv_V_R = NULL;
int *rgb_to_yuv_V_G = NULL;
int *rgb_to_yuv_V_B = NULL;

static int myround(double n)
{
  if (n >= 0) 
    return (int)(n + 0.5);
  else
    return (int)(n - 0.5);
}

static void rgb_to_yuv_table_init(void)
{
  int i;
  double cyr, cyg, cyb;
  double cur, cug, cub;
  double cvr, cvg, cvb;

  if (rgb_to_yuv_Y_R)
    return;

  rgb_to_yuv_Y_R = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_Y_G = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_Y_B = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_U_R = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_U_G = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_U_B = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_V_R = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_V_G = (int*)malloc(sizeof(int)*256);
  rgb_to_yuv_V_B = (int*)malloc(sizeof(int)*256);

  cyr = coef[matrix_coefficients][0];
  cyg = coef[matrix_coefficients][1];
  cyb = coef[matrix_coefficients][2];

  cur = -cyr/(2.-2.*cyb);
  cug = -cyg/(2.-2.*cyb);
  cub = (1.-cyb)/(2.-2.*cyb);

  cvr = (1.-cyr)/(2.-2.*cyr);
  cvg = -cyg/(2.-2.*cyr);
  cvb = -cyb/(2.-2.*cyr);

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-point-factor)
   *
   * to one of each, add the following:
   *             + (fixed-point-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-point-factor)  --- to add the offset
   *             
   */
  for (i = 0; i < 256; i++) {
    rgb_to_yuv_Y_R[i] = myround(cyr * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_Y_G[i] = myround(cyg * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_Y_B[i] = myround((cyb * (double)i 
		      * 219.0 / 255.0 * (double)(1<<FP_BITS))
		     + (double)(1<<(FP_BITS-1))
		     + (16.0 * (double)(1<<FP_BITS)));

    rgb_to_yuv_U_R[i] = myround(cur * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_U_G[i] = myround(cug * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_U_B[i] = myround((cub * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));

    rgb_to_yuv_V_R[i] = myround(cvr * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_V_G[i] = myround(cvg * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    rgb_to_yuv_V_B[i] = myround((cvb * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));
  }
}

