/*
 * yuv_to_argb32.c
 *
 * code from mpeg2dec <http://libmpeg2.sourceforge.net>
 *
 * Copyright: see below
 */

/*
 * yuv2rgb.c
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "csp_conv.h"
#include "csp_conv_yuv_to_rgb.h"
#include "yuv_to_rgb.h"

static uint32_t ** table_rV = NULL;
static uint32_t ** table_gU = NULL;
static int      *  table_gV = NULL;
static uint32_t ** table_bU = NULL;
static uint32_t *  table_32 = NULL;

void csp_yuv_to_argb32_init (void)
{
  YUV_to_RGB_Tables tables;

  if (table_32)
    return;

  yuv_to_rgb_init_table(YUV_TO_RGB_MODE_ARGB, 32, &tables);

  table_rV = (uint32_t**)tables.table_rV;
  table_gU = (uint32_t**)tables.table_gU;
  table_gV = (int      *)tables.table_gV;
  table_bU = (uint32_t**)tables.table_bU;
  table_32 = (uint32_t *)tables.table;
}

void csp_yuv_to_argb32_quit(void)
{
  if (table_rV) free(table_rV);
  if (table_gU) free(table_gU);
  if (table_gV) free(table_gV);
  if (table_bU) free(table_bU);
  if (table_32) free(table_32);

  table_rV = table_gU = table_bU = NULL;
  table_gV = NULL;
  table_32 = NULL;
}


#define RGB(i)                                              \
        U = pu[i];                                          \
        V = pv[i];                                          \
        r = table_rV[V];                                    \
        g = table_gU[U] + table_gV[V];                      \
        b = table_bU[U];

#define DST2(py,dst,i)                            \
        Y = py[2*i];                              \
        dst[2*i] = r[Y] + g[Y] + b[Y];            \
        Y = py[2*i+1];                            \
        dst[2*i+1] = r[Y] + g[Y] + b[Y];

#define DST1(py,dst,i)                          \
        Y = py[i];                              \
        dst[i] = r[Y] + g[Y] + b[Y];


void csp_yuv420_to_argb32 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint32_t * r, * g, * b;
  uint32_t * dst1, * dst2;
  uint8_t *py1, *py2, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_32)
//    csp_yuv_to_argb32_init ();

  dst1 = (uint32_t *) rgb[0];
  dst2 = dst1 + rgb_stride[0];
  py1 = yuv_planes[0];
  py2 = py1 + yuv_stride[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0]*2 - width;
  next_rgb = rgb_stride[0]*2 - width;
  next_uv = yuv_stride[1];

  width >>= 1;
  height >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST2(py1,dst1,0);
      DST2(py2,dst2,0);

      py1 +=2; py2 +=2;
      dst1+=2; dst2+=2;
    }
    pu += next_uv;
    pv += next_uv;
    py1 += next_y;
    py2 += next_y;
    dst1 += next_rgb;
    dst2 += next_rgb;
  }
}

void csp_yuv422_to_argb32 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint32_t * r, * g, * b;
  uint32_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_32)
//    csp_yuv_to_argb32_init ();

  dst = (uint32_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0] - width;
  next_rgb = rgb_stride[0] - width;
  next_uv = yuv_stride[1];

  width >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST2(py,dst,0);

      py++;
      dst++;
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}

void csp_yuv444_to_argb32 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint32_t * r, * g, * b;
  uint32_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_32)
//    csp_yuv_to_argb32_init ();

  dst = (uint32_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0];
  next_rgb = rgb_stride[0];
  next_uv = yuv_stride[1];

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST1(py,dst,w);
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}



#if 0

/* precomputed tables */

static int RGB_Y[256];
static int R_V[256];
static int G_U[256];
static int G_V[256];
static int B_U[256];
static int conv_YR_inited = 0;

static const int32_t tInverse_Table_6_9[8][4] = {
  {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
  {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
  {104597, 132201, 25675, 53279}, /* unspecified */
  {104597, 132201, 25675, 53279}, /* reserved */
  {104448, 132798, 24759, 53109}, /* FCC */
  {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
  {104597, 132201, 25675, 53279}, /* SMPTE 170M */
  {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

static unsigned char Clip_table[1024];
static unsigned char *Clip;

static void init_YCbCr_to_RGB_tables(void)
{
  int i;

  int crv = tInverse_Table_6_9[6][0];
  int cbu = tInverse_Table_6_9[6][1];
  int cgu = -tInverse_Table_6_9[6][2];
  int cgv = -tInverse_Table_6_9[6][3];

  for (i = 0; i < 16; i++)
    RGB_Y[i] = 76309 * 16;
  for (i = 16; i < 236; i++)
    RGB_Y[i] = 76309 * (i-16);
  for (i = 236; i < 256; i++)
    RGB_Y[i] = 76309 * 235;
    
  /* clip Cb/Cr values below 16 */	 
  for (i = 0; i < 16; i++) {
    R_V[i] = crv * -112;
    G_U[i] = cgu * -112;
    G_V[i] = cgv * -112;
    B_U[i] = cbu * -112;
  }
  for (i = 16; i < 241; i++) {
    R_V[i] = crv * (i-128);
    G_U[i] = cgu * (i-128);
    G_V[i] = cgv * (i-128);
    B_U[i] = cbu * (i-128);
  }
  /* clip Cb/Cr values above 240 */	 
  for (i = 241; i < 256; i++) {
    R_V[i] = crv * (112);
    G_U[i] = cgu * (112);
    G_V[i] = cgv * (i-128);
    B_U[i] = cbu * (112);
  }
  conv_YR_inited = 1;

  Clip = Clip_table + 384;
  for (i = -384; i < 640; i++)
    Clip[i] = (i<0) ? 0 : (i>255) ? 255 : i;
}


void csp_yuv420_to_rgb32 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint32_t * dst1, * dst2;
  uint8_t *py1, *py2, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

  if (!conv_YR_inited)
    init_YCbCr_to_RGB_tables();

  dst1 = (uint32_t *) rgb[0];
  dst2 = dst1 + rgb_stride[0];
  py1 = yuv_planes[0];
  py2 = py1 + yuv_stride[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0]*2 - width;
  next_rgb = rgb_stride[0]*2 - width;
  next_uv = yuv_stride[1];

  width >>= 1;
  height >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      U = pu[w];
      V = pv[w];
      Y = py1[0];
      dst1[0] = (Clip[((RGB_Y[Y] + R_V[V]) >> 16)] << 16) |
                (Clip[((RGB_Y[Y] + G_U[U] + G_V[V]) >> 16)] << 8) |
                (Clip[((RGB_Y[Y] + B_U[U]) >> 16)]);
      Y = py1[1];
      dst1[1] = (Clip[((RGB_Y[Y] + R_V[V]) >> 16)] << 16) |
                (Clip[((RGB_Y[Y] + G_U[U] + G_V[V]) >> 16)] << 8) |
                (Clip[((RGB_Y[Y] + B_U[U]) >> 16)]);
      Y = py2[0];
      dst2[0] = (Clip[((RGB_Y[Y] + R_V[V]) >> 16)] << 16) |
                (Clip[((RGB_Y[Y] + G_U[U] + G_V[V]) >> 16)] << 8) |
                (Clip[((RGB_Y[Y] + B_U[U]) >> 16)]);
      Y = py2[1];
      dst2[1] = (Clip[((RGB_Y[Y] + R_V[V]) >> 16)] << 16) |
                (Clip[((RGB_Y[Y] + G_U[U] + G_V[V]) >> 16)] << 8) |
                (Clip[((RGB_Y[Y] + B_U[U]) >> 16)]);

      py1 +=2; py2 +=2;
      dst1+=2; dst2+=2;
    }
    pu += next_uv;
    pv += next_uv;
    py1 += next_y;
    py2 += next_y;
    dst1 += next_rgb;
    dst2 += next_rgb;
  }
}
#endif

