/*
 * yuv_to_rgb.c
 *
 * code from mpeg2dec <http://libmpeg2.sourceforge.net>
 *
 */

/*
 * yuv2rgb.c
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "yuv_to_rgb.h"

static uint32_t matrix_coefficients = 6;

#if 0
static double yuv_coef[7][3] = {
  {0.2125,0.7154,0.0721}, /* ITU-R Rec. 709 (1990) */
  {0.299, 0.587, 0.114},  /* unspecified */
  {0.299, 0.587, 0.114},  /* reserved */
  {0.30,  0.59,  0.11},   /* FCC */
  {0.299, 0.587, 0.114},  /* ITU-R Rec. 624-4 System B, G */
  {0.299, 0.587, 0.114},  /* SMPTE 170M */
  {0.212, 0.701, 0.087}}; /* SMPTE 240M (1987) */
#endif


/* color space conversion coefficients
 * for YCbCr -> RGB mapping
 *
 * entries are {crv,cbu,cgu,cgv}
 *
 * crv=(255/224)*65536*(1-cr)/0.5
 * cbu=(255/224)*65536*(1-cb)/0.5
 * cgu=(255/224)*65536*(cb/cg)*(1-cb)/0.5
 * cgv=(255/224)*65536*(cr/cg)*(1-cr)/0.5
 *
 * where Y=cr*R+cg*G+cb*B (cr+cg+cb=1)
 */

/* ISO/IEC 13818-2 section 6.3.6 sequence_display_extension() */

static const int32_t Inverse_Table_6_9[8][4] = {
    {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
    {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
    {104597, 132201, 25675, 53279}, /* unspecified */
    {104597, 132201, 25675, 53279}, /* reserved */
    {104448, 132798, 24759, 53109}, /* FCC */
    {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
    {104597, 132201, 25675, 53279}, /* SMPTE 170M */
    {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

#ifdef ARCH_X86

#include "attributes.h"
#include "mmx.h"

mmx_t csp_conv_mmx_80w = {0x0080008000800080LL};
mmx_t csp_conv_mmx_U_green = {0xf37df37df37df37dLL};
mmx_t csp_conv_mmx_U_blue = {0x4093409340934093LL};
mmx_t csp_conv_mmx_V_red = {0x3312331233123312LL};
mmx_t csp_conv_mmx_V_green = {0xe5fce5fce5fce5fcLL};
mmx_t csp_conv_mmx_10w = {0x1010101010101010LL};
mmx_t csp_conv_mmx_00ffw = {0x00ff00ff00ff00ffLL};
mmx_t csp_conv_mmx_Y_coeff = {0x253f253f253f253fLL};

mmx_t csp_conv_mmx_bluemask = {0xf8f8f8f8f8f8f8f8LL};
mmx_t csp_conv_mmx_greenmask = {0xfcfcfcfcfcfcfcfcLL};
mmx_t csp_conv_mmx_greenmask5 = {0xf8f8f8f8f8f8f8f8LL};
mmx_t csp_conv_mmx_redmask = {0xf8f8f8f8f8f8f8f8LL};

void yuv_to_rgb_mmx_init(void)
{
  int U_green, U_blue, V_red, V_green, Y_coeff;
  int i;

  V_red = Inverse_Table_6_9[matrix_coefficients][0];
  U_blue = Inverse_Table_6_9[matrix_coefficients][1];
  U_green = -Inverse_Table_6_9[matrix_coefficients][2];
  V_green = -Inverse_Table_6_9[matrix_coefficients][3];
  Y_coeff = 76309/8;

  for (i = 0; i < 4; i++) {
    csp_conv_mmx_U_green.w[i] = U_green;
    csp_conv_mmx_U_blue.w[i] = U_blue;
    csp_conv_mmx_V_red.w[i] = V_red;
    csp_conv_mmx_V_green.w[i] = V_green;
    csp_conv_mmx_Y_coeff.w[i] = Y_coeff;
  }
}

#endif

static int div_round (int dividend, int divisor)
{
    if (dividend > 0)
	return (dividend + (divisor>>1)) / divisor;
    else
	return -((-dividend + (divisor>>1)) / divisor);
}

int yuv_to_rgb_init_table (int order, int bpp, YUV_to_RGB_Tables *tables_ret)
{
    int i;
    uint8_t table_Y[1024];
    uint32_t * table_32 = 0;
    uint16_t * table_16 = 0;
    uint8_t * table_8 = 0;
    int entry_size = 0;
    void * table_r = 0;
    void * table_g = 0;
    void * table_b = 0;
    void * table = NULL;

    void ** table_rV = NULL;
    void ** table_gU = NULL;
    int  *  table_gV = NULL;
    void ** table_bU = NULL;

    int crv = Inverse_Table_6_9[matrix_coefficients][0];
    int cbu = Inverse_Table_6_9[matrix_coefficients][1];
    int cgu = -Inverse_Table_6_9[matrix_coefficients][2];
    int cgv = -Inverse_Table_6_9[matrix_coefficients][3];

    tables_ret->table_rV = NULL;
    tables_ret->table_gU = NULL;
    tables_ret->table_gV = NULL;
    tables_ret->table_bU = NULL;
    tables_ret->table = NULL;

    for (i = 0; i < 1024; i++) {
	int j;

	/* 76309 = (255/219)*65536 */
	j = (76309 * (i - 384 - 16) + 32768) >> 16;
	j = (j < 0) ? 0 : ((j > 255) ? 255 : j);
	table_Y[i] = j;
    }

    switch (bpp) {
    case 32:
	table_32 = (uint32_t *) malloc ((197 + 2*682 + 256 + 132) *
					sizeof (uint32_t));
	table = table_32;
        table_rV = (void**) malloc (sizeof(uint32_t*) * 256);
        table_gU = (void**) malloc (sizeof(uint32_t*) * 256);
        table_gV = (int * ) malloc (sizeof(int      ) * 256);
        table_bU = (void**) malloc (sizeof(uint32_t*) * 256);

	entry_size = sizeof (uint32_t);
	table_r = table_32 + 197;
	table_b = table_32 + 197 + 685;
	table_g = table_32 + 197 + 2*682;

#ifdef WORDS_BIGENDIAN
	for (i = -197; i < 256+197; i++)
	    ((uint32_t *) table_r)[i] =
		table_Y[i+384] << ((order == YUV_TO_RGB_MODE_ARGB) ? 16 : 0);
	for (i = -132; i < 256+132; i++)
	    ((uint32_t *) table_g)[i] = table_Y[i+384] << 8;
	for (i = -232; i < 256+232; i++)
	    ((uint32_t *) table_b)[i] =
		table_Y[i+384] << ((order == YUV_TO_RGB_MODE_ARGB) ? 0 : 16);
#else
	for (i = -197; i < 256+197; i++)
	    ((uint32_t *) table_r)[i] =
		table_Y[i+384] << ((order == YUV_TO_RGB_MODE_ARGB) ? 8 : 16);
	for (i = -132; i < 256+132; i++)
//	    ((uint32_t *) table_g)[i] = table_Y[i+384] << 8;
	    ((uint32_t *) table_g)[i] =
		table_Y[i+384] << ((order == YUV_TO_RGB_MODE_ARGB) ? 16 : 8);
	for (i = -232; i < 256+232; i++)
	    ((uint32_t *) table_b)[i] =
		table_Y[i+384] << ((order == YUV_TO_RGB_MODE_ARGB) ? 24 : 0);
#endif
	break;

    case 24:
	table_8 = (uint8_t *) malloc ((256 + 2*232) * sizeof (uint8_t));
	table = table_8;
        table_rV = (void**) malloc (sizeof(uint8_t*) * 256);
        table_gU = (void**) malloc (sizeof(uint8_t*) * 256);
        table_gV = (int * ) malloc (sizeof(int     ) * 256);
        table_bU = (void**) malloc (sizeof(uint8_t*) * 256);

	entry_size = sizeof (uint8_t);
	table_r = table_g = table_b = table_8 + 232;

	for (i = -232; i < 256+232; i++)
	    ((uint8_t * )table_b)[i] = table_Y[i+384];
	break;

    case 15:
    case 16:
	table_16 = (uint16_t *) malloc ((197 + 2*682 + 256 + 132) *
					sizeof (uint16_t));
	table = table_16;
        table_rV = (void**) malloc (sizeof(uint16_t*) * 256);
        table_gU = (void**) malloc (sizeof(uint16_t*) * 256);
        table_gV = (int * ) malloc (sizeof(int      ) * 256);
        table_bU = (void**) malloc (sizeof(uint16_t*) * 256);

	entry_size = sizeof (uint16_t);
	table_r = table_16 + 197;
	table_b = table_16 + 197 + 685;
	table_g = table_16 + 197 + 2*682;

	for (i = -197; i < 256+197; i++) {
	    int j = table_Y[i+384] >> 3;

	    if (order == YUV_TO_RGB_MODE_RGB)
		j <<= ((bpp==16) ? 11 : 10);

	    ((uint16_t *)table_r)[i] = j;
	}
	for (i = -132; i < 256+132; i++) {
	    int j = table_Y[i+384] >> ((bpp==16) ? 2 : 3);

	    ((uint16_t *)table_g)[i] = j << 5;
	}
	for (i = -232; i < 256+232; i++) {
	    int j = table_Y[i+384] >> 3;

	    if (order == YUV_TO_RGB_MODE_BGR)
		j <<= ((bpp==16) ? 11 : 10);

	    ((uint16_t *)table_b)[i] = j;
	}
	break;

    default:
	fprintf (stderr, "%ibpp not supported by yuv_to_rgb\n", bpp);
	exit (1);
    }

    for (i = 0; i < 256; i++) {
	table_rV[i] = (((uint8_t *)table_r) +
		       entry_size * div_round (crv * (i-128), 76309));
	table_gU[i] = (((uint8_t *)table_g) +
		       entry_size * div_round (cgu * (i-128), 76309));
//	table_gV[i] = entry_size * div_round (cgv * (i-128), 76309);
	table_gV[i] = div_round (cgv * (i-128), 76309);
	table_bU[i] = (((uint8_t *)table_b) +
		       entry_size * div_round (cbu * (i-128), 76309));
    }

    tables_ret->table_rV = table_rV;
    tables_ret->table_gU = table_gU;
    tables_ret->table_gV = table_gV;
    tables_ret->table_bU = table_bU;
    tables_ret->table = table;

    return 0;
}

