/*
 * yuv_to_rgb555.c
 *
 * code from mpeg2dec <http://libmpeg2.sourceforge.net>
 *
 */

/*
 * yuv2rgb.c
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "csp_conv.h"
#include "csp_conv_yuv_to_rgb.h"
#include "yuv_to_rgb.h"

static uint16_t ** table_rV = NULL;
static uint16_t ** table_gU = NULL;
static int      *  table_gV = NULL;
static uint16_t ** table_bU = NULL;
static uint16_t *  table_555 = NULL;

void csp_yuv_to_rgb555_init (void)
{
  YUV_to_RGB_Tables tables;

  if (table_555)
    return;

  yuv_to_rgb_init_table(YUV_TO_RGB_MODE_RGB, 15, &tables);

  table_rV  = (uint16_t**)tables.table_rV;
  table_gU  = (uint16_t**)tables.table_gU;
  table_gV  = (int      *)tables.table_gV;
  table_bU  = (uint16_t**)tables.table_bU;
  table_555 = (uint16_t *)tables.table;
}

void csp_yuv_to_rgb555_quit(void)
{
  if (table_rV) free(table_rV);
  if (table_gU) free(table_gU);
  if (table_gV) free(table_gV);
  if (table_bU) free(table_bU);
  if (table_555)   free(table_555);

  table_rV = table_gU = table_bU = NULL;
  table_gV = NULL;
  table_555 = NULL;
}

#define RGB(i)                                                  \
        U = pu[i];                                              \
        V = pv[i];                                              \
        r = table_rV[V];                                        \
        g = table_gU[U] + table_gV[V];                          \
        b = table_bU[U];

#define DST2(py,dst,i)                          \
        Y = py[2*i];                            \
        dst[2*i] = r[Y] + g[Y] + b[Y];          \
        Y = py[2*i+1];                          \
        dst[2*i+1] = r[Y] + g[Y] + b[Y];

#define DST1(py,dst,i)                          \
        Y = py[i];                              \
        dst[i] = r[Y] + g[Y] + b[Y];


void csp_yuv420_to_rgb555 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint16_t * r, * g, * b;
  uint16_t * dst1, * dst2;
  uint8_t *py1, *py2, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_555)
//    csp_yuv_to_rgb555_init ();

  dst1 = (uint16_t *) rgb[0];
  dst2 = dst1 + rgb_stride[0];
  py1 = yuv_planes[0];
  py2 = py1 + yuv_stride[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0]*2 - width;
  next_rgb = rgb_stride[0]*2 - width;
  next_uv = yuv_stride[1];

  width >>= 1;
  height >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST2(py1,dst1,0);
      DST2(py2,dst2,0);

      py1 +=2; py2 +=2;
      dst1+=2; dst2+=2;
    }
    pu += next_uv;
    pv += next_uv;
    py1 += next_y;
    py2 += next_y;
    dst1 += next_rgb;
    dst2 += next_rgb;
  }
}

void csp_yuv422_to_rgb555 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint16_t * r, * g, * b;
  uint16_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_555)
//    csp_yuv_to_rgb555_init ();

  dst = (uint16_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0] - width;
  next_rgb = rgb_stride[0] - width;
  next_uv = yuv_stride[1];

  width >>= 1;
  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST2(py,dst,0);

      py +=2;
      dst+=2;
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}

void csp_yuv444_to_rgb555 (uint8_t *yuv_planes[], uint8_t *rgb[],
	  int yuv_stride[], int rgb_stride[], int width, int height)
{
  int U, V, Y;
  uint16_t * r, * g, * b;
  uint16_t * dst;
  uint8_t *py, *pu, *pv;
  int w;
  int next_y, next_rgb, next_uv;

//  if (!table_555)
//    csp_yuv_to_rgb555_init ();

  dst = (uint16_t *) rgb[0];
  py = yuv_planes[0];
  pu = yuv_planes[1];
  pv = yuv_planes[2];

  next_y = yuv_stride[0];
  next_rgb = rgb_stride[0];
  next_uv = yuv_stride[1];

  for (; height > 0; height--)
  {
    for (w = 0; w < width; w++)
    {
      RGB(w);
      DST1(py,dst,w);
    }
    pu  += next_uv;
    pv  += next_uv;
    py  += next_y;
    dst += next_rgb;
  }
}

