/*
 * yuv2rgb_mmx.c
 * Copyright (C) 2000-2002 Silicon Integrated System Corp.
 * All Rights Reserved.
 *
 * Author: Olie Lho <ollie@sis.com.tw>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#ifdef ARCH_X86

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "attributes.h"
#include "mmx.h"

#include "yuv_to_rgb_mmx.h"

static inline void mmxext_unpack_16rgb (uint8_t * image)
{
    /*
     * convert RGB plane to RGB 16 bits
     * mm0 -> B, mm1 -> R, mm2 -> G
     * mm4 -> GB, mm5 -> AR pixel 4-7
     * mm6 -> GB, mm7 -> AR pixel 0-3
     */

    pand_m2r (csp_conv_mmx_bluemask, mm0);	/* mm0 = b7b6b5b4b3______ */
    pand_m2r (csp_conv_mmx_greenmask, mm2);	/* mm2 = g7g6g5g4g3g2____ */
    pand_m2r (csp_conv_mmx_redmask, mm1);	/* mm1 = r7r6r5r4r3______ */
    psrlq_i2r (3, mm0);			/* mm0 = ______b7b6b5b4b3 */
    pxor_r2r (mm4, mm4);		/* mm4 = 0 */
    movq_r2r (mm0, mm5);		/* mm5 = ______b7b6b5b4b3 */
    movq_r2r (mm2, mm7);		/* mm7 = g7g6g5g4g3g2____ */

    punpcklbw_r2r (mm4, mm2);
    punpcklbw_r2r (mm1, mm0);
    psllq_i2r (3, mm2);
    por_r2r (mm2, mm0);
    movntq_r2m (mm0, *image);

    punpckhbw_r2r (mm4, mm7);
    punpckhbw_r2r (mm1, mm5);
    psllq_i2r (3, mm7);
    por_r2r (mm7, mm5);
    movntq_r2m (mm5, *(image+8));
}

static inline void mmx_unpack_16rgb (uint8_t * image)
{
    /*
     * convert RGB plane to RGB 16 bits
     * mm0 -> B, mm1 -> R, mm2 -> G
     * mm4 -> GB, mm5 -> AR pixel 4-7
     * mm6 -> GB, mm7 -> AR pixel 0-3
     */

    pand_m2r (csp_conv_mmx_bluemask, mm0);	/* mm0 = b7b6b5b4b3______ */
    pand_m2r (csp_conv_mmx_greenmask, mm2);	/* mm2 = g7g6g5g4g3g2____ */
    pand_m2r (csp_conv_mmx_redmask, mm1);	/* mm1 = r7r6r5r4r3______ */
    psrlq_i2r (3, mm0);			/* mm0 = ______b7b6b5b4b3 */
    pxor_r2r (mm4, mm4);		/* mm4 = 0 */
    movq_r2r (mm0, mm5);		/* mm5 = ______b7b6b5b4b3 */
    movq_r2r (mm2, mm7);		/* mm7 = g7g6g5g4g3g2____ */

    punpcklbw_r2r (mm4, mm2);
    punpcklbw_r2r (mm1, mm0);
    psllq_i2r (3, mm2);
    por_r2r (mm2, mm0);
    movq_r2m (mm0, *image);

    punpckhbw_r2r (mm4, mm7);
    punpckhbw_r2r (mm1, mm5);
    psllq_i2r (3, mm7);
    por_r2r (mm7, mm5);
    movq_r2m (mm5, *(image+8));
}

void csp_yuv420_to_rgb565_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmxext_unpack_16rgb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
	if (height & 1) {
	    pu += u_stride;
	    pv += v_stride;
	} else {
	    pu -= 4 * width;
	    pv -= 4 * width;
	}
    } while (--height);
    emms();
}


void csp_yuv422_to_rgb565_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmxext_unpack_16rgb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

void csp_yuv444_to_rgb565_mmxext (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - width;
    v_stride = yuv_stride[2] - width;
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv44_to_rgb (py, pu, pv);
	    mmxext_unpack_16rgb (image);
	    py += 8;
	    pu += 8;
	    pv += 8;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}


void csp_yuv420_to_rgb565_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmx_unpack_16rgb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
	if (height & 1) {
	    pu += u_stride;
	    pv += v_stride;
	} else {
	    pu -= 4 * width;
	    pv -= 4 * width;
	}
    } while (--height);
    emms();
}


void csp_yuv422_to_rgb565_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - (width >> 1);
    v_stride = yuv_stride[2] - (width >> 1);
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv42_to_rgb (py, pu, pv);
	    mmx_unpack_16rgb (image);
	    py += 8;
	    pu += 4;
	    pv += 4;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

void csp_yuv444_to_rgb565_mmx (uint8_t *yuv_planes[], uint8_t *rgb[],
                int yuv_stride[], int rgb_stride[], int width, int height)
{
    int i;
    int y_stride, u_stride, v_stride;
    int rgb_std;
    uint8_t *image;
    uint8_t *py, *pu, *pv;

    image = rgb[0];
    py = yuv_planes[0];
    pu = yuv_planes[1];
    pv = yuv_planes[2];

    y_stride = yuv_stride[0] - width;
    u_stride = yuv_stride[1] - width;
    v_stride = yuv_stride[2] - width;
    rgb_std  = (rgb_stride[0] - width) * 2;
    width >>= 3;

    do {
	i = width;
	do {
	    mmx_yuv44_to_rgb (py, pu, pv);
	    mmx_unpack_16rgb (image);
	    py += 8;
	    pu += 8;
	    pv += 8;
	    image += 16;
	} while (--i);

	py += y_stride;
	image += rgb_std;
        pu += u_stride;
        pv += v_stride;
    } while (--height);
    emms();
}

#endif

