/* mgr.c */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "parseopt.h"
#include "preset.h"
#include "csp.h"
#include "get_cur_u_time.h"
#include "vcodec.h"
#include "vframe_mgr.h"
#include "v4l.h"
#include "vgrab.h"
#include "venc.h"
#include "acodec.h"
#include "aframe_mgr.h"
#include "aread.h"
#include "aenc.h"
#include "ffmt.h"
#include "oss.h"

#include "mgr.h"

#define AUDIO_BUFFER_TIME 3.0

#ifndef NDEBUG
#define debug(str) fprintf(stderr, str);fflush(stderr)
#else
#define debug(str)
#endif /* NDEBUG */

static MGR_PARAM mgr_param;

static SEL_COMPO video_codecs[] = {
#ifdef HAVE_LIBXVIDCORE
  { "XVID",  (int)VCODEC_XVID },
#endif /* HAVE_LIBXVIDCORE */

  { "RGB24", (int)VCODEC_RGB24 },
  { "YV12",  (int)VCODEC_YV12 },
  { "RTJPEG",  (int)VCODEC_RTJPEG },

#ifdef ARCH_X86
  { "RTJ1",  (int)VCODEC_RTJ1 },
#endif

#ifdef HAVE_LIBAVCODEC
  { "MPEG1",  (int)VCODEC_MPEG1 },
  { "H263",  (int)VCODEC_H263 },
  { "H263P",  (int)VCODEC_H263P },
  { "MJPEG",  (int)VCODEC_MJPEG },
  { "MPEG4",  (int)VCODEC_MPEG4 },
  { "MSMPEG4V1",  (int)VCODEC_MSMPEG4V1 },
  { "MSMPEG4V2",  (int)VCODEC_MSMPEG4V2 },
  { "MSMPEG4V3",  (int)VCODEC_MSMPEG4V3 },
  { "WMV1",  (int)VCODEC_WMV1 },
  { "WMV2",  (int)VCODEC_WMV2 },
//  { "HUFFYUV",  (int)VCODEC_HUFFYUV },
#endif /* HAVE_LIBAVCODEC */
};

#define NTVREC_PRESET_FILENAME "~/.ntvrec/ntvrec.preset"

static int
opt_set_preset(const char *arg)
{
  if (preset_init(NTVREC_PRESET_FILENAME) != OK)
    return OK;
//    return FAIL;
  if (preset_set(arg) != OK)
    return FAIL;
  if (preset_quit() != OK)
    return FAIL;
  return OK;
}

static int
opt_set_fps (const char *arg)
{
  mgr_param.fps = strtod (arg, 0);
  return 0;
}

static int
opt_set_size (const char *arg)
{
  char *p;
  char *ep;
  int width;
  int height;

  p = (char*)arg;
  width = strtol(p, &ep, 10);
  if (ep == NULL) {
    fprintf (stderr, "option frame size error. '%s'\n", arg);
    return -1;
  }
  p = ++ep;
  height = strtol(p, &ep, 10);
  if (width <= 0 || height <= 0) {
    fprintf(stderr, "Incorrect frame size\n");
    return -1;
  }
  if ((width % 16) != 0) {
    width = width - (width % 16);
    fprintf(stderr, "Frame width must be a multiple of 16, %d\n", width);
    fprintf(stderr, "force width %d\n", width);
  }
  if ((height % 16) != 0) {
    height = height - (height % 16);
    fprintf(stderr, "Frame height must be a multiple of 16, %d\n", height);
    fprintf(stderr, "force height %d\n", height);
  }
  mgr_param.width = width;
  mgr_param.height = height;
  return 0;
}

#if 0
char *use_fourcc = NULL;
const char *fourcc_str = NULL;
#endif

int
opt_set_use_vfourcc (const char *arg)
{
#if 0
  if (use_fourcc)
    free(use_fourcc);
  use_fourcc = (char*)strndup(arg, 4);
#endif
  mgr_param.use_vfourcc = arg;
  return 0;
}

int
opt_set_use_afourcc (const char *arg)
{
  mgr_param.use_afourcc = arg;
  return 0;
}

static SEL_COMPO grab_csps[] = {
//  { "I420",  CSP_I420 },
//  { "RGB24", CSP_RGB24 },
//  { "RGB32", CSP_RGB32 },
  { "YUV420",  CSP_YUV420 },
  { "YV12",  CSP_YV12 },
};

OptionDef mgr_option[] = {
#ifdef HAVE_LIBXVIDCORE
  {"vcodec", "VIDEO_CODEC", HAS_ARG|OPT_SEL, {(void*)&mgr_param.vcodec}, {VCODEC_XVID}, {(int)video_codecs}, sizeof(video_codecs)/sizeof(SEL_COMPO), "Output video codec", "codec"},
#else
  {"vcodec", "VIDEO_CODEC", HAS_ARG|OPT_SEL, {(void*)&mgr_param.vcodec}, {VCODEC_YV12}, {(int)video_codecs}, sizeof(video_codecs)/sizeof(SEL_COMPO), "Output video codec", "codec"},
#endif /* HAVE_LIBXVIDCORE */
  {"fps", "VIDEO_FPS", HAS_ARG|OPT_FUNC, {(void*)&opt_set_fps}, {0}, {0}, 0, "Video frames/second", "r"},
  {"s", "VIDEO_SIZE", HAS_ARG|OPT_FUNC, {(void*)&opt_set_size}, {(int)"320x240"}, {0}, 0, "frame size (WxH)", "WxH"},
  {"max-buffers", "VIDEO_MAX_BUFFERS", HAS_ARG, {(void*)&mgr_param.max_buffers}, {32}, {4}, 64, "Max number of buffers", "b"},
  {"use-video-fourcc", "USE_VIDEO_FOURCC", HAS_ARG|OPT_FUNC, {(void*)&opt_set_use_vfourcc}, {0}, {0}, 5, "use fourcc", "f"},
  {"grab-csp", "GRAB_CSP", HAS_ARG|OPT_SEL, {(void*)&mgr_param.grab_csp}, {CSP_UNKNOWN}, {(int)grab_csps}, sizeof(grab_csps)/sizeof(SEL_COMPO), "color space specified when grabing", "b"},
  {"preset", NULL, HAS_ARG|OPT_FUNC|OPT_BEFORE, {(void*)&opt_set_preset}, {0}, {0}, 0, "read preset from preset file.", "name"},
};

int mgr_option_num = sizeof(mgr_option) / sizeof(OptionDef);

static SEL_COMPO vd_norm[] = {
  { "PAL",  V4L_NORM_PAL },
  { "NTSC",  V4L_NORM_NTSC },
  { "SECAM",  V4L_NORM_SECAM },
  { "AUTO",  V4L_NORM_AUTO },
};

OptionDef vdev_param[] = {
  {"vd", "VIDEO_DEVICE", HAS_ARG|OPT_STR, {(void*)&mgr_param.vdev_name}, {(int)"/dev/video0"}, {0}, 1024, "video device", "name"},
  {"vd-source", "VIDEO_DEVICE_SOURCE", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_source_channel}, {0}, {0}, 4, "video device source [0,1,...]", "num"},
  {"vd-tuner", "VIDEO_DEVICE_TUNER", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_tuner}, {0}, {0}, 4, "video device tuner [0,1,...]", "num"},
  {"vd-norm", "VIDEO_DEVICE_NORM", HAS_ARG|OPT_SEL, {(void*)&mgr_param.vdev_norm}, {V4L_NORM_NTSC}, {(int)vd_norm}, sizeof(vd_norm)/sizeof(SEL_COMPO), "video device norm", "norm"},
  {"vd-audio", "VIDEO_DEVICE_AUDIO", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_audio}, {0}, {0}, 4, "video device audio [0,1,...]", "num"},
  {"vd-brightness", "VIDEO_DEVICE_BRIGHTNESS", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_brightness}, {-1}, {-1}, 100, "video device brightness", "num"},
  {"vd-hue", "VIDEO_DEVICE_HUE", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_hue}, {-1}, {-1}, 100, "video device hue", "num"},
  {"vd-color", "VIDEO_DEVICE_COLOUR", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_colour}, {-1}, {-1}, 100, "video device color", "num"},
  {"vd-contrast", "VIDEO_DEVICE_CONTRAST", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_contrast}, {-1}, {-1}, 100, "video device contrast", "num"},
  {"vd-whiteness", "VIDEO_DEVICE_WHITENESS", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_whiteness}, {-1}, {-1}, 100, "video device whiteness", "num"},
  {"vd-freq", "VIDEO_DEVICE_FREQ", HAS_ARG|OPT_INT, {(void*)&mgr_param.vdev_freq}, {-1}, {-1}, INT_MAX, "video device tuner frequency", "num"},
};
int vdev_param_num = sizeof(vdev_param) / sizeof(OptionDef);

static SEL_COMPO audio_codecs[] = {
#ifdef HAVE_LIBMP3LAME
  {"MP3", ACODEC_MP3},
#endif /* HAVE_LIBMP3LAME */
  {"PCM", ACODEC_PCM},
#ifdef HAVE_OGGVORBIS
  {"oggvorbis", ACODEC_OGGVORBIS},
#endif /* HAVE_OGGVORBIS */
};

#if 0
static SEL_COMPO audio_formats[] = {
  {"U8", AUDIO_FMT_U8},
  {"S16LE", AUDIO_FMT_S16_LE}
};
#endif

OptionDef audio_param[] = {
  {"ad", "AUDIO_DEVICE", HAS_ARG|OPT_STR, {(void*)&mgr_param.adev_name}, {(int)"/dev/dsp"}, {0}, 1024, "audio device", "name"},
  {"ad-mixer", "AUDIO_MIXER_DEVICE", HAS_ARG|OPT_STR, {(void*)&mgr_param.adev_mixer_name}, {(int)"/dev/mixer"}, {0}, 1024, "audio mixer device", "name"},
  {"ad-source", "AUDIO_DEVICE_SOURCE", HAS_ARG|OPT_STR, {(void*)&mgr_param.adev_source}, {(int)NULL}, {0}, 32, "audio record source", "name"},

#ifdef HAVE_LIBMP3LAME
  {"acodec", "AUDIO_CODEC", HAS_ARG|OPT_SEL, {(void*)&mgr_param.acodec}, {ACODEC_MP3}, {(int)audio_codecs}, sizeof(audio_codecs)/sizeof(SEL_COMPO), "force audio codec ", "c"},
#else
  {"acodec", "AUDIO_CODEC", HAS_ARG|OPT_SEL, {(void*)&mgr_param.acodec}, {ACODEC_PCM}, {(int)audio_codecs}, sizeof(audio_codecs)/sizeof(SEL_COMPO), "force audio codec ", "c"},
#endif /* HAVE_LIBMP3LAME */
  {"audio-rate", "AUDIO_RATE", HAS_ARG, {(void*)&mgr_param.audio_rate}, {22050}, {0}, 96000, "set audio sampling rate [in Hz]", "r"},
  {"audio-channels", "AUDIO_CHANNEL", HAS_ARG, {(void*)&mgr_param.audio_channels}, {2}, {0}, 2, "set number of audio channels", "c"},
  {"use-audio-fourcc", "USE_AUDIO_FOURCC", HAS_ARG|OPT_FUNC, {(void*)&opt_set_use_afourcc}, {0}, {0}, 5, "use audio fourcc", "s"},
//  {"audio_format", "AUDIO_FORMAT", HAS_ARG|OPT_SEL, {(void*)&mgr_param.audio_format}, {AUDIO_FMT_S16_LE}, {(int)audio_formats}, sizeof(audio_formats)/sizeof(SEL_COMPO), "set audio format ", "c"},
};

int audio_param_num = sizeof(audio_param) / sizeof(OptionDef);

#if 0
static SEL_COMPO fmts[] = {
  { "AVI",  FFMT_FORMAT_AVI },
  { "RAW",  FFMT_FORMAT_RAW },
  { "MOV",  FFMT_FORMAT_MOV},
};
#endif

int
mgr_set_output_file_name (const char *arg)
{
  mgr_param.output_fname = (char*)strndup(arg, MAX_FILE_NAME_LENGTH);
  return 0;
}

static SEL_COMPO sync_modes[] = {
  { "no",  SYNC_NONE },
  { "system",  SYNC_SYSTEM },
  { "audio",  SYNC_AUDIO  },
};

static SEL_COMPO rec_media[] = {
  { "both",   BOTH_MEDIA },
  { "video",  VIDEO_MEDIA },
  { "audio",  AUDIO_MEDIA },
};

OptionDef ffmt_param[] = {
  {"sync", "SYNC", HAS_ARG|OPT_SEL, {(void*)&mgr_param.sync_mode}, {SYNC_AUDIO}, {(int)sync_modes}, sizeof(sync_modes)/sizeof(SEL_COMPO), "video and audio sync mode", "sel"},
  {"media", "MEDIA", HAS_ARG|OPT_SEL, {(void*)&mgr_param.record_media}, {BOTH_MEDIA}, {(int)rec_media}, sizeof(rec_media)/sizeof(SEL_COMPO), "select record media", "sel"},
//  {"file-format", "FILE_FORMAT", HAS_ARG|OPT_SEL, {(void*)&mgr_param.file_format}, {FFMT_FORMAT_MOV}, {(int)fmts}, sizeof(fmts)/sizeof(SEL_COMPO), "Output file format", "sel"},
  {"o", NULL, HAS_ARG|OPT_FUNC, {(void*)&mgr_set_output_file_name}, {0}, {0}, MAX_FILE_NAME_LENGTH, "Output file name", "name"},
};

int ffmt_param_num = sizeof(ffmt_param) / sizeof(OptionDef);


static void mgr_error(const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "mgr: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);
  exit(1);
}

static void
mgr_error_cb_vgrab(void *cb_data)
{
  mgr_t *mgr = (mgr_t*)cb_data;
  pthread_mutex_lock(&mgr->mutex);
  mgr->error_state |= VGRAB_THREAD;
  pthread_mutex_unlock(&mgr->mutex);
}

static void
mgr_error_cb_aread(void *cb_data)
{
  mgr_t *mgr = (mgr_t*)cb_data;
  pthread_mutex_lock(&mgr->mutex);
  mgr->error_state |= AREAD_THREAD;
  pthread_mutex_unlock(&mgr->mutex);
}

static void
mgr_error_cb_venc(void *cb_data)
{
  mgr_t *mgr = (mgr_t*)cb_data;
  pthread_mutex_lock(&mgr->mutex);
  mgr->error_state |= VENC_THREAD;
  pthread_mutex_unlock(&mgr->mutex);
}

static void
mgr_error_cb_aenc(void *cb_data)
{
  mgr_t *mgr = (mgr_t*)cb_data;
  pthread_mutex_lock(&mgr->mutex);
  mgr->error_state |= AENC_THREAD;
  pthread_mutex_unlock(&mgr->mutex);
}

static void
mgr_error_cb_ffmt(void *cb_data)
{
  mgr_t *mgr = (mgr_t*)cb_data;
  pthread_mutex_lock(&mgr->mutex);
  mgr->error_state |= FFMT_THREAD;
  pthread_mutex_unlock(&mgr->mutex);
}

int
mgr_error_check(mgr_t *mgr)
{
  int ret = 0;
  pthread_mutex_lock(&mgr->mutex);
  ret = mgr->error_state;
  pthread_mutex_unlock(&mgr->mutex);
  return ret;
}

static u_time_t
get_time(void)
{
  return get_cur_u_time();
}

double
mgr_get_bitrate(mgr_t *mgr)
{
  return ffmt_get_bitrate(mgr->ffmt);
}

double
mgr_get_record_time(mgr_t *mgr)
{
  return ffmt_get_record_time(mgr->ffmt);
}

long long
mgr_get_total_data_size(mgr_t *mgr)
{
  return ffmt_get_total_data_size(mgr->ffmt);
}

int
mgr_get_video_frame_count(mgr_t *mgr)
{
  return ffmt_get_video_frame_count(mgr->ffmt);
}

int
mgr_get_record_media(mgr_t *mgr)
{
  if (!mgr) return -1;
  return (mgr_param.record_media);
}

int
mgr_state_start(mgr_t *mgr)
{
  int retval = MGR_FAIL;
  u_time_t state_time;

  if (mgr->state == MGR_STATE_START)
    return MGR_OK;

  state_time = get_time();

  if (mgr->vgrab && vgrab_state_start(mgr->vgrab, &state_time) != VGRAB_OK) {
    mgr_error("mgr_state_start: vgrab_state_start failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_start: vgrab_state_start done\n");
  if (mgr->aread && aread_state_start(mgr->aread, &state_time) != AREAD_OK) {
    mgr_error("mgr_state_start: aread_state_start failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_start: aread_state_start done\n");
  if (mgr->venc && venc_state_start(mgr->venc) != VENC_OK) {
    mgr_error("mgr_state_start: venc_state_start failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_start: venc_state_start done\n");
  if (mgr->aenc && aenc_state_start(mgr->aenc) != AENC_OK) {
    mgr_error("mgr_state_start: aenc_state_start failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_start: aenc_state_start done\n");
  if (ffmt_state_start(mgr->ffmt) != FFMT_OK) {
    mgr_error("mgr_state_start: ffmt_state_start failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_start: ffmt_state_start done\n");
  mgr->state = MGR_STATE_START;
  debug("mgr_state_start: done\n");
  return MGR_OK;
}

int
mgr_state_toggle_pause(mgr_t *mgr)
{
  int retval = MGR_FAIL;
  u_time_t state_time;

  if (mgr->state != MGR_STATE_START && mgr->state != MGR_STATE_PAUSE)
    return MGR_OK;

  state_time = get_time();

  if (mgr->vgrab &&
      vgrab_state_toggle_pause(mgr->vgrab, &state_time) != VGRAB_OK) {
    mgr_error("mgr_state_toggle_pause: vgrab_state_toggle_pause failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_toggle_pause: vgrab_state_toggle_pause done.\n");
  if (mgr->aread &&
      aread_state_toggle_pause(mgr->aread, &state_time) != AREAD_OK) {
    mgr_error("mgr_state_toggle_pause: aread_state_toggle_pause failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_toggle_pause: aread_state_toggle_pause done.\n");
#if 0
  if (mgr->venc && venc_state_toggle_pause(mgr->venc) != VENC_OK) {
    mgr_error("mgr_state_toggle_pause: venc_state_toggle_pause failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_toggle_pause: venc_state_toggle_pause done.\n");
  if (mgr->aenc && aenc_state_toggle_pause(mgr->aenc) != AENC_OK) {
    mgr_error("mgr_state_toggle_pause: aenc_state_toggle_pause failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_toggle_pause: aenc_state_toggle_pause done.\n");
  if (ffmt_state_toggle_pause(mgr->ffmt) != FFMT_OK) {
    mgr_error("mgr_state_toggle_pause: ffmt_state_toggle_pause failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_toggle_pause: ffmt_state_toggle_pause done.\n");
#endif
  if (mgr->state == MGR_STATE_START)
    mgr->state = MGR_STATE_PAUSE;
  else if (mgr->state == MGR_STATE_PAUSE)
    mgr->state = MGR_STATE_PAUSE;
  else
    return MGR_FAIL;
  debug("mgr_state_toggle_pause: done.\n");
  return MGR_OK;
}

int
mgr_state_stop(mgr_t *mgr)
{
  int retval = MGR_FAIL;
  u_time_t state_time;

  if (mgr->state & MGR_STATE_STOP)
    return MGR_OK;

  state_time = get_time();

  if (mgr->vgrab && vgrab_state_stop(mgr->vgrab, &state_time) != VGRAB_OK) {
    mgr_error("mgr_state_stop: vgrab_state_stop failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_stop: vgrab_state_stop done\n");
  if (mgr->aread && aread_state_stop(mgr->aread, &state_time) != AREAD_OK) {
    mgr_error("mgr_state_stop: aread_state_stop failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_stop: aread_state_stop done\n");
  if (mgr->venc && venc_state_stop(mgr->venc) != VENC_OK) {
    mgr_error("mgr_state_stop: venc_state_stop failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_stop: venc_state_stop done\n");
  if (mgr->aenc && aenc_state_stop(mgr->aenc) != AENC_OK) {
    mgr_error("mgr_state_stop: aenc_state_stop failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_stop: aenc_state_stop done\n");
  if (ffmt_state_stop(mgr->ffmt) != FFMT_OK) {
    mgr_error("mgr_state_stop: ffmt_state_stop failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_stop: ffmt_state_stop done\n");
  mgr->state = MGR_STATE_STOP;
  debug("mgr_state_stop: done\n");
  return MGR_OK;
}

int
mgr_state_quit(mgr_t *mgr)
{
  int retval = MGR_FAIL;
  u_time_t state_time;
  double time;

  if (mgr->state == MGR_STATE_QUIT)
    return MGR_OK;

  state_time = get_time();
  time = state_time;

  if (mgr->vgrab && vgrab_state_quit(mgr->vgrab, &state_time) != VGRAB_OK) {
    mgr_error("mgr_state_quit: vgrab_state_quit failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_quit: vgrab_state_quit done\n");
//fprintf(stderr, "time %f\n", ((double)state_time - time)/1000000.);
  if (mgr->aread && aread_state_quit(mgr->aread, &state_time) != AREAD_OK) {
    mgr_error("mgr_state_quit: aread_state_quit failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_quit: aread_state_quit done\n");
  if (mgr->venc && venc_state_quit(mgr->venc) != VENC_OK) {
    mgr_error("mgr_state_quit: venc_state_quit failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_quit: venc_state_quit done\n");
  if (mgr->aenc && aenc_state_quit(mgr->aenc) != AENC_OK) {
    mgr_error("mgr_state_quit: aenc_state_quit failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_quit: aenc_state_quit done\n");
  if (ffmt_state_quit(mgr->ffmt) != FFMT_OK) {
    mgr_error("mgr_state_quit: ffmt_state_quit failed.\n");
    mgr_error_quit(mgr);
    return retval;
  }
  debug("mgr_state_quit: ffmt_state_quit done\n");
  mgr->state = MGR_STATE_QUIT;
  debug("mgr_state_quit: done\n");
  return MGR_OK;
}

int
mgr_state_error(mgr_t *mgr)
{
  int retval = MGR_FAIL;
  u_time_t state_time;

  if (mgr->state == MGR_STATE_ERROR)
    return MGR_OK;

  state_time = get_time();

  if (mgr->vgrab && vgrab_state_error(mgr->vgrab, &state_time) != VGRAB_OK) {
    mgr_error("mgr_state_error: vgrab_state_error failed.\n");
    return retval;
  }
  debug("mgr_state_error: vgrab_state_error done\n");
  if (mgr->aread && aread_state_error(mgr->aread, &state_time) != AREAD_OK) {
    mgr_error("mgr_state_error: aread_state_error failed.\n");
    return retval;
  }
  debug("mgr_state_error: aread_state_error done\n");
  if (mgr->venc && venc_state_error(mgr->venc) != VENC_OK) {
    mgr_error("mgr_state_error: venc_state_error failed.\n");
    return retval;
  }
  debug("mgr_state_error: venc_state_error done\n");
  if (mgr->aenc && aenc_state_error(mgr->aenc) != AENC_OK) {
    mgr_error("mgr_state_error: aenc_state_error failed.\n");
    return retval;
  }
  debug("mgr_state_error: aenc_state_error done\n");
  if (ffmt_state_error(mgr->ffmt) != FFMT_OK) {
    mgr_error("mgr_state_error: ffmt_state_error failed.\n");
    return retval;
  }
  debug("mgr_state_error: ffmt_state_error done\n");
  mgr->state = MGR_STATE_ERROR;
  debug("mgr_state_error: done\n");
  return MGR_OK;
}

int
mgr_error_quit(mgr_t *mgr)
{
  int ret = MGR_OK;
  u_time_t state_time;

  if (!mgr)
    return MGR_OK;

  if (mgr->state != MGR_STATE_ERROR)
    ret = mgr_state_error(mgr);

  state_time = get_time();

  if (mgr->vgrab)
    vgrab_error_quit(mgr->vgrab, &state_time);
debug("mgr_error_quit: vgrab_error_quit done\n");
  if (mgr->aread)
    aread_error_quit(mgr->aread, &state_time);
debug("mgr_error_quit: aread_error_quit done\n");
  if (mgr->venc)
    venc_error_quit(mgr->venc);
debug("mgr_error_quit: venc_error_quit done\n");
  if (mgr->aenc)
    aenc_error_quit(mgr->aenc);
debug("mgr_error_quit: aenc_error_quit done\n");
  if (mgr->ffmt)
    ffmt_error_quit(mgr->ffmt);
debug("mgr_error_quit: ffmt_error_quit done\n");

  if (mgr->vfrm_mgr && vframe_mgr_quit(mgr->vfrm_mgr) != VFRAME_MGR_OK) {
    mgr_error("mgr_quit: vframe_mgr_quit() failed\n");
    ret = MGR_FAIL;
  }

  if (mgr->afrm_mgr && aframe_mgr_quit(mgr->afrm_mgr) != AFRAME_MGR_OK) {
    mgr_error("mgr_quit: aframe_mgr_quit() failed\n");
    ret = MGR_FAIL;
  }

  free(mgr);

  debug("mgr_error_quit: done.\n");

  return ret;
}

int
mgr_quit(mgr_t *mgr)
{
  int ret = MGR_OK;
  u_time_t state_time;

  if (mgr->state != MGR_STATE_ERROR && mgr->state != MGR_STATE_QUIT)
    ret = mgr_state_quit(mgr);

  state_time = get_time();

  if (mgr->vgrab)
    vgrab_quit(mgr->vgrab, &state_time);
debug("mgr_quit: vgrab_quit done\n");
  if (mgr->aread)
    aread_quit(mgr->aread, &state_time);
debug("mgr_quit: aread_quit done\n");
  if (mgr->venc)
    venc_quit(mgr->venc);
debug("mgr_quit: venc_quit done\n");
  if (mgr->aenc)
    aenc_quit(mgr->aenc);
debug("mgr_quit: aenc_quit done\n");
  ffmt_quit(mgr->ffmt);
debug("mgr_quit: ffmt_quit done\n");

  if (mgr->vfrm_mgr && vframe_mgr_quit(mgr->vfrm_mgr) != VFRAME_MGR_OK) {
    mgr_error("mgr_quit: vframe_mgr_quit() failed\n");
    ret = MGR_FAIL;
  }

  if (mgr->afrm_mgr && aframe_mgr_quit(mgr->afrm_mgr) != AFRAME_MGR_OK) {
    mgr_error("mgr_quit: aframe_mgr_quit() failed\n");
    ret = MGR_FAIL;
  }

  free(mgr);

debug("mgr_quit: done.\n");
  return ret;
}

static int
mgr_video_init(mgr_t *mgr)
{
  MGR_PARAM *prm = &mgr_param;
  int grab_csp;
  double fps;
  v4l_param_t v4l_param;
  u_time_t state_time;

  if (vcodec_encode_set_codec(prm->vcodec) == VCODEC_FAIL) {
    mgr_error("mgr_video_init: vcodec_encode_set_codec failed\n");
    return MGR_FAIL;
  }
  grab_csp = vcodec_encode_csp_cap(prm->grab_csp);
  if (grab_csp == CSP_UNKNOWN) {
    mgr_error("mgr_video_init: grab_csp invalied.\n");
    return MGR_FAIL;
  }
  prm->grab_csp = grab_csp;

  v4l_param.device_name = prm->vdev_name;
  v4l_param.source_channel = prm->vdev_source_channel;
  v4l_param.tuner = prm->vdev_tuner;
  v4l_param.norm = prm->vdev_norm;
  v4l_param.audio = prm->vdev_audio;
  v4l_param.brightness = prm->vdev_brightness;
  v4l_param.hue = prm->vdev_hue;
  v4l_param.colour = prm->vdev_colour;
  v4l_param.contrast = prm->vdev_contrast;
  v4l_param.whiteness = prm->vdev_whiteness;
  v4l_param.freq = prm->vdev_freq;

  mgr->vgrab = vgrab_init(prm->width, prm->height, prm->fps, grab_csp, &v4l_param);
  if (!mgr->vgrab) {
    mgr_error("mgr_video_init: vgrab_init failed.\n");
    return MGR_FAIL;
  }

  prm->width = vgrab_get_width(mgr->vgrab);
  prm->height = vgrab_get_height(mgr->vgrab);
  fps = vgrab_get_fps(mgr->vgrab);

  mgr->venc = venc_init(prm->width, prm->height, fps, prm->vcodec, grab_csp, mgr->vgrab);
  if (!mgr->venc) {
    mgr_error("mgr_video_init: venc_init failed.\n");
    vgrab_quit(mgr->vgrab, &state_time);
    mgr->vgrab = NULL;
    return MGR_FAIL;
  }

  mgr->vfrm_mgr = vframe_mgr_init(prm->width, prm->height, 0, 0, grab_csp, prm->max_buffers);
  if (!mgr->vfrm_mgr) {
    mgr_error("mgr_video_init: vframe_mgr_init failed.\n");
    vgrab_quit(mgr->vgrab, &state_time);
    mgr->vgrab = NULL;
    venc_quit(mgr->venc);
    mgr->venc = NULL;
    return MGR_FAIL;
  }

  vgrab_set_vframe_mgr(mgr->vgrab, mgr->vfrm_mgr);
  venc_set_vframe_mgr(mgr->venc, mgr->vfrm_mgr);

  return MGR_OK;
}

static int
mgr_audio_init(mgr_t *mgr)
{
  MGR_PARAM *prm = &mgr_param;
  int rate, channels, bits;
  uint32_t acodec;
  int sync_system;
  u_time_t state_time;
  double buffer_time;


  buffer_time = AUDIO_BUFFER_TIME;

  if (prm->sync_mode == SYNC_SYSTEM)
    sync_system = 1;
  else
    sync_system = 0;

  rate = prm->audio_rate;
  channels = prm->audio_channels;
  bits = prm->audio_bits;
  acodec = prm->acodec;

  if (acodec_encode_set_codec(acodec) == ACODEC_FAIL) {
    mgr_error("mgr_audio_init: acodec_encode_set_codec failed\n");
    return MGR_FAIL;
  }

  if ((rate = acodec_encode_rate_cap(prm->audio_rate)) != prm->audio_rate) {
    mgr_error("mgr_audio_init: audio rate failed\n");
    return MGR_FAIL;
  }
  if ((channels = acodec_encode_channels_cap(prm->audio_channels)) != prm->audio_channels) {
    mgr_error("mgr_audio_init: audio channels failed\n");
    return MGR_FAIL;
  }
  prm->audio_bits = 16;
  if ((bits = acodec_encode_bits_cap(prm->audio_bits)) != prm->audio_bits) {
    mgr_error("mgr_audio_init: audio bits failed\n");
    return MGR_FAIL;
  }

  mgr->aread = aread_init(rate, channels, bits, sync_system, prm->adev_name, prm->adev_mixer_name, prm->adev_source);
  if (!mgr->aread) {
    mgr_error("mgr_audio_init: aread init failed\n");
    return MGR_FAIL;
  }
  rate = aread_get_rate(mgr->aread);
  channels = aread_get_channels(mgr->aread);
  bits = aread_get_bits_per_sample(mgr->aread);

  if (rate != prm->audio_rate || channels != prm->audio_channels ||
      bits != prm->audio_bits) {
    mgr_error("mgr_audio_init: oss_init parameter fialed.\n");
    aread_quit(mgr->aread, &state_time);
    mgr->aread = NULL;
    return MGR_FAIL;
  }
  prm->audio_rate = rate;
  prm->audio_channels = channels;
  prm->audio_bits = bits;

  mgr->aenc = aenc_init(rate, channels, bits, acodec, mgr->aread);
  if (!mgr->aenc) {
    mgr_error("mgr_audio_init: aenc init failed\n");
    aread_quit(mgr->aread, &state_time);
    mgr->aread = NULL;
    return MGR_FAIL;
  }
  prm->audio_frame_sample_size = aenc_get_frame_sample_size(mgr->aenc);

  mgr->afrm_mgr = aframe_mgr_init(rate, channels, bits, prm->audio_frame_sample_size, buffer_time);
  if (!mgr->afrm_mgr) {
    mgr_error("mgr_audio_init: aframe_mgr_audio_init failed\n");
    aread_quit(mgr->aread, &state_time);
    mgr->aread = NULL;
    aenc_quit(mgr->aenc);
    mgr->aenc = NULL;
    return MGR_FAIL;
  }
  aread_set_aframe_mgr(mgr->aread, mgr->afrm_mgr);
  aenc_set_aframe_mgr(mgr->aenc, mgr->afrm_mgr);

  prm->audio_out_rate = aenc_get_out_rate(mgr->aenc);

  return MGR_OK;
}

mgr_t*
mgr_init(void)
{
  int ret;
  mgr_t *mgr;
  MGR_PARAM *prm = &mgr_param;
  int sync_audio;


  mgr = (mgr_t*)malloc(sizeof(mgr_t));
  if (!mgr) {
    fprintf(stderr, "mgr_init: memory error.\n");
    return NULL;
  }

  pthread_mutex_init(&mgr->mutex, NULL);

  mgr->state = MGR_STATE_STOP;
  mgr->error_state = 0;
  mgr->vgrab = NULL;
  mgr->venc  = NULL;
  mgr->aread = NULL;
  mgr->aenc  = NULL;
  mgr->ffmt  = NULL;
  mgr->vfrm_mgr = NULL;
  mgr->afrm_mgr = NULL;

//  if (prm->file_format == FFMT_FORMAT_MOV &&
  if (prm->acodec == ACODEC_PCM) {
    prm->acodec = ACODEC_PCM_BE;
  }

/* audio */
  if (!(prm->record_media == VIDEO_MEDIA)) {
    if (mgr_audio_init(mgr) == MGR_FAIL) {
      mgr_error("mgr_init: mgr_audio_init failed\n");
      mgr_error_quit(mgr);
      return NULL;
    }
  }

/* video */
  if (!(prm->record_media == AUDIO_MEDIA)) {
    if (mgr_video_init(mgr) == MGR_FAIL) {
      mgr_error("mgr_init: mgr_video_init failed\n");
      mgr_error_quit(mgr);
      return NULL;
    }
  }

/* file format */
  if (prm->sync_mode == SYNC_AUDIO)
    sync_audio = 1;
  else
    sync_audio = 0;

  mgr->ffmt = ffmt_init(prm->output_fname,
      prm->width, prm->height, prm->fps, prm->vcodec, prm->use_vfourcc,
      prm->acodec, prm->audio_out_rate, prm->audio_channels, prm->audio_bits,
      prm->use_afourcc, sync_audio, mgr->venc, mgr->aenc);

  if (!mgr->ffmt) {
    mgr_error("mgr_init: ffmt_init failer.\n");
    mgr_error_quit(mgr);
    return NULL;
  }
  ffmt_set_vframe_mgr(mgr->ffmt, mgr->vfrm_mgr);
  ffmt_set_aframe_mgr(mgr->ffmt, mgr->afrm_mgr);

  vgrab_set_error_cb(mgr->vgrab, mgr_error_cb_vgrab, mgr);
  aread_set_error_cb(mgr->aread, mgr_error_cb_aread, mgr);
  venc_set_error_cb(mgr->venc, mgr_error_cb_venc, mgr);
  aenc_set_error_cb(mgr->aenc, mgr_error_cb_aenc, mgr);
  ffmt_set_error_cb(mgr->ffmt, mgr_error_cb_ffmt, mgr);

  ret = mgr_state_stop(mgr);
  if (ret != MGR_OK) {
    mgr_error("mgr_init: mgr_state_stop failer.\n");
    mgr_error_quit(mgr);
    return NULL;
  }

  return mgr;
}

static void
mgr_print_param_video(mgr_t *mgr)
{
  MGR_PARAM *prm = &mgr_param;

  fprintf(stdout, "\n");
  fprintf(stdout, "VIDEO SIZE:              %dx%d\n", prm->width, prm->height);
  if (prm->fps == 0)
    fprintf(stdout, "VIDEO FPS:               0 [auto]\n");
  else
    fprintf(stdout, "VIDEO FPS:               %f\n", prm->fps);
  if (prm->grab_csp != CSP_UNKNOWN)
    fprintf(stdout, "VIDEO GRAB CSP:          %s\n",csp_to_name(prm->grab_csp));
  fprintf(stdout, "VIDEO MAX BUFFERS:       %d\n", prm->max_buffers);
  v4l_print_param(mgr->vgrab->v4l);
  vcodec_encode_print_param();
}

static void
mgr_print_param_audio(mgr_t *mgr)
{
  MGR_PARAM *prm = &mgr_param;

  fprintf(stdout, "\n");
  fprintf(stdout, "AUDIO RATE:              %d\n", prm->audio_rate);
  fprintf(stdout, "AUDIO CHANNELS:          %d\n", prm->audio_channels);
  fprintf(stdout, "AUDIO BITS:              %d\n", prm->audio_bits);
  oss_print_param(mgr->aread->oss);
  acodec_encode_print_param();
}

void
mgr_print_param(mgr_t *mgr)
{
  fprintf(stdout, "ntvrec version %s\n", VERSION);
  mgr_print_param_video(mgr);
  mgr_print_param_audio(mgr);
  ffmt_print_param(mgr->ffmt);
}

