/* deno2.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <malloc.h>

#include "denoise2.h"


#define CLIP(v) clip[(v)]

#if 1
#define CLIP_CHECK(v)
#else
#define CLIP_CHECK(v) \
  if ((v)<-384 || (v)>640) {       \
    fprintf(stderr,"deno2: clipping value %d\n",(v)); \
    (v) = (((v)<0)?0:((v)>255)?255:(v)); \
  }
#endif

#define BUF_NUM 4
#define BORDER  4

typedef double (*FILT_FUNC)(double x);

static int filt_coef0[5];
static int filt_coef1[5];

static int Width, Height;
static int Width2, Height2;
static int Threshold = 6;
static int uninited = 1;
static int start_offset[3];
static int stride[3];
static unsigned char *frmbuf[BUF_NUM][3];
static unsigned char *clip;
static unsigned char clip_buf[1024];

static double filt_cubic(double x)         /* 4th order (cubic) b-spline */
{
    double t;

    if (x<-2.) return 0.;
    if (x<-1.) {t = 2.+x; return t*t*t/6.;}
    if (x<0.) return (4.+x*x*(-6.+x*-3.))/6.;
    if (x<1.) return (4.+x*x*(-6.+x*3.))/6.;
    if (x<2.) {t = 2.-x; return t*t*t/6.;}
    return 0.;
}

static double filt_catrom(double x)        /* Catmull-Rom spline, Overhauser spline */
{
    if (x<-2.) return 0.;
    if (x<-1.) return .5*(4.+x*(8.+x*(5.+x)));
    if (x<0.) return .5*(2.+x*x*(-5.+x*-3.));
    if (x<1.) return .5*(2.+x*x*(-5.+x*3.));
    if (x<2.) return .5*(4.+x*(-8.+x*(5.-x)));
    return 0.;
}

static int
make_filt_coef(double blur, FILT_FUNC func, int coefs[], double supp)
{
  double filt_scale;
  double den, sc, tr;
  int i, c, ti, sum;

  if (func == NULL)
    func = filt_catrom;
  filt_scale = blur;
  if (supp == 0.)
    supp = 2.0;

  den = 0;
  for (i = -2; i <= 2; i++)
    den += func(((double)i) / filt_scale);
  sc = den==0. ? 1 : 1/den;
  sum = 0;
  c = 0;
  for (i = -2; i <= 2; i++) {
    tr = func(((double)i) / filt_scale);
    tr = sc * tr * 256;
    ti = tr + 0.5;
    sum = sum + ti;
    coefs[c++] = ti;
  }

  if (sum != 256)
    coefs[2] += 256 - sum;

  return 0;
}

static void
lowp_filt(int width, int height, unsigned char *src, int src_stride, unsigned char *dst, int dst_stride, int coef[])
{
  int i, j;
  unsigned char *sp, *dp;
  unsigned char *bpm2, *bpm1, *bp, *bpp1, *bpp2;
  unsigned char buf[width*height];
  int v;

  sp = src;
  bp = buf;
  for (j = 0; j < height; j++) {
    i = 0;
    v = (  coef[0]*sp[i]   + coef[1]*sp[i]   + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+2] + 128) >> 8;
    CLIP_CHECK(v)
    bp[i++] = CLIP(v);
    v = (  coef[0]*sp[i-1] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+2] + 128) >> 8;
    CLIP_CHECK(v)
    bp[i++] = CLIP(v);
    for (; i < width-2; i++) {
      v     = (  coef[0]*sp[i-2]
               + coef[1]*sp[i-1]
               + coef[2]*sp[i]
               + coef[3]*sp[i+1]
               + coef[4]*sp[i+2] + 128) >> 8;
      CLIP_CHECK(v)
      bp[i] = CLIP(v);
    }
    v = (  coef[0]*sp[i-2] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i+1] + coef[4]*sp[i+1] + 128) >> 8;
    CLIP_CHECK(v)
    bp[i++] = CLIP(v);
    v = (  coef[0]*sp[i-2] + coef[1]*sp[i-1] + coef[2]*sp[i]
         + coef[3]*sp[i] + coef[4]*sp[i] + 128) >> 8;
    CLIP_CHECK(v)
    bp[i] = CLIP(v);
    sp += src_stride;
    bp += width;
  }

  dp = dst;
  bpm2 = bpm1 = bp = buf;
  bpp1 = buf + width;
  bpp2 = bpp1 + width;
  for (j = 0; j < height; j++) {
    for (i = 0; i < width; i++) {
      v     = (  coef[0]*bpm2[i]
	       + coef[1]*bpm1[i]
	       + coef[2]*bp[i]
	       + coef[3]*bpp1[i]
	       + coef[4]*bpp2[i] + 128) >> 8;
      CLIP_CHECK(v)
      dp[i] = CLIP(v);
    }
    dp += dst_stride;
    bpm2 = bpm1; bpm1 = bp; bp = bpp1; bpp1 = bpp2;
    if (j < height-2)
      bpp2 += width;
  }
}

/* num must be odd */
static inline int
med_sort(int data[], int num, int shift)
{
  int i;
  int big_num, small_num;
  int big_count, small_count;
  int *big, *small, *src, *tmp;
  int arr[3][num];
  int val;

  big_num = small_num = num >> 1;
  if (!(num & 1)) {
    big_num -= 1;
  }
  big_num -= shift;
  small_num += shift;
  if (big_num <= 0 || small_num <= 0) {
    fprintf(stderr, "med_sort: invalid number, big %d, small %d.\n", big_num, small_num);
    exit(1);
  }

  src = arr[0];
  big = arr[1];
  small = arr[2];

  val = data[0];
  big_count = small_count = 0;
  for (i = 1; i < num; i++) {
    if (data[i] > val) {
      big[big_count] = data[i];
      big_count++;
//    } else if (data[i] <= val) {
    } else {
      small[small_count] = data[i];
      small_count++;
    }
  }

  if (big_count == big_num || small_count == small_num) {
    return val;
  }
  else if (big_count < big_num) {
    num = num - big_count - 1;
    big_num = num - small_num - 1;
    tmp = src; src = small; small = tmp;
  }
//  else if (small_count < small_num) {
  else {
    num = num - small_count - 1;
    small_num = num - big_num - 1;
    tmp = src; src = big; big = tmp;
  }

  do {
    val = src[0];
    big_count = small_count = 0;
    for (i = 1; i < num; i++) {
      if (src[i] > val) {
        big[big_count] = src[i];
        big_count++;
//    } else if (src[i] <= val) {
      } else {
        small[small_count] = src[i];
        small_count++;
      }
    }

    if (big_count == big_num || small_count == small_num) {
      return val;
    }
    else if (big_count < big_num) {
      num = num - big_count - 1;
      big_num = num - small_num - 1;
      tmp = src; src = small; small = tmp;
    }
//  else if (small_count < small_num) {
    else {
      num = num - small_count - 1;
      small_num = num - big_num - 1;
      tmp = src; src = big; big = tmp;
    }
  } while (1);

  // never reached
  return val;
}

static inline int
med_filt(unsigned char *frm, int stride)
{
  int point[13];
  int v;

  point[0]  = *frm;
  point[1]  = frm[-1];
  point[2]  = frm[1];

  point[3]  = frm[-stride];
  point[4]  = frm[stride];
  point[5]  = frm[-stride*2];
  point[6]  = frm[stride*2];

  point[7]  = *frm - Threshold;
  point[8]  = *frm + Threshold;

//  v = med_sort(point, 9, 0);

  point[9]  = frm[-1] - Threshold;
  point[10] = frm[-1] + Threshold;
  point[11] = frm[1] - Threshold;
  point[12] = frm[1] + Threshold;

  v = med_sort(point, 13, 0);
  return v;
}

static void
laplacian(int width, int height, unsigned char *src, int src_stride, unsigned char *dst, int dst_stride)
{
  unsigned char *sp, *spm1, *spp1, *dp;
  int v, i, j;

  spm1 = sp = src;
  spp1 = sp + src_stride;
  dp = dst;
  for (j = 0; j < height; j++)
  {
    i = 0;
    v = (12*sp[i] - sp[i] - sp[i+1] - spm1[i] - spp1[i])>>3;
    CLIP_CHECK(v)
    dp[i++] = CLIP(v);
    for (i = 0; i < width-1; i++)
    {
      v = (12*sp[i] - sp[i-1] - sp[i+1] - spm1[i] - spp1[i])>>3;
      CLIP_CHECK(v)
      dp[i] = CLIP(v);
    }
    v = (12*sp[i] - sp[i-1] - sp[i] - spm1[i] - spp1[i])>>3;
    CLIP_CHECK(v)
    dp[i++] = CLIP(v);
    dp += dst_stride;
    spm1 = sp; sp = spp1;
    if (j < height-1)
      spp1 += src_stride;
  }
}

static void
deno_frame(int width, int height, unsigned char *frmm1, unsigned char *frm, unsigned char *frmp1, int frm_stride)
{
  int i, j;
  int v;
  int d0, d1;
  int thre = Threshold;
  unsigned char *fp;;

  fp = frm;
  for (j = 0; j < height; j++)
  {
    for (i = 0; i < width; i++)
    {
      v = frm[i];

      d0 = v - frmm1[i];
      d0 = (d0<0) ? -d0 : d0;

      if (d0 > thre) {
        v = med_filt(&frm[i], frm_stride);
        d0 = v - frmm1[i];
        d0 = (d0<0) ? -d0 : d0;
      }

      d1 = v - frmp1[i];
      d1 = (d1<0) ? -d1 : d1;

      if (d0 <= thre && d1 <= thre) {
	v = (v + frmm1[i]*2 + frmp1[i])>>2;
      }
      else if (d0 <= thre) {
        //v = (v + frmm1[i]*3) >> 2;
        //v = (v*(256-thre*2) + frmm1[i]*(thre*2)) >> 8;
        v = (v*3 + frmm1[i]) >> 2;
      }
      else if (d1 <= thre) {
        v = (v*3 + frmp1[i]) >> 2;
        //v = (v*(256-thre*2) + frmp1[i]*(thre*2)) >> 8;
      }

      CLIP_CHECK(v)
      frm[i] = CLIP(v);
    }
    frm += frm_stride;
    frmm1 += frm_stride;
    frmp1 += frm_stride;
  }

  fp += frm_stride;
  for (j = 1; j < height-1; j++)
  {
    for (i = 1; i < width-1; i++)
    {
      v = fp[i];

      d0 = v - fp[i-1];
      d0 = (d0<0) ? -d0 : d0;

      d1 = v - fp[i+1];
      d1 = (d1<0) ? -d1 : d1;

      if (d0 <= thre && d1 <= thre) {
	v = (v*2 + fp[i-1] + fp[i+1])>>2;
      }

      d0 = v - fp[i-frm_stride];
      d0 = (d0<0) ? -d0 : d0;

      d1 = v - fp[i+frm_stride];
      d1 = (d1<0) ? -d1 : d1;

      if (d0 <= thre && d1 <= thre) {
	v = (v*2 + fp[i-frm_stride] + fp[i+frm_stride])>>2;
      }

      CLIP_CHECK(v)
      fp[i] = CLIP(v);
    }
    fp += frm_stride;
  }
}

static void
rotate_buf(void)
{
  int i;
  unsigned char *tmp[3];
  tmp[0] = frmbuf[BUF_NUM-1][0];
  tmp[1] = frmbuf[BUF_NUM-1][1];
  tmp[2] = frmbuf[BUF_NUM-1][2];
  for (i = BUF_NUM-2; i >= 0; i--) {
    frmbuf[i+1][0] = frmbuf[i][0];
    frmbuf[i+1][1] = frmbuf[i][1];
    frmbuf[i+1][2] = frmbuf[i][2];
  }
  frmbuf[0][0] = tmp[0];
  frmbuf[0][1] = tmp[1];
  frmbuf[0][2] = tmp[2];
}

int
denoise2(unsigned char *src, unsigned char *dst)
{
  int i, j, offset;
  unsigned char *frm;
  unsigned char *dp;
//  unsigned char buf[Width*Height*3/2];


  frm = frmbuf[0][0] + start_offset[0];
  /* Y top field */
  lowp_filt(Width, Height/2, src, Width*2, frm, stride[0]*2, filt_coef0);
  /* Y bottom field */
  lowp_filt(Width, Height/2, src+Width, Width*2, frm+stride[0], stride[0]*2, filt_coef0);

  /* U */
  offset = Width*Height;
  frm = frmbuf[0][1] + start_offset[1];
  lowp_filt(Width2, Height2, src+offset, Width2, frm, stride[1], filt_coef1);

  /* V */
  offset += offset>>2;
  frm = frmbuf[0][2] + start_offset[2];
  lowp_filt(Width2, Height2, src+offset, Width2, frm, stride[2], filt_coef1);

  if (uninited == 2) {
    rotate_buf();
    uninited = 0;
    return DENO2_NEED_MORE_DATA;
  }

#if 0
  frm = frmbuf[0][0]+start_offset[0];
  dp = buf;
  for (j = 0; j < Height; j++) {
    for (i = 0; i < Width; i++) {
      dp[i] = frm[i];
    }
    dp += Width;
    frm += stride[0];
  }
  frm = frmbuf[0][1]+start_offset[1];
  dp = buf+Width*Height;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++) {
      dp[i] = frm[i];
    }
    dp += Width2;
    frm += stride[1];
  }
  frm = frmbuf[0][2]+start_offset[2];
  dp = buf+Width*Height*5/4;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++) {
      dp[i] = frm[i];
    }
    dp += Width2;
    frm += stride[2];
  }
#endif

  deno_frame(Width, Height,
      frmbuf[2][0]+start_offset[0],
      frmbuf[1][0]+start_offset[0],
      frmbuf[0][0]+start_offset[0], stride[0]);

  deno_frame(Width2, Height2,
      frmbuf[2][1]+start_offset[1],
      frmbuf[1][1]+start_offset[1],
      frmbuf[0][1]+start_offset[1], stride[1]);

  deno_frame(Width2, Height2,
      frmbuf[2][2]+start_offset[2],
      frmbuf[1][2]+start_offset[2],
      frmbuf[0][2]+start_offset[2], stride[2]);

  laplacian(Width, Height, frmbuf[1][0]+start_offset[0], stride[0], dst, Width);
#if 0
  offset = Width * Height;
  laplacian(Width2, Height2, frmbuf[1][1]+start_offset[1], stride[1], dst+offset, Width2);
  offset = Width * Height * 5 / 4;
  laplacian(Width2, Height2, frmbuf[1][2]+start_offset[2], stride[2], dst+offset, Width2);
#endif

  offset = Width * Height;
  frm = frmbuf[1][1] + start_offset[1];
  dp = dst + offset;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++)
      dp[i] = frm[i];
    dp += Width2;
    frm += stride[1];
  }

  offset = Width * Height * 5 / 4;
  frm = frmbuf[1][2] + start_offset[2];
  dp = dst + offset;
  for (j = 0; j < Height2; j++) {
    for (i = 0; i < Width2; i++)
      dp[i] = frm[i];
    dp += Width2;
    frm += stride[2];
  }

  rotate_buf();

  return DENO2_OK;
}

int
denoise2_init(int width, int height, int threshold)
{
  int i;
  double thre, thre2;
  int border;
  int bufsize;

  Width = width;
  Width2 = width>>1;
  Height = height;
  Height2 = height>>1;
  Threshold = threshold;

  thre = threshold;
  thre2 = thre * 4.0;
  if (thre2 > 255.)
    thre2 = 255.;
  make_filt_coef((255.+thre2)/255., filt_catrom, filt_coef0, 2.0);
  make_filt_coef((255.+thre)/255., filt_cubic, filt_coef1, 2.0);

  border = BORDER;
  bufsize = (width + border*2) * (height + border*2);
  for (i = 0; i < BUF_NUM; i++)
  {
    frmbuf[i][0] = (unsigned char*)memalign(16, bufsize);
//    posix_memalign(&frmbuf[i][0], 16, bufsize);
    memset(frmbuf[i][0], 0, bufsize);
  }
  width = Width2;
  height = Height2;
  bufsize = (width + border*2) * (height + border*2);
  for (i = 0; i < BUF_NUM; i++)
  {
    frmbuf[i][1] = (unsigned char*)memalign(16, bufsize);
    frmbuf[i][2] = (unsigned char*)memalign(16, bufsize);
//    posix_memalign(&frmbuf[i][1], 16, bufsize);
//    posix_memalign(&frmbuf[i][2], 16, bufsize);
    memset(frmbuf[i][1], 128, bufsize);
    memset(frmbuf[i][2], 128, bufsize);
  }

  stride[0] = Width + border*2;
  stride[1] = Width2 + border*2;
  stride[2] = Width2 + border*2;
  start_offset[0] = stride[0] * border + border;
  start_offset[1] = stride[1] * border + border;
  start_offset[2] = stride[2] * border + border;

  clip = clip_buf + 384;
  for (i = -384; i < 640; i++)
    clip[i] = (i<0) ? 0 : (i>255) ? 255 : i;

  uninited = 2;

  return DENO2_OK;
}

