/* xwin.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>
#define FCC_YUY2 844715353 /* "YUYV", planes=1, bits=16 */
#define FCC_YV12 842094169 /* "YVU", planes=3, bits=12 */
#define FCC_I420 808596553 /* "YUV", planes=3, bits=12 */
#define FCC_UYVY 808596553 /* "UYVY", planes=1, bits=16 */

#include "xwin.h"

#include "csp.h"
#include "csp_conv.h"

static csp_conv_func_t csp_conv_func;

static int haderror;
static int (*origerrorhandler)(Display *, XErrorEvent *);

static int
shmerrorhandler (Display *d, XErrorEvent *e)
{
  haderror++;
  if (e->error_code == BadAccess) {
    fprintf (stderr, "failed to attach shared memory\n");
    return 0;
  } else
    return (*origerrorhandler)(d, e);
}

static int
create_shm (xwin_t *xw, XShmSegmentInfo *info, int size)
{
  info->shmid = shmget (IPC_PRIVATE, size, IPC_CREAT|0777);
  if (info->shmid == -1) {
    fprintf (stderr, "create_shm(): cannot create shared memory\n");
    return -1;
  } 
  info->shmaddr = (char*) shmat (info->shmid, 0, 0);
  if (info->shmaddr == ((char*)-1)) {
    perror ("shmat");
    shmctl (info->shmid, IPC_RMID, 0);
    return -1;
  } 
  info->readOnly = True;
  XSync (xw->dpy, True);
  haderror = False;
  origerrorhandler = XSetErrorHandler(shmerrorhandler);
  XShmAttach (xw->dpy, info);
  XSync (xw->dpy, True);
  XSetErrorHandler (origerrorhandler);
  if (haderror) {
    if (shmdt (info->shmaddr) == -1)
      perror ("shmdt:");
    if (shmctl (info->shmid, IPC_RMID, 0) == -1)
      perror ("shmctl rmid:");
    return -1;
  }
  return 0;
}

struct XWIN_CSP_COMP {
  unsigned int id;
  int csp;
  int xcsp;
  const char *name;
} xwin_csp_comp[] = {
  { FCC_YUY2, CSP_YUY2, XWIN_CSP_YUY2, "YUY2" },
  { FCC_YUY2, CSP_YUV422, XWIN_CSP_YUV422, "YUV422" },
  { FCC_YV12, CSP_YV12, XWIN_CSP_YV12, "YV12" },
  { FCC_I420, CSP_I420, XWIN_CSP_I420, "I420" },
  { FCC_I420, CSP_YUV420, XWIN_CSP_YUV420, "YUV420" },
  { FCC_UYVY, CSP_UYVY, XWIN_CSP_UYVY, "UYVY" },
  { 0, CSP_RGB24, XWIN_CSP_RGB24, "RGB24" },
  { 0, CSP_BGR24, XWIN_CSP_BGR24, "BGR24" },
  { 0, CSP_ARGB32, XWIN_CSP_ARGB32, "BGRA32" },
  { 0, CSP_BGRA32, XWIN_CSP_BGRA32, "BGRA32" },
  { 0, CSP_RGB565, XWIN_CSP_RGB565, "RGB565" },
  { 0, CSP_RGB555, XWIN_CSP_RGB555, "RGB555" },
};
static int num_xwin_csp_comp = sizeof(xwin_csp_comp)/sizeof(struct XWIN_CSP_COMP);

static int
csp_to_xv_id (int csp)
{
  int i;
  for (i = 0; i < num_xwin_csp_comp; i++)
    if (xwin_csp_comp[i].csp == csp && xwin_csp_comp[i].id != 0)
      return xwin_csp_comp[i].id;
  return 0;
}

static int
xcsp_to_xv_id (int xcsp)
{
  int i;
  for (i = 0; i < num_xwin_csp_comp; i++)
    if (xwin_csp_comp[i].xcsp == xcsp && xwin_csp_comp[i].id != 0)
      return xwin_csp_comp[i].id;
  return 0;
}

static int
csp_to_xcsp (int csp)
{
  int i;
  for (i = 0; i < num_xwin_csp_comp; i++)
    if (xwin_csp_comp[i].csp == csp)
      return xwin_csp_comp[i].xcsp;
  return XWIN_CSP_UNKNOWN;
}

static int
xcsp_to_csp (int xcsp)
{
  int i;
  for (i = 0; i < num_xwin_csp_comp; i++)
    if (xwin_csp_comp[i].xcsp == xcsp)
      return xwin_csp_comp[i].csp;
  return CSP_UNKNOWN;
}

static int xv_check_csp (Display *dpy, XvPortID xport, int xcsp)
{
    XvImageFormatValues * formatValues;
    int formats;
    int i;
    unsigned int id;

    if ((id = xcsp_to_xv_id(xcsp)) == 0) {
      return -1;
    }

    formatValues = XvListImageFormats (dpy, xport, &formats);
    for (i = 0; i < formats; i++)
//        if ((formatValues[i].id == FCC_YV12) &&
//            (! (strcmp (formatValues[i].guid, "YV12")))) {
        if (formatValues[i].id == (int)id) {
            XFree (formatValues);
            return 0;
        }
    XFree (formatValues);
    return -1;
}

static int xv_check_extension (xwin_t *xw, int xcsp)
{
  unsigned int version;
  unsigned int release;
  unsigned int dummy;
  int adaptors;
  int i;
  unsigned long j;
  XvAdaptorInfo * adaptorInfo;

  if ((XvQueryExtension (xw->dpy, &version, &release,
                         &dummy, &dummy, &dummy) != Success) ||
      (version < 2) || ((version == 2) && (release < 2))) {
      fprintf (stderr, "No xv extension\n");
      return -1;
  }

  XvQueryAdaptors (xw->dpy, xw->win, &adaptors,
                   &adaptorInfo);

  for (i = 0; i < adaptors; i++)
    if (adaptorInfo[i].type & XvImageMask)
      for (j = 0; j < adaptorInfo[i].num_ports; j++)
        if ((! (xv_check_csp (xw->dpy, adaptorInfo[i].base_id + j, xcsp))) &&
            (XvGrabPort (xw->dpy, adaptorInfo[i].base_id + j, 0) == Success)) {
          xw->port = adaptorInfo[i].base_id + j;
          xw->disp_csp = xcsp;
          XvFreeAdaptorInfo (adaptorInfo);
          return 0;
        }

  XvFreeAdaptorInfo (adaptorInfo);
  fprintf (stderr, "xv_check_extension: Cannot find xv port\n");
  return -1;
}

int
open_x (xwin_t *xw)
{
  unsigned int mask;
  int major, minor;
  Bool pixmaps;
  XVisualInfo visualTemplate;
  XVisualInfo *XvisualInfoTable;
  XVisualInfo *XvisualInfo;
  int number;
  int i;
  XSetWindowAttributes attr;
  XGCValues gcValues;
  XVisualInfo xvinfo;
  XSizeHints size_hints;
  int width, height;
  int blue_mask = 0;
  int bits_per_pixel = 0;
  int scrn;

  width = xw->disp_width = xw->win_width = xw->image_width;
  height = xw->disp_height = xw->win_height = xw->image_height;

  xw->dpy = XOpenDisplay (xw->dpyname);
  if (!xw->dpy) {
    fprintf (stderr, "open_x(): Can not open display\n");
    return -1;
  }
  scrn = DefaultScreen (xw->dpy);
  xw->gc = DefaultGC (xw->dpy, scrn);

  visualTemplate.class = TrueColor;
  visualTemplate.screen = scrn;
  XvisualInfoTable = XGetVisualInfo (xw->dpy, VisualScreenMask|VisualClassMask,
      &visualTemplate, &number);
  if (XvisualInfoTable == NULL) {
    fprintf (stderr, "open_x(): No truecolor visual\n");
    XCloseDisplay (xw->dpy);
    xw->dpy = NULL;
    return -1;
  }

  XvisualInfo = XvisualInfoTable;
  for (i = 0; i < number; i++)
    if (XvisualInfoTable[i].depth > XvisualInfo->depth)
      XvisualInfo = XvisualInfoTable + i;

  xvinfo = *XvisualInfo;
  XFree (XvisualInfoTable);

  attr.background_pixmap = None;
  attr.backing_store = NotUseful;
  attr.border_pixel = 0;
  attr.event_mask = 0;
  attr.colormap = DefaultColormap (xw->dpy, scrn);
  //attr.colormap = XCreateColormap(xw->dpy, RootWindow(xw->dpy, xvinfo.screen),
  //    xvinfo.visual, AllocNone);

  xw->win = XCreateWindow (xw->dpy, DefaultRootWindow(xw->dpy),
      0, 0, width, height, 0,
      xvinfo.depth, InputOutput, xvinfo.visual,
      (CWBackPixmap | CWBackingStore | CWBorderPixel |
       CWEventMask | CWColormap), &attr);

  mask = KeyPressMask|ExposureMask|StructureNotifyMask;
  XSelectInput (xw->dpy, xw->win, mask);

  xw->gc = XCreateGC (xw->dpy, xw->win, 0, &gcValues);

  xw->xv_image = NULL;
  xw->shm_image = NULL;
  xw->x11_image = NULL;
  if (xw->vo_type == XWIN_VO_XV) {
    if (xv_check_extension (xw, xw->image_csp)) {
      fprintf (stderr, "open_x(): not support xv extension\n");
      //return -1;
      xw->vo_type = XWIN_VO_XSHM;
    }
    else {
      xw->xv_image = XvShmCreateImage (xw->dpy, xw->port,
	  xcsp_to_xv_id(xw->image_csp), NULL,
	          width, height, &xw->shm_info);
      if (xw->xv_image == NULL) {
        fprintf (stderr, "open_x(): Cannot create xv image\n");
	xw->vo_type = XWIN_VO_XSHM;
      } else {
        xw->disp_data_size = xw->xv_image->data_size;
      }
    }
  }

  if (xw->vo_type == XWIN_VO_XSHM) {
    if ((XShmQueryVersion (xw->dpy, &major, &minor, &pixmaps) == 0) ||
        (major < 1) || ((major == 1) && (minor < 1))) {
      fprintf (stderr, "open_x(): No shm extension\n");
      xw->vo_type = XWIN_VO_X11;
    } else {
      xw->shm_image = XShmCreateImage (xw->dpy, xvinfo.visual, xvinfo.depth,
	  ZPixmap, NULL, &xw->shm_info, width, height);
      if (xw->shm_image == NULL) {
        fprintf (stderr, "open_x(): Cannot create x shm image\n");
	xw->vo_type = XWIN_VO_X11;
      } else {
        xw->disp_data_size=xw->shm_image->bytes_per_line*xw->shm_image->height;
      }
    }
  }

  xw->shm_flag = 0;
  if ((xw->vo_type == XWIN_VO_XV) || (xw->vo_type == XWIN_VO_XSHM)) {
    if (create_shm(xw, &xw->shm_info, xw->disp_data_size) < 0) {
      fprintf (stderr, "open_x(): cannot create shared memory\n");
      if (xw->xv_image) {
	XFree (xw->xv_image);
	xw->xv_image = XvCreateImage (xw->dpy, xw->port, FCC_YV12, NULL,
	    width, height);
	if (xw->xv_image == NULL) {
          fprintf (stderr, "open_x(): cannot create xv image.\n");
	  xw->vo_type = XWIN_VO_X11;
	} else {
	  xw->disp_data_size = xw->xv_image->data_size;
	  xw->xv_image->data = (char*) malloc (xw->disp_data_size);
	}
      }
      else if (xw->shm_image) {
        XDestroyImage (xw->shm_image);
	xw->shm_image = NULL;
        xw->vo_type = XWIN_VO_X11;
      }
    } else {
      xw->shm_flag = 1;
      if (xw->vo_type == XWIN_VO_XV) {
        xw->xv_image->data = xw->shm_info.shmaddr;
        xw->xv_image->obdata = (char*) &xw->shm_info;
      } else {
        xw->shm_image->data = xw->shm_info.shmaddr;
        xw->shm_image->obdata = (char*) &xw->shm_info;
        bits_per_pixel = xw->shm_image->bits_per_pixel;
        blue_mask = xw->shm_image->blue_mask;
        xw->disp_bytes_per_line = xw->shm_image->bytes_per_line;
	switch (xvinfo.depth) {
	  case 32:
	    xw->disp_csp = XWIN_CSP_RGB32;
	    break;
	  case 24:
	    if (bits_per_pixel == 32)
	      xw->disp_csp = XWIN_CSP_RGB32;
	    else if (bits_per_pixel == 24)
	      xw->disp_csp = XWIN_CSP_RGB24;
	    break;
	  case 16:
	    xw->disp_csp = XWIN_CSP_RGB565;
	    break;
	  case 15:
	    xw->disp_csp = XWIN_CSP_RGB555;
	    break;
	}
      }
    }
  }

  if (xw->vo_type == XWIN_VO_X11) {
    xw->x11_image = XCreateImage (xw->dpy, xvinfo.visual, xvinfo.depth,
       	ZPixmap, 0, NULL, width, height, 8, 0);
    bits_per_pixel = xw->x11_image->bits_per_pixel;
    blue_mask = xw->x11_image->blue_mask;
    xw->disp_bytes_per_line = xw->x11_image->bytes_per_line;
    xw->disp_data_size = xw->disp_bytes_per_line * height;
    xw->x11_image->data = (char*) malloc (xw->disp_bytes_per_line * height);
    switch (xvinfo.depth) {
      case 32:
        xw->disp_csp = XWIN_CSP_RGB32;
        break;
      case 24:
        if (bits_per_pixel == 32)
          xw->disp_csp = XWIN_CSP_RGB32;
        else if (bits_per_pixel == 24)
          xw->disp_csp = XWIN_CSP_RGB24;
        break;
      case 16:
        xw->disp_csp = XWIN_CSP_RGB565;
        break;
      case 15:
        xw->disp_csp = XWIN_CSP_RGB555;
        break;
    }
  }

  xw->disp_depth = (xvinfo.depth == 24) ? bits_per_pixel : xvinfo.depth;
  xw->disp_mode_rgb = (blue_mask & 0x01);

  xw->wm_protocols = XInternAtom (xw->dpy, "WM_PROTOCOLS", False);
  xw->wm_delete_window = XInternAtom (xw->dpy, "WM_DELETE_WINDOW", False);
  (void) XSetWMProtocols (xw->dpy, xw->win, &xw->wm_delete_window, 1);

  XmbSetWMProperties (xw->dpy, xw->win, xw->title, xw->title,
      NULL, 0, NULL, NULL, NULL);

  if (!xw->xv_image) {
    size_hints.flags = PMinSize | PMaxSize;
    size_hints.min_width = width;
    size_hints.min_height = height;
    size_hints.max_width = width;
    size_hints.max_height = height;
    XSetWMNormalHints (xw->dpy, xw->win, &size_hints);
  }

  XMapWindow (xw->dpy, xw->win);

  return 1;
}

int
close_x (xwin_t *xw)
{
  if (xw->dpy == NULL)
    return 0;

  if (xw->shm_flag) {
    XShmDetach (xw->dpy, &xw->shm_info);
    XSync (xw->dpy, False);
    if (shmdt (xw->shm_info.shmaddr) == -1)
      perror ("shmdt:");
    if (shmctl (xw->shm_info.shmid, IPC_RMID, 0) == -1)
      perror ("shmctl rmid:");
  }

  if (xw->xv_image) {
    XvUngrabPort (xw->dpy, xw->port, 0);
    if (!xw->shm_flag)
      free (xw->xv_image->data);
    XFree (xw->xv_image);
    xw->xv_image = NULL;
  }
  else if (xw->shm_image) {
    XDestroyImage (xw->shm_image);
    xw->shm_image = NULL;
  }
  else if (xw->x11_image) {
    XDestroyImage (xw->x11_image);
    xw->x11_image = NULL;
  }

  XFreeGC (xw->dpy, xw->gc);
  XDestroyWindow (xw->dpy, xw->win);
  XCloseDisplay (xw->dpy);
  xw->dpy = NULL;
  return 1;
}

int
xwin_dispatch_event (xwin_t *xw)
{
  XEvent ev;
  KeySym keysym;
  int ret = -1;

  while (XPending(xw->dpy)) {
    XNextEvent (xw->dpy, &ev);
    switch (ev.type) {
      case KeyPress:
        keysym = XKeycodeToKeysym (xw->dpy, ev.xkey.keycode, 0);
	switch (keysym) {
	  case XK_q: ret = 'q'; return ret; break;
	  case XK_k: ret = 'k'; return ret; break;
	  case XK_j: ret = 'j'; return ret; break;
	  case XK_l: ret = 'l'; return ret; break;
	  case XK_h: ret = 'h'; return ret; break;
	  case XK_semicolon: ret = ';'; return ret; break;
	  case XK_g: ret = 'g'; return ret; break;
	  case XK_m: ret = 'm'; return ret; break;
	  case XK_n: ret = 'n'; return ret; break;
	  case XK_x: ret = 'x'; return ret; break;
	  default: ret = -1; return ret; break;
	}
	break;
      case Expose:
	if (xw->xv_image) {
	  if (xw->shm_flag)
	    XvShmPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
	        0, 0, xw->disp_width, xw->disp_height,
	       	0, 0, xw->win_width, xw->win_height, True);
	  else
	    XvPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
		0, 0, xw->disp_width, xw->disp_height,
	       	0, 0, xw->win_width, xw->win_height);
	}
	else if (xw->shm_image)
          XShmPutImage(xw->dpy, xw->win, xw->gc, xw->shm_image,
	      0, 0, 0, 0, xw->disp_width, xw->disp_height, True);
	ret = 1;
	break;
      case ClientMessage:
        if (ev.xclient.message_type == xw->wm_protocols &&
            ev.xclient.data.l[0] == (int)xw->wm_delete_window)
          return ((int)'q');
        break;
      case ConfigureNotify:
	if (xw->xv_image) {
	  xw->win_width = ev.xconfigure.width;
	  xw->win_height = ev.xconfigure.height;
	}
	ret = 2;
        break;
    }
  }
  return ret;
}

int
xwin_put_pic (xwin_t *xw, void *data)
{
  uint8_t *image_planes[3];

  image_planes[0] = data + xw->image_offset[0];
  image_planes[1] = data + xw->image_offset[1];
  image_planes[2] = data + xw->image_offset[2];

  csp_conv_func(image_planes, xw->disp_planes, xw->image_stride,
      xw->disp_stride, xw->image_width, xw->image_height);

  if (xw->xv_image) {
    if (xw->shm_flag)
      XvShmPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
          0, 0, xw->disp_width, xw->disp_height,
	  0, 0, xw->win_width, xw->win_height, False);
    else
      XvPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
	  0, 0, xw->disp_width, xw->disp_height,
	  0, 0, xw->win_width, xw->win_height);
    return 0;
  }
  else if (xw->shm_image) {
    XShmPutImage (xw->dpy, xw->win, xw->gc, xw->shm_image,
       	0, 0, 0, 0, xw->disp_width, xw->disp_height, False);
  }
  else {
    XPutImage (xw->dpy, xw->win, xw->gc, xw->x11_image,
       	0, 0, 0, 0, xw->disp_width, xw->disp_height);
  }
  XSync (xw->dpy, True);

  return 0;
}

int
xwin_put_image (xwin_t *xw, uint8_t *img_planes[], int img_stride[])
{
  csp_conv_func(img_planes, xw->disp_planes, img_stride,
      xw->disp_stride, xw->image_width, xw->image_height);

  if (xw->xv_image) {
    if (xw->shm_flag)
      XvShmPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
          0, 0, xw->disp_width, xw->disp_height,
	  0, 0, xw->win_width, xw->win_height, False);
    else
      XvPutImage (xw->dpy, xw->port, xw->win, xw->gc, xw->xv_image,
	  0, 0, xw->disp_width, xw->disp_height,
	  0, 0, xw->win_width, xw->win_height);
    return 0;
  }
  else if (xw->shm_image) {
    XShmPutImage (xw->dpy, xw->win, xw->gc, xw->shm_image,
       	0, 0, 0, 0, xw->disp_width, xw->disp_height, False);
  }
  else {
    XPutImage (xw->dpy, xw->win, xw->gc, xw->x11_image,
       	0, 0, 0, 0, xw->disp_width, xw->disp_height);
  }
//  XSync (dpy, True);

  return 0;
}

int
xwin_quit (xwin_t *xw)
{
  if (!xw)
    return 0;
  close_x (xw);
  free(xw);
  return 0;
}

xwin_t *
xwin_init (int width, int height, int in_csp, int vo_type, const char *display_name, const char *title)
{
  char *vo_name;
  xwin_t *xw;
  int src_csp, dst_csp, offset, i;
  unsigned char *out_buf;

  xw = (xwin_t*)malloc(sizeof(xwin_t));
  if (!xw) {
    fprintf(stderr, "xwin_init: memory allocate error.\n");
    return NULL;
  }
  memset(xw, 0, sizeof(xwin_t));

  xw->title = title;
  xw->dpyname = display_name;
  xw->image_width = width;
  xw->image_height = height;
  xw->vo_type = vo_type;

  xw->image_csp = csp_to_xcsp(in_csp);
  if (xw->image_csp == XWIN_CSP_UNKNOWN) {
    fprintf(stderr, "xwin_init: xwin doesn't support csp \"%s\".\n",
       	csp_to_name(in_csp));
    free(xw);
    return NULL;
  }

  if (csp_to_xv_id(in_csp) == 0) {
    if (vo_type == XWIN_VO_XV) {
      fprintf(stderr, "xwin_init: XV extension doesn't support RGB24 output, force XSHM.\n");
      xw->vo_type = XWIN_VO_XSHM;
    }
  }

  if (open_x (xw) < 0) {
    xwin_quit(xw);
    return NULL;
  }

  switch (xw->vo_type) {
    case XWIN_VO_XV: vo_name = "XV"; break;
    case XWIN_VO_X11: vo_name = "X11"; break;
    case XWIN_VO_XSHM: vo_name = "XSHM"; break;
    default: vo_name = "UNKNOWN"; break;
  }
  xw->vo_name = vo_name;

  src_csp = dst_csp = CSP_UNKNOWN;

  src_csp = xcsp_to_csp(xw->image_csp);
  dst_csp = xcsp_to_csp(xw->disp_csp);

  if (xw->xv_image)
    out_buf = xw->xv_image->data;
  else if (xw->shm_image)
    out_buf = xw->shm_image->data;
  else
    out_buf = xw->x11_image->data;

  switch (xw->disp_csp) {
    case XWIN_CSP_RGB24:
#ifdef WORDS_BIGENDIAN
      dst_csp = CSP_RGB24;
#else
      dst_csp = CSP_BGR24;
#endif
      break;
    case XWIN_CSP_RGB32:
#ifdef WORDS_BIGENDIAN
      dst_csp = CSP_ARGB32;
#else
      dst_csp = CSP_BGRA32;
#endif
      break;
    case XWIN_CSP_YUV422:
      dst_csp = CSP_YUY2;
  }

  csp_calc_buf_size(src_csp, width, height, 0, 0, xw->image_size, xw->image_offset, xw->image_stride);
  csp_calc_buf_size(dst_csp, width, height, 0, 0, xw->disp_size, xw->disp_offset, xw->disp_stride);

  offset = 0;
  for (i = 0; i < 3; i++) {
    xw->image_offset[i] += offset;
    offset += xw->image_size[i];
  }

  offset = 0;
  for (i = 0; i < 3; i++) {
    xw->disp_offset[i] += offset;
    xw->disp_planes[i] = out_buf + xw->disp_offset[i];
    offset += xw->disp_size[i];
  }

  csp_conv_func = csp_conv_get_convert_func(src_csp, dst_csp);
  if (!csp_conv_func)
    fprintf(stderr, "xwin_init: csp_conv_func is NULL, %d, %d, %d, %d\n", src_csp, dst_csp, xw->image_csp, xw->disp_csp);

  fprintf(stdout, "xwin: vo type \"%s\".\n", xw->vo_name);

  return xw;
}

