/* v4l.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <linux/videodev.h>

#include "v4l.h"

#define DEFAULT_DEVICE_NAME "/dev/video0"

typedef struct {
  int csp;
  int palette;
  const char *name;
} CSP_PALETTE;

static CSP_PALETTE csp_palette[] = {
//  { V4L_CSP_RGB565, VIDEO_PALETTE_RGB565, "RGB565" },
//  { V4L_CSP_RGB555, VIDEO_PALETTE_RGB555, "RGB555" },
  { V4L_CSP_RGB24, VIDEO_PALETTE_RGB24, "RGB24" },
  { V4L_CSP_ARGB32, VIDEO_PALETTE_RGB32, "RGB32" },
//  { V4L_CSP_YUV422PK, VIDEO_PALETTE_YUV422, "YUV422PK" },
  { V4L_CSP_YUYV, VIDEO_PALETTE_YUYV, "YUYV" },
  { V4L_CSP_YUY2, VIDEO_PALETTE_YUYV, "YUY2" },
  { V4L_CSP_UYVY, VIDEO_PALETTE_UYVY, "UYVY" },
  { V4L_CSP_YUV420, VIDEO_PALETTE_YUV420P, "YUV420" },
  { V4L_CSP_YUV422, VIDEO_PALETTE_YUV422P, "YUV422" },
};
static int csp_palette_num = sizeof(csp_palette)/sizeof(CSP_PALETTE);

static int
csp_to_palette(int csp)
{
  int i;
  for (i = 0; i < csp_palette_num; i++)
    if (csp_palette[i].csp == csp)
      break;
  if (i == csp_palette_num)
    return -1;
  return csp_palette[i].palette;
}

static int
palette_to_csp(int palette)
{
  int i;
  for (i = 0; i < csp_palette_num; i++)
    if (csp_palette[i].palette == palette)
      break;
  if (i == csp_palette_num)
    return -1;
  return csp_palette[i].csp;
}

#if 0
static const char*
csp_to_name(int csp)
{
  int i;
  for (i = 0; i < csp_palette_num; i++)
    if (csp_palette[i].csp == csp)
      break;
  if (i == csp_palette_num)
    return NULL;
  return csp_palette[i].name;
}
#endif

static const char*
palette_to_name(int palette)
{
  int i;
  for (i = 0; i < csp_palette_num; i++)
    if (csp_palette[i].palette == palette)
      break;
  if (i == csp_palette_num)
    return NULL;
  return csp_palette[i].name;
}

typedef struct {
  int norm;
  int mode;
  const char *name;
} NORM_TABLE;

static NORM_TABLE norm_table[] = {
  { V4L_NORM_PAL, VIDEO_MODE_PAL, "PAL" },
  { V4L_NORM_NTSC, VIDEO_MODE_NTSC, "NTSC" },
  { V4L_NORM_SECAM, VIDEO_MODE_SECAM, "SECAM" },
  { V4L_NORM_AUTO, VIDEO_MODE_AUTO, "AUTO" },
};
static int norm_table_num = sizeof(norm_table)/sizeof(NORM_TABLE);

const char*
norm_to_str(int norm)
{
  int i;

  for (i = 0; i < norm_table_num; i++) {
    if (norm_table[i].norm == norm)
      break;
  }
  if (i == norm_table_num)
    return NULL;
  return norm_table[i].name;
}

static int
norm_to_mode(int norm)
{
  int i;

  for (i = 0; i < norm_table_num; i++) {
    if (norm_table[i].norm == norm)
      break;
  }
  if (i == norm_table_num)
    return -1;
  return norm_table[i].mode;
}

static int
mode_to_norm(int mode)
{
  int i;

  for (i = 0; i < norm_table_num; i++) {
    if (norm_table[i].mode == mode)
      break;
  }
  if (i == norm_table_num)
    return -1;
  return norm_table[i].norm;
}

int
v4l_capture_cap(v4l_t *v4l)
{
  if (v4l->fd == -1)
    return V4L_FAIL;
  if (!(v4l->type & VID_TYPE_CAPTURE)) {
    fprintf (stderr, "v4l_grab_set_mmap: device cannot capture.\n");
    return V4L_FAIL;
  }
  return V4L_OK;
}

int
v4l_set_tuner (v4l_t *v4l, int tuner)
{
  struct video_tuner vid_tun;
  int freq;

  if (v4l->fd == -1)
    return V4L_FAIL;
  
  if (!(v4l->channel_flags & VIDEO_VC_TUNER) || !v4l->channel_tuners){
    v4l->tuner = -1;
    return V4L_FAIL;
  }

  if (tuner < 0)
    tuner = 0;
  else if (tuner >= v4l->channel_tuners)
    tuner = v4l->channel_tuners-1;

  vid_tun.tuner = tuner;
  if (ioctl (v4l->fd, VIDIOCGTUNER, &vid_tun) < 0) {
    perror ("v4l_set_tuner: VIDIOCGTUNER error");
    return V4L_FAIL;
  }
  vid_tun.tuner = tuner;
  vid_tun.mode = norm_to_mode(v4l->norm);
  vid_tun.signal = 65535;
  if (ioctl (v4l->fd, VIDIOCSTUNER, &vid_tun) < 0) {
    perror ("v4l_set_tuner: VIDIOCSTUNER error");
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGTUNER, &vid_tun) < 0) {
    perror ("v4l_set_tuner: VIDIOCGTUNER error");
    return V4L_FAIL;
  }
  v4l->tuner = vid_tun.tuner;
  strncpy(v4l->tuner_name, vid_tun.name, 32);
  v4l->rangelow = vid_tun.rangelow;
  v4l->rangehigh = vid_tun.rangehigh;
  v4l->tuner_flags = vid_tun.flags;
  v4l->tuner_mode = vid_tun.mode;
  v4l->tuner_signal = vid_tun.signal;

  if (v4l->tuner_mode != norm_to_mode(v4l->norm)) {
    fprintf(stderr, "v4l_set_tuner: norm failed.\n");
  }

  if (ioctl (v4l->fd, VIDIOCGFREQ, &freq) < 0) {
    perror ("v4l_get_freq: VIDIOCGFREQ error");
    return V4L_FAIL;
  }
  v4l->freq = freq;

  return V4L_OK;
}

int
v4l_get_tuner (v4l_t *v4l)
{
  if (v4l->fd == -1 ||
      !(v4l->channel_flags & VIDEO_VC_TUNER) || !v4l->channel_tuners)
    return V4L_FAIL;
  return v4l->tuner;
}

int
v4l_set_source_channel (v4l_t *v4l, int source_channel)
{
  struct video_channel vid_chan;
  int tuner;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (source_channel < 0)
    source_channel = 0;
  if (source_channel >= v4l->channels)
    source_channel = v4l->channels-1;

  vid_chan.channel = source_channel;
  if (ioctl (v4l->fd, VIDIOCGCHAN, &vid_chan) < 0) {
    perror ("v4l_set_source_channel: VIDIOCGCHAN error");
    return V4L_FAIL;
  }
  if (v4l->norm == V4L_NORM_UNKNOWN)
    v4l->norm = mode_to_norm(vid_chan.norm);
  else
    vid_chan.norm = norm_to_mode(v4l->norm);
  vid_chan.channel = source_channel;
  if (ioctl (v4l->fd, VIDIOCSCHAN, &vid_chan) < 0) {
    perror ("v4l_set_source_channel(): VIDIOCSCHAN error");
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGCHAN, &vid_chan) < 0) {
    perror ("v4l_set_source_channel: VIDIOCGCHAN error");
    return V4L_FAIL;
  }
  v4l->source_channel = vid_chan.channel;
  strncpy(v4l->channel_name, vid_chan.name, 32);
  v4l->channel_tuners = vid_chan.tuners;
  v4l->channel_flags = vid_chan.flags;
  v4l->channel_type = vid_chan.type;
  v4l->channel_norm = vid_chan.norm;

  if (v4l->channel_norm != norm_to_mode(v4l->norm)) {
    fprintf(stderr, "v4l_set_source_channel: norm failed.\n");
    v4l->norm = mode_to_norm(v4l->channel_norm);
  }

  tuner = v4l->tuner;
  v4l->tuner = -1;
  if (v4l->channel_flags & VIDEO_VC_TUNER) {
    if (tuner == -1)
      tuner = 0;
    v4l_set_tuner(v4l, tuner);
  }

  return V4L_OK;
}

int
v4l_get_source_channel (v4l_t *v4l)
{
  if (v4l->fd == -1)
    return V4L_FAIL;
  return v4l->source_channel;
}

int
v4l_get_norm (v4l_t *v4l)
{
  if (v4l->fd == -1)
    return V4L_FAIL;
  return v4l->norm;
}

int
v4l_set_norm (v4l_t *v4l, int n)
{
  int mode;

  if (v4l->fd == -1)
    return V4L_FAIL;

  mode = norm_to_mode(n);
  if (mode == V4L_FAIL) {
    fprintf(stderr, "v4l_set_norm: ERROR unknown norm.\n");
    return V4L_FAIL;
  }

  v4l->norm = mode_to_norm(mode);
  if (mode != v4l->channel_norm ||
     (v4l->channel_flags&VIDEO_VC_TUNER && v4l->tuner_mode!=mode)){
    if (v4l_set_source_channel(v4l, v4l->source_channel) == V4L_FAIL)
      return V4L_FAIL;
  }

  return V4L_OK;
}

int
v4l_set_freq (v4l_t *v4l, int freq)
{
  int f;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (!(v4l->channel_flags & VIDEO_VC_TUNER)) {
    return V4L_FAIL;
  }

  f = freq;
  if (ioctl (v4l->fd, VIDIOCSFREQ, &f) < 0) {
    perror ("v4l_set_freq() VIDIOCSFREQ error");
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGFREQ, &f) < 0) {
    perror ("v4l_set_freq: VIDIOCGFREQ error");
    return V4L_FAIL;
  }
  if (f != freq)
    return V4L_FAIL;
  v4l->freq = freq;
  return V4L_OK;
}

int
v4l_get_freq (v4l_t *v4l)
{
  int freq;

  if (v4l->fd == -1)
    return V4L_FAIL;
  if (!(v4l->channel_flags & VIDEO_VC_TUNER)) {
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGFREQ, &freq) < 0) {
    perror ("v4l_get_freq: VIDIOCGFREQ error");
    return V4L_FAIL;
  }
  return freq;
}

int
v4l_set_csp(v4l_t *v4l, int csp)
{
  struct video_picture vid_pic;
  int palette;

  if (v4l->fd == -1)
    return V4L_FAIL;

  palette = csp_to_palette(csp);
  if (palette == -1) {
    fprintf(stderr, "v4l_set_csp: unsupported csp.\n");
    return V4L_FAIL;
  }

  if (ioctl(v4l->fd, VIDIOCGPICT, &vid_pic) < 0) {
    perror ("v4l_set_csp: VIDIOCGPICT error");
    return V4L_FAIL;
  }
  if (vid_pic.palette == palette)
    return V4L_OK;

  vid_pic.palette = palette;
  if (ioctl(v4l->fd, VIDIOCSPICT, &vid_pic) < 0) {
    perror ("v4l_set_csp: VIDIOCSPICT error");
    fprintf(stderr, "v4l_set_csp: Device doesn't support %s capture.\n",
	palette_to_name(palette));
    return V4L_FAIL;
  }
  if (ioctl(v4l->fd, VIDIOCGPICT, &vid_pic) < 0) {
    perror ("v4l_set_csp: VIDIOCGPICT error");
    return V4L_FAIL;
  }
  if (vid_pic.palette != palette) {
    fprintf(stderr, "v4l_set_csp: unsupported csp.\n");
    return V4L_FAIL;
  }
  v4l->palette = palette_to_csp(palette);
  v4l->csp = csp;

  return V4L_OK;
}

int
v4l_get_csp(v4l_t *v4l)
{
  if (v4l->fd == -1)
    return V4L_FAIL;
  return v4l->csp;
}

int
v4l_set_audio(v4l_t *v4l, int audio, int on)
{
  struct video_audio vid_aud;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (v4l->audios <= 0 || !(v4l->channel_flags & VIDEO_VC_AUDIO)) {
    v4l->audio = -1;
    return V4L_FAIL;
  }

  if (audio < 0)
    audio = 0;
  else if (audio >= v4l->audios)
    audio = v4l->audios - 1;

  vid_aud.audio = audio;
  if (ioctl (v4l->fd, VIDIOCGAUDIO, &vid_aud) < 0) {
    perror ("v4l_set_audio: VIDIOCGAUDIO error");
    return V4L_FAIL;
  }
  if (on) {
    vid_aud.flags &= ~VIDEO_AUDIO_MUTE;
  } else {
    if (vid_aud.flags & VIDEO_AUDIO_MUTABLE)
      vid_aud.flags |= VIDEO_AUDIO_MUTE;
  }
  if (ioctl(v4l->fd, VIDIOCSAUDIO, &vid_aud) < 0) {
    perror ("v4l_set_audio: VIDIOCSAUDIO error");
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGAUDIO, &vid_aud) < 0) {
    perror ("v4l_set_audio: VIDIOCGAUDIO error");
    return V4L_FAIL;
  }
  v4l->audio = vid_aud.audio;
  v4l->audio_volume = vid_aud.volume;
  v4l->audio_bass = vid_aud.bass;
  v4l->audio_treble = vid_aud.treble;
  v4l->audio_flags = vid_aud.flags;
  strncpy(v4l->audio_name, vid_aud.name, 16);
  v4l->audio_mode = vid_aud.mode;
  v4l->audio_balance = vid_aud.balance;
  v4l->audio_step = vid_aud.step;
  return V4L_OK;
}

int
v4l_set_window_size (v4l_t *v4l, int w, int h)
{
  struct video_window vid_win;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (ioctl(v4l->fd, VIDIOCGWIN, &vid_win) < 0) {
    perror ("v4l_window_size: VIDIOCGWIN error");
    return V4L_FAIL;
  }

  if (w > v4l->maxwidth)
    w = v4l->maxwidth;
  else if (w < v4l->minwidth)
    w = v4l->minwidth;

  if (h > v4l->maxheight)
    h = v4l->maxheight;
  else if (h < v4l->minheight)
    h = v4l->minheight;

  if (w > 0)
    vid_win.width = w;
  if (h > 0)
    vid_win.height = h;

  vid_win.x         = 0;
  vid_win.y         = 0;
//  vid_win.flags     = VIDEO_WINDOW_INTERLACE;
  vid_win.flags     = 0;
  vid_win.clips     = NULL;
  vid_win.clipcount = 0;
  if (ioctl(v4l->fd, VIDIOCSWIN, &vid_win) < 0) {
    perror ("v4l_set_window_size: VIDIOCSWIN error");
    return V4L_FAIL;
  }
  if (ioctl (v4l->fd, VIDIOCGWIN, &vid_win) < 0) {
    perror ("v4l_set_window_size: VIDIOCGWIN error");
    return V4L_FAIL;
  }
  v4l->x = vid_win.x;
  v4l->y = vid_win.y;
  v4l->width = vid_win.width;
  v4l->height = vid_win.height;
  v4l->chromakey = vid_win.chromakey;
  v4l->win_flags = vid_win.flags;

  return V4L_OK;
}

int
v4l_get_window_size (v4l_t *v4l, int *width, int *height)
{
  if (v4l->fd == -1) {
    *width = *height = 0;
    return V4L_FAIL;
  }
  *width  = v4l->width;
  *height = v4l->height;
  return V4L_OK;
}

int
v4l_color(v4l_t *v4l, int brightness, int hue, int colour, int contrast, int whiteness)
{
  struct video_picture vid_pic;
  if (v4l->fd == -1)
    return V4L_FAIL;

  if (ioctl(v4l->fd, VIDIOCGPICT, &vid_pic) < 0) {
    perror ("v4l_set_color: VIDIOCGPICT error");
    return V4L_FAIL;
  }
  if (brightness >= 0)
    vid_pic.brightness = brightness;
  if (hue >= 0)
    vid_pic.hue        = hue;
  if (colour >= 0)
    vid_pic.colour     = colour;
  if (contrast >= 0)
    vid_pic.contrast   = contrast;
  if (whiteness >= 0)
    vid_pic.whiteness  = whiteness;
  if (ioctl(v4l->fd, VIDIOCSPICT, &vid_pic) < 0) {
    perror ("v4l_set_color: VIDIOCSPICT error");
    return V4L_FAIL;
  }
  if (ioctl(v4l->fd, VIDIOCGPICT, &vid_pic) < 0) {
    perror ("v4l_set_color: VIDIOCGPICT error");
    return V4L_FAIL;
  }
  v4l->brightness = vid_pic.brightness;
  v4l->hue        = vid_pic.hue;
  v4l->colour     = vid_pic.colour;
  v4l->contrast   = vid_pic.contrast;
  v4l->whiteness  = vid_pic.whiteness;
  v4l->depth      = vid_pic.depth;
  v4l->palette    = palette_to_csp(vid_pic.palette);

  if (brightness >= 0 && vid_pic.brightness != brightness)
    return V4L_FAIL;
  else if (hue >= 0 && vid_pic.hue != hue)
    return V4L_FAIL;
  else if (colour >= 0 && vid_pic.colour != colour)
    return V4L_FAIL;
  else if (contrast >= 0 && vid_pic.contrast != contrast)
    return V4L_FAIL;
  else if (whiteness >= 0 && vid_pic.whiteness  != whiteness)
    return V4L_FAIL;

  return V4L_OK;
}

int
v4l_set_color (v4l_t *v4l, float percent, int element)
{
  float max;
  int val;
  int brightness, hue, colour, contrast, whiteness;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (percent < 0)
    return V4L_OK;

  max = (float)65535;
  if (percent > 100.0)
    percent = 100.0;
  val = percent / 100.0 * max;
  if (val < 0)
    val = 0;
  else if (val > 65535)
    val = 65535;

  brightness = hue = colour = contrast = whiteness = -1;
  switch (element) {
    case V4L_BRIGHTNESS:
      brightness = val;
      break;
    case V4L_HUE:
      hue = val;
      break;
    case V4L_COLOUR:
      colour = val;
      break;
    case V4L_CONTRAST:
      contrast = val;
      break;
    case V4L_WHITENESS:
      whiteness = val;
      break;
  }

  if (v4l_color(v4l, brightness, hue, colour, contrast, whiteness) == V4L_FAIL)
      return V4L_FAIL;

  return V4L_OK;
}

int
v4l_set_brightness(v4l_t *v4l, float percent)
{
  return v4l_set_color(v4l, percent, V4L_BRIGHTNESS);
}

int
v4l_set_hue(v4l_t *v4l, float percent)
{
  return v4l_set_color(v4l, percent, V4L_HUE);
}

int
v4l_set_colour(v4l_t *v4l, float percent)
{
  return v4l_set_color(v4l, percent, V4L_COLOUR);
}

int
v4l_set_contrast(v4l_t *v4l, float percent)
{
  return v4l_set_color(v4l, percent, V4L_CONTRAST);
}

int
v4l_set_whiteness(v4l_t *v4l, float percent)
{
  return v4l_set_color(v4l, percent, V4L_WHITENESS);
}

float
v4l_get_color (v4l_t *v4l, int element)
{
  float max;
  float percent;
  float val = 0;

  if (v4l->fd == -1)
    return V4L_FAIL;

  switch (element) {
    case V4L_BRIGHTNESS:
      val = (float)v4l->brightness;
      break;
    case V4L_HUE:
      val = (float)v4l->hue;
      break;
    case V4L_COLOUR:
      val = (float)v4l->colour;
      break;
    case V4L_CONTRAST:
      val = (float)v4l->contrast;
      break;
    case V4L_WHITENESS:
      val = (float)v4l->whiteness;
      break;
  }
  max = (float)65535;
  percent = val / max * 100.0;
  return percent;
}

float
v4l_get_brightness(v4l_t *v4l)
{
  return v4l_get_color(v4l, V4L_BRIGHTNESS);
}

float
v4l_get_hue(v4l_t *v4l)
{
  return v4l_get_color(v4l, V4L_HUE);
}

float
v4l_get_colour(v4l_t *v4l)
{
  return v4l_get_color(v4l, V4L_COLOUR);
}

float
v4l_get_contrast(v4l_t *v4l)
{
  return v4l_get_color(v4l, V4L_CONTRAST);
}

float
v4l_get_whiteness(v4l_t *v4l)
{
  return v4l_get_color(v4l, V4L_WHITENESS);
}

#if 0
int
v4l_overlay_on (v4l_t *v4l)
{
  int one = 1;

  if (v4l->fd == -1)
    return V4L_FAIL;
  if ((v4l->type & VID_TYPE_CAPTURE) &&
      (v4l->type & VID_TYPE_OVERLAY)) {
    if (ioctl (v4l->fd, VIDIOCCAPTURE, &one) < 0) {
      perror ("v4l_overlay_on: VIDIOCCAPTURE error");
      return V4L_FAIL;
    }
  } else {
    return V4L_FAIL;
  }
  return V4L_OK;
}

int
v4l_overlay_off (v4l_t *v4l)
{
  int zero = 0;

  if (v4l->fd == -1)
    return V4L_FAIL;
  if ((v4l->type & VID_TYPE_CAPTURE) &&
      (v4l->type & VID_TYPE_OVERLAY)) {
    if (ioctl(v4l->fd, VIDIOCCAPTURE, &zero) < 0) {
      perror ("v4l_overlay_off: VIDIOCCAPTURE error");
      return V4L_FAIL;
    }
  } else {
    return V4L_FAIL;
  }
  return V4L_OK;
}
#endif

static int
v4l_grab_set_mmap (v4l_t *v4l)
{
  struct video_mbuf vid_mbuf;
  int i;

  if (v4l->fd == -1)
    return V4L_FAIL;

  if (v4l->offsets) {
    fprintf(stderr, "v4l_grab_set_mmap: offsets pointer remains.\n");
    free(v4l->offsets);
    v4l->offsets = NULL;
  }
  if (v4l->map_data) {
    fprintf(stderr, "v4l_grab_set_mmap: map_data pointer remains.\n");
    munmap(v4l->map_data, v4l->map_size);
    v4l->map_data = NULL;
  }

  if (ioctl(v4l->fd, VIDIOCGMBUF, &vid_mbuf) < 0) {
    perror ("v4l_grab_set_mmap: VIDIOCGMBUF error");
    return V4L_FAIL;
  }
  v4l->frames = vid_mbuf.frames;
  v4l->offsets = (int*) calloc (sizeof(int*), vid_mbuf.frames);
  for (i = 0; i < vid_mbuf.frames; i++)
    v4l->offsets[i] = vid_mbuf.offsets[i];
  v4l->map_size = vid_mbuf.size;
  v4l->map_data = (char*) mmap (0, v4l->map_size,
      PROT_READ|PROT_WRITE, MAP_SHARED, v4l->fd, 0);
  if (v4l->map_data == MAP_FAILED) {
    perror ("v4l_grab_set_mmap: mmap error");
    if (v4l->offsets) {
      free(v4l->offsets);
      v4l->offsets = NULL;
    }
    v4l->map_data = NULL;
    return V4L_FAIL;
  }
  return V4L_OK;
}

int
v4l_grab_pic_start (v4l_t *v4l)
{
  int i;

//  v4l_overlay_on ();
  if (v4l->fd == -1)
    return V4L_FAIL;

  if (!(v4l->type & VID_TYPE_CAPTURE)) {
    fprintf (stderr, "v4l_grab_set_mmap: device cannot capture.\n");
    return V4L_FAIL;
  }

  if (v4l->map_data != NULL || v4l->offsets != NULL) {
    fprintf(stderr, "v4l_grab_pic_start: map_data pointer remains.\n");
    return V4L_FAIL;
  }

  if (v4l_grab_set_mmap(v4l) == V4L_FAIL)
    return V4L_FAIL;

  v4l->vid_mmap.width  = v4l->width;
  v4l->vid_mmap.height = v4l->height;
  v4l->vid_mmap.format = csp_to_palette(v4l->palette);
  for (i = 0; i < v4l->frames; i++) {
    v4l->vid_mmap.frame = i;
    if (ioctl(v4l->fd, VIDIOCMCAPTURE, &v4l->vid_mmap) < 0) {
      perror("v4l_grab_pic_start: VIDIOCMCAPTURE error");
      if (v4l->map_data) {
        munmap (v4l->map_data, v4l->map_size);
        v4l->map_data = NULL;
      }
      if (v4l->offsets) {
        free (v4l->offsets);
        v4l->offsets = NULL;
      }
      return V4L_FAIL;
    }
  }
  v4l->vid_mmap.frame = 0;
  return V4L_OK;
}

char*
v4l_grab_pic (v4l_t *v4l)
{
  if (ioctl(v4l->fd, VIDIOCMCAPTURE, &v4l->vid_mmap) < 0) {
//    perror("v4l_grab_pic: VIDIOCMCAPTURE error");
//    return NULL;
  }
  if (++v4l->vid_mmap.frame >= (uint32_t)v4l->frames)
    v4l->vid_mmap.frame = 0;

  if (ioctl(v4l->fd, VIDIOCSYNC, &v4l->vid_mmap.frame) < 0) {
    /*
    perror ("v4l_grab_pic() VIDIOCSYNC error");
    fprintf (stderr, "v4l_grab_pic: cur_frame %d, frames %d\n",
	cur_frame, frames);
    */
    return NULL;
  }
  return &v4l->map_data[v4l->offsets[v4l->vid_mmap.frame]];
}

char*
v4l_grab_pic_sync (v4l_t *v4l)
{
  if (ioctl(v4l->fd, VIDIOCSYNC, &v4l->vid_mmap.frame) < 0) {
    /*
    perror ("v4l_grab_pic_sync() VIDIOCSYNC error");
    fprintf (stderr, "v4l_grab_pic_sync(): cur_frame %d, frames %d\n",
	cur_frame, frames);
    */
    return NULL;
  }
  return &v4l->map_data[v4l->offsets[v4l->vid_mmap.frame]];
}

int
v4l_grab_pic_capture (v4l_t *v4l)
{
  if (ioctl(v4l->fd, VIDIOCMCAPTURE, &v4l->vid_mmap) < 0) {
    perror("v4l_grab_pic_capture: VIDIOCMCAPTURE error");
    return V4L_FAIL;
  }
  if (++v4l->vid_mmap.frame >= (uint32_t)v4l->frames)
    v4l->vid_mmap.frame = 0;
  return V4L_OK;
}

int
v4l_grab_pic_stop (v4l_t *v4l)
{
  int i;

  if (v4l->map_data == NULL && v4l->offsets == NULL)
    return V4L_OK;

  for (i = 0; i < v4l->frames; i++) {
    if (ioctl(v4l->fd, VIDIOCSYNC, &v4l->vid_mmap.frame) < 0) {
      perror ("v4l_grab_pic_stop: VIDIOCSYNC error");
//      fprintf (stderr, "v4l_grab_pic_stop: cur_frame %d, frames %d\n",
//	cur_frame, frames);
      break;
    }
    if (++v4l->vid_mmap.frame >= (uint32_t)v4l->frames)
      v4l->vid_mmap.frame = 0;
  }

  if (v4l->map_data) {
    munmap (v4l->map_data, v4l->map_size);
    v4l->map_data = NULL;
  }
  if (v4l->offsets) {
    free (v4l->offsets);
    v4l->offsets = NULL;
  }
//  v4l_overlay_off ();
  return V4L_OK;
}

int
v4l_open (v4l_t *v4l)
{
  const char *name;
  struct video_capability vid_cap;
  int brightness, hue, colour, contrast, whiteness;
  int width, height;

  if (v4l->device_name)
    name = v4l->device_name;
  else
    name = DEFAULT_DEVICE_NAME;

  if ((v4l->fd = open(name, O_RDWR)) < 0) {
    fprintf (stderr, "v4l_open: %s open error", name);
    perror (" ");
    v4l->fd = -1;
    return V4L_FAIL;
  }

  if (ioctl (v4l->fd, VIDIOCGCAP, &vid_cap) < 0) {
    perror ("v4l_open: VIDIOCGCAP error");
    v4l_close (v4l);
    return V4L_FAIL;
  }

  v4l->type = vid_cap.type;
  v4l->channels = vid_cap.channels;
  v4l->audios = vid_cap.audios;
  v4l->maxwidth = vid_cap.maxwidth;
  v4l->maxheight = vid_cap.maxheight;
  v4l->minwidth = vid_cap.minwidth;
  v4l->minheight = vid_cap.minheight;

  v4l_set_source_channel(v4l, v4l->source_channel);

  brightness = hue = colour = contrast = whiteness = -1;
  v4l_color(v4l, brightness, hue, colour, contrast, whiteness);

  width = height = 0;
  v4l_set_window_size(v4l, width, height);

  v4l_set_audio(v4l, v4l->audio, 1);

  return V4L_OK;
}

int
v4l_close (v4l_t *v4l)
{
  if (v4l->fd == -1)
    return V4L_OK;

  if (v4l->map_data) {
    v4l_grab_pic_stop(v4l);
    v4l->map_data = NULL;
  }
  if (v4l->offsets) {
    v4l->offsets = NULL;
  }

  v4l_set_audio (v4l, v4l->audio, 0);

  if (v4l->fd != -1 && (close(v4l->fd)) < 0) {
    perror ("v4l_close: Device close error");
    v4l->fd = -1;
    return V4L_FAIL;
  }
  v4l->fd = -1;
  v4l->device_name = NULL;

  return V4L_OK;
}

v4l_t*
v4l_init (const char *device_name, int source_channel, int tuner, int norm, int audio)
{
  v4l_t *v4l;

  v4l = (v4l_t*)malloc(sizeof(v4l_t));
  if (!v4l) {
    fprintf(stderr, "v4l_init: memory error.\n");
    return NULL;
  }
  memset(v4l, 0, sizeof(v4l_t));
  v4l->fd = -1;

  if (device_name)
    v4l->device_name = strdup(device_name);
  else
    v4l->device_name = NULL;

  v4l->source_channel = source_channel;
  v4l->tuner = tuner;
  v4l->norm = norm;
  v4l->audio = audio;
  if (v4l_open(v4l) == V4L_FAIL) {
    v4l_quit(v4l);
    return NULL;
  }

  return v4l;
}

int
v4l_quit (v4l_t *v4l)
{
  int ret;
  ret = v4l_close (v4l);
  if (v4l->device_name)
    free((char*)v4l->device_name);
  free(v4l);
  return ret;
}

void
v4l_print_param(v4l_t *v4l)
{
  const char *devname = v4l->device_name;
  if (!v4l->device_name)
    devname = DEFAULT_DEVICE_NAME;
  fprintf(stdout, "VIDEO DEVICE:            %s\n", devname);
  fprintf(stdout, "video device has %d source, %d audios.\n", v4l->channels, v4l->audios);
  fprintf(stdout, "VIDEO DEVICE SOURCE:     %d [%s]\n", v4l->source_channel, v4l->channel_name);
  if (v4l->channel_flags & VIDEO_VC_TUNER) {
    fprintf(stdout, "source %d has %d tuners.\n", v4l->source_channel, v4l->channel_tuners);
    fprintf(stdout, "VIDEO DEVICE TUNER:      %d [%s]\n", v4l->tuner, v4l->tuner_name);
  } else {
    fprintf(stdout, "source %d has no tuners.\n", v4l->source_channel);
  }
  if (v4l->channel_flags & VIDEO_VC_AUDIO) {
    fprintf(stdout, "source %d has audio.\n", v4l->source_channel);
    fprintf(stdout, "VIDEO DEVICE AUDIO:      %d [%s]\n", v4l->audio, v4l->audio_name);
  } else {
    fprintf(stdout, "source %d has no audio.\n", v4l->source_channel);
  }
  fprintf(stdout, "VIDEO DEVICE NORM:       %s\n", norm_to_str(v4l->norm));
  fprintf(stdout, "VIDEO DEVICE BRIGHTNESS: %d%%\n", (int)v4l_get_brightness(v4l));
  fprintf(stdout, "VIDEO DEVICE HUE:        %d%%\n", (int)v4l_get_hue(v4l));
  fprintf(stdout, "VIDEO DEVICE COLOR:      %d%%\n", (int)v4l_get_colour(v4l));
  fprintf(stdout, "VIDEO DEVICE CONTRAST:   %d%%\n", (int)v4l_get_contrast(v4l));
  if (v4l->channel_flags & VIDEO_VC_TUNER) {
    fprintf(stdout, "VIDEO DEVICE FREQ:       %d\n",v4l->freq * 1000 / 16);
  }

  fflush(stdout);
}

