/* v4l.h */

#ifndef V4L_H
#define V4L_H

#include <sys/types.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <inttypes.h>

#define V4L_OK    0
#define V4L_FAIL -1

#define V4L_NORM_UNKNOWN 0
#define V4L_NORM_PAL     1
#define V4L_NORM_NTSC    2
#define V4L_NORM_SECAM   3
#define V4L_NORM_AUTO    4

#define V4L_CSP_UNKNOWN  0
//#define V4L_CSP_RGB565   1
#define V4L_CSP_RGB24    2       /* if little endian "BGR" */
#define V4L_CSP_ARGB32   3       /* if little endian "BGRA" */
//#define V4L_CSP_RGB555   4
//#define V4L_CSP_YUV422PK 5     /* YUV 4:2:2 packed, "YUYV" */
#define V4L_CSP_YUY2     6       /* YUV 4:2:2 packed, "YUYV" */
#define V4L_CSP_YUYV     7       /* YUV 4:2:2 packed, "YUYV" */
#define V4L_CSP_UYVY     8       /* YUV 4:2:2 packed, "UYVY" */

#define V4L_CSP_YUV422   10      /* YUV 4:2:2 planer */
#define V4L_CSP_YUV420   11      /* YUV 4:2:0 planer */

typedef struct {
  const char *device_name;
  int source_channel;
  int tuner;
  int norm;
  int audio;
  int brightness;
  int hue;
  int colour;
  int contrast;
  int whiteness;
  int freq;
} v4l_param_t;

typedef struct {
  const char *device_name;
  int fd;

  int name[32];
  int type;
  int channels;
  int audios;
  int maxwidth;
  int maxheight;
  int minwidth;
  int minheight;

  int source_channel;
  char channel_name[32];
  int channel_tuners;
  uint32_t channel_flags;
  int channel_type;
  int channel_norm;

  int tuner;
  char tuner_name[32];
  uint32_t rangelow;
  uint32_t rangehigh;
  uint32_t tuner_flags;
  int tuner_mode;
  int tuner_signal;

  uint32_t freq;
  int norm;

  int brightness;
  int hue;
  int colour;
  int contrast;
  int whiteness;
  int depth;
  int palette;
  int csp;

  int audio;
  int audio_volume;
  int audio_bass;
  int audio_treble;
  uint32_t audio_flags;
  char audio_name[16];
  int audio_mode;
  int audio_balance;
  int audio_step;

  int x;
  int y;
  int width;
  int height;
  int chromakey;
  uint32_t win_flags;

  int frames;
  int cur_frame;
  int *offsets;
  int map_size;
  char *map_data;

  struct video_mmap vid_mmap;

} v4l_t;

v4l_t *v4l_init (const char *device_name, int source_channel, int tuner, int norm, int audio);
int v4l_quit (v4l_t *v4l);
int v4l_open (v4l_t *v4l);
int v4l_close (v4l_t *v4l);
int v4l_capture_cap(v4l_t *v4l);

int   v4l_grab_pic_start (v4l_t *v4l);
char *v4l_grab_pic_sync (v4l_t *v4l);
int   v4l_grab_pic_capture (v4l_t *v4l);
int   v4l_grab_pic_stop (v4l_t *v4l);
char *v4l_grab_pic (v4l_t *v4l);

int  v4l_overlay_on (v4l_t *v4l);
int  v4l_overlay_off (v4l_t *v4l);

int  v4l_set_source_channel (v4l_t *v4l, int source_channel);
int  v4l_get_channel (v4l_t *v4l);
int  v4l_set_tuner (v4l_t *v4l, int tuner);
int  v4l_get_tuner (v4l_t *v4l);
int  v4l_set_freq (v4l_t *v4l, int freq);
int  v4l_get_freq (v4l_t *v4l);
int  v4l_set_audio (v4l_t *v4l, int audio, int on);
int  v4l_set_window_size (v4l_t *v4l, int width, int height);
int  v4l_get_window_size (v4l_t *v4l, int *width, int *height);

int v4l_set_csp (v4l_t *v4l, int csp);
int v4l_get_csp (v4l_t *v4l);

int v4l_color(v4l_t *v4l, int brightness, int hue, int colour, int contrast, int whiteness);
#define V4L_BRIGHTNESS 0
#define V4L_HUE        1
#define V4L_COLOUR     2
#define V4L_CONTRAST   3
#define V4L_WHITENESS  4
int   v4l_set_color (v4l_t *v4l, float percent, int element);
int   v4l_set_brightness(v4l_t *v4l, float percent);
int   v4l_set_hue(v4l_t *v4l, float percent);
int   v4l_set_colour(v4l_t *v4l, float percent);
int   v4l_set_contrast(v4l_t *v4l, float percent);
int   v4l_set_whiteness(v4l_t *v4l, float percent);
float v4l_get_color (v4l_t *v4l, int element);
float v4l_get_brightness(v4l_t *v4l);
float v4l_get_hue(v4l_t *v4l);
float v4l_get_colour(v4l_t *v4l);
float v4l_get_contrast(v4l_t *v4l);
float v4l_get_whiteness(v4l_t *v4l);

int  v4l_set_norm (v4l_t *v4l, int norm);
int  v4l_get_norm (v4l_t *v4l);

void v4l_print_param(v4l_t *v4l);

#endif /* V4L_H */

