/* rtj1.c */

/*
 * Reference:
 * mpeg2vidcodec_v12.tar.gz <http://www.mpeg.org>
 * mjpegtools-1.6.1.tar.gz <http://mjpeg.sourceforge.net>
 *
 */


/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <limits.h>
#include <malloc.h>

#include "attributes.h"
#include "mmx.h"

#include "rtj1.h"

/* default intra quantization matrix */
static const uint8_t quantizer_matrix[64] =
{
   8, 16, 19, 22, 26, 27, 29, 34,
  16, 16, 22, 24, 27, 29, 34, 37,
  19, 22, 26, 27, 29, 34, 34, 38,
  22, 22, 26, 27, 29, 34, 37, 40,
  22, 26, 27, 29, 32, 35, 40, 48,
  26, 27, 29, 32, 35, 40, 48, 58,
  26, 27, 29, 34, 38, 46, 56, 69,
  27, 29, 35, 38, 46, 56, 69, 83
};

#define SCAN alternate_scan
#if 0
/* zig-zag scan */
static unsigned char zig_zag_scan[64] =
{
  0,1,8,16,9,2,3,10,17,24,32,25,18,11,4,5,
  12,19,26,33,40,48,41,34,27,20,13,6,7,14,21,28,
  35,42,49,56,57,50,43,36,29,22,15,23,30,37,44,51,
  58,59,52,45,38,31,39,46,53,60,61,54,47,55,62,63
};
#endif
static unsigned char alternate_scan[64] =
{ /* Alternate scan pattern */
  0,8,16,24,1,9,2,10,17,25,32,40,48,56,57,49,
  41,33,26,18,3,11,4,12,19,27,34,42,50,58,35,43,
  51,59,20,28,5,13,6,14,21,29,36,44,52,60,37,45,
  53,61,22,30,7,15,23,31,38,46,54,62,39,47,55,63
};

typedef struct
{
  unsigned char code; /* right justified */
  char len;
} VLCtable;

/* for codes longer than 8 bits (excluding leading zeroes) */
typedef struct
{
  unsigned short code; /* right justified */
  char len;
} sVLCtable;


/* Table B-12, variable length codes for dct_dc_size_luminance
 *
 * indexed by [dct_dc_size_luminance]
 */

static sVLCtable DClumtab[12]=
{
  {0x0004,3}, {0x0000,2}, {0x0001,2}, {0x0005,3}, {0x0006,3}, {0x000e,4},
  {0x001e,5}, {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x01ff,9}
};


/* Table B-13, variable length codes for dct_dc_size_chrominance
 *
 * indexed by [dct_dc_size_chrominance]
 */

static sVLCtable DCchromtab[12]=
{
  {0x0000,2}, {0x0001,2}, {0x0002,2}, {0x0006,3}, {0x000e,4}, {0x001e,5},
  {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x03fe,10},{0x03ff,10}
};


/* Table B-15, DCT coefficients table one
 *
 * indexed by [run][level-1]
 * split into two tables (dct_code_tab1a, dct_code_tab2a) to reduce size
 * 'End of Block' is treated elsewhere
 * codes do not include s (sign bit)
 */

static VLCtable dct_code_tab1a[2][40]=
{
 /* run = 0, level = 1...40 */
 {
  {0x02, 2}, {0x06, 3}, {0x07, 4}, {0x1c, 5},
  {0x1d, 5}, {0x05, 6}, {0x04, 6}, {0x7b, 7},
  {0x7c, 7}, {0x23, 8}, {0x22, 8}, {0xfa, 8},
  {0xfb, 8}, {0xfe, 8}, {0xff, 8}, {0x1f,14},
  {0x1e,14}, {0x1d,14}, {0x1c,14}, {0x1b,14},
  {0x1a,14}, {0x19,14}, {0x18,14}, {0x17,14},
  {0x16,14}, {0x15,14}, {0x14,14}, {0x13,14},
  {0x12,14}, {0x11,14}, {0x10,14}, {0x18,15},
  {0x17,15}, {0x16,15}, {0x15,15}, {0x14,15},
  {0x13,15}, {0x12,15}, {0x11,15}, {0x10,15}
 },
 /* run = 1, level = 1...18 */
 {
  {0x02, 3}, {0x06, 5}, {0x79, 7}, {0x27, 8},
  {0x20, 8}, {0x16,13}, {0x15,13}, {0x1f,15},
  {0x1e,15}, {0x1d,15}, {0x1c,15}, {0x1b,15},
  {0x1a,15}, {0x19,15}, {0x13,16}, {0x12,16},
  {0x11,16}, {0x10,16}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}
 }
};

static VLCtable dct_code_tab2a[30][5]=
{
  /* run = 2...31, level = 1...5 */
  {{0x05, 5}, {0x07, 7}, {0xfc, 8}, {0x0c,10}, {0x14,13}},
  {{0x07, 5}, {0x26, 8}, {0x1c,12}, {0x13,13}, {0x00, 0}},
  {{0x06, 6}, {0xfd, 8}, {0x12,12}, {0x00, 0}, {0x00, 0}},
  {{0x07, 6}, {0x04, 9}, {0x12,13}, {0x00, 0}, {0x00, 0}},
  {{0x06, 7}, {0x1e,12}, {0x14,16}, {0x00, 0}, {0x00, 0}},
  {{0x04, 7}, {0x15,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x05, 7}, {0x11,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x78, 7}, {0x11,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x7a, 7}, {0x10,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x21, 8}, {0x1a,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x25, 8}, {0x19,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x24, 8}, {0x18,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x05, 9}, {0x17,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x07, 9}, {0x16,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x0d,10}, {0x15,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1a,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x19,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x17,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x16,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}}
};


typedef struct {
  char val, len;
} VLCtab; 

typedef struct {
  char run, level, len;
} DCTtab;

#define ERROR (-1)

/* Table B-12, dct_dc_size_luminance, codes 00xxx ... 11110 */
static VLCtab DClumtab0[32] =
{ {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2},
  {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2},
  {0, 3}, {0, 3}, {0, 3}, {0, 3}, {3, 3}, {3, 3}, {3, 3}, {3, 3},
  {4, 3}, {4, 3}, {4, 3}, {4, 3}, {5, 4}, {5, 4}, {6, 5}, {ERROR, 0}
};

/* Table B-12, dct_dc_size_luminance, codes 111110xxx ... 111111111 */
static VLCtab DClumtab1[16] =
{ {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6}, {7, 6},
  {8, 7}, {8, 7}, {8, 7}, {8, 7}, {9, 8}, {9, 8}, {10,9}, {11,9}
};

/* Table B-13, dct_dc_size_chrominance, codes 00xxx ... 11110 */
static VLCtab DCchromtab0[32] =
{ {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2}, {0, 2},
  {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2}, {1, 2},
  {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2}, {2, 2},
  {3, 3}, {3, 3}, {3, 3}, {3, 3}, {4, 4}, {4, 4}, {5, 5}, {ERROR, 0}
};

/* Table B-13, dct_dc_size_chrominance, codes 111110xxxx ... 1111111111 */
static VLCtab DCchromtab1[32] =
{ {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6},
  {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6}, {6, 6},
  {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7}, {7, 7},
  {8, 8}, {8, 8}, {8, 8}, {8, 8}, {9, 9}, {9, 9}, {10,10}, {11,10}
};

/* Table B-15, DCT coefficients table one,
 * codes 000001xx ... 11111111
*/
DCTtab DCTtab0a[252] =
{
  {65,0,6}, {65,0,6}, {65,0,6}, {65,0,6}, /* Escape */
  {7,1,7}, {7,1,7}, {8,1,7}, {8,1,7},
  {6,1,7}, {6,1,7}, {2,2,7}, {2,2,7},
  {0,7,6}, {0,7,6}, {0,7,6}, {0,7,6},
  {0,6,6}, {0,6,6}, {0,6,6}, {0,6,6},
  {4,1,6}, {4,1,6}, {4,1,6}, {4,1,6},
  {5,1,6}, {5,1,6}, {5,1,6}, {5,1,6},
  {1,5,8}, {11,1,8}, {0,11,8}, {0,10,8},
  {13,1,8}, {12,1,8}, {3,2,8}, {1,4,8},
  {2,1,5}, {2,1,5}, {2,1,5}, {2,1,5},
  {2,1,5}, {2,1,5}, {2,1,5}, {2,1,5},
  {1,2,5}, {1,2,5}, {1,2,5}, {1,2,5},
  {1,2,5}, {1,2,5}, {1,2,5}, {1,2,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5},
  {3,1,5}, {3,1,5}, {3,1,5}, {3,1,5},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {1,1,3}, {1,1,3}, {1,1,3}, {1,1,3},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4}, /* EOB */
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {64,0,4}, {64,0,4}, {64,0,4}, {64,0,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,3,4}, {0,3,4}, {0,3,4}, {0,3,4},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,1,2}, {0,1,2}, {0,1,2}, {0,1,2},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,2,3}, {0,2,3}, {0,2,3}, {0,2,3},
  {0,4,5}, {0,4,5}, {0,4,5}, {0,4,5},
  {0,4,5}, {0,4,5}, {0,4,5}, {0,4,5},
  {0,5,5}, {0,5,5}, {0,5,5}, {0,5,5},
  {0,5,5}, {0,5,5}, {0,5,5}, {0,5,5},
  {9,1,7}, {9,1,7}, {1,3,7}, {1,3,7},
  {10,1,7}, {10,1,7}, {0,8,7}, {0,8,7},
  {0,9,7}, {0,9,7}, {0,12,8}, {0,13,8},
  {2,3,8}, {4,2,8}, {0,14,8}, {0,15,8}
};

/* Table B-15, DCT coefficients table one,
 * codes 000000100x ... 000000111x
 */
DCTtab DCTtab1a[8] =
{
  {5,2,9}, {5,2,9}, {14,1,9}, {14,1,9},
  {2,4,10}, {16,1,10}, {15,1,9}, {15,1,9}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 000000010000 ... 000000011111
 */
DCTtab DCTtab2[16] =
{
  {0,11,12}, {8,2,12}, {4,3,12}, {0,10,12},
  {2,4,12}, {7,2,12}, {21,1,12}, {20,1,12},
  {0,9,12}, {19,1,12}, {18,1,12}, {1,5,12},
  {3,3,12}, {0,8,12}, {6,2,12}, {17,1,12}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 0000000010000 ... 0000000011111
 */
DCTtab DCTtab3[16] =
{
  {10,2,13}, {9,2,13}, {5,3,13}, {3,4,13},
  {2,5,13}, {1,7,13}, {1,6,13}, {0,15,13},
  {0,14,13}, {0,13,13}, {0,12,13}, {26,1,13},
  {25,1,13}, {24,1,13}, {23,1,13}, {22,1,13}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 00000000010000 ... 00000000011111
 */
DCTtab DCTtab4[16] =
{
  {0,31,14}, {0,30,14}, {0,29,14}, {0,28,14},
  {0,27,14}, {0,26,14}, {0,25,14}, {0,24,14},
  {0,23,14}, {0,22,14}, {0,21,14}, {0,20,14},
  {0,19,14}, {0,18,14}, {0,17,14}, {0,16,14}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 000000000010000 ... 000000000011111
 */
DCTtab DCTtab5[16] =
{
  {0,40,15}, {0,39,15}, {0,38,15}, {0,37,15},
  {0,36,15}, {0,35,15}, {0,34,15}, {0,33,15},
  {0,32,15}, {1,14,15}, {1,13,15}, {1,12,15},
  {1,11,15}, {1,10,15}, {1,9,15}, {1,8,15}
};

/* Table B-14/15, DCT coefficients table zero / one,
 * codes 0000000000010000 ... 0000000000011111
 */
DCTtab DCTtab6[16] =
{
  {1,18,16}, {1,17,16}, {1,16,16}, {1,15,16},
  {6,3,16}, {16,2,16}, {15,2,16}, {14,2,16},
  {13,2,16}, {12,2,16}, {11,2,16}, {31,1,16},
  {30,1,16}, {29,1,16}, {28,1,16}, {27,1,16}
};


/* bitstream stuff */

#define BSR(v,p)   asm ("bsrl    %1,%0\n\t" : "=r" (p) : "r" (v), "0" (p))
#define BSWAP32(v) asm ("bswap  %0": "=r" (v) : "0" (v))
#define BSWAP16(v) asm ("xchgb  %b0,%h0": "=q" (v) : "0" (v))

#define BITMASK(v, n)                                      \
  { register unsigned int mask = 0;                        \
    asm volatile ("btsl    %1, %0\n\t"                     \
                  "decl    %0\n"                           \
                  : "=r" (mask) : "r" ((n)), "0" (mask));  \
    (v) &= mask; }

typedef struct {
  unsigned int bitbfr;
  int bitcnt;
  unsigned char *stream;
  int bytecnt;
} bitstream_t;

#define MODE_GET 0
#define MODE_PUT 1

static void bitstream_init(bitstream_t *bs, unsigned char *stream, int mode)
{
  if (mode == MODE_GET) {
    bs->bitbfr = ((*stream)<<24)|((*(stream+1))<<16)|
                  ((*(stream+2))<<8)|(*(stream+3));
    bs->stream = stream + 4;
  } else {
    bs->bitbfr = 0;
    bs->stream = stream;
  }
  bs->bitcnt = 0;
  bs->bytecnt = 0;
}

/* putbits */

#if 0
/* zero bit stuffing to next byte boundary (5.2.3, 6.2.1) */
static void alignbits(bitstream_t *bs)
{
  bs->bitcnt = ((bs->bitcnt+7)&~7);
  if (bs->bitcnt == 32) {
    register unsigned int v;
    v = bs->bitbfr;
    BSWAP32(v);
    *((unsigned int*)bs->stream) = v;
    bs->bitbfr = 0;
    bs->stream += 4;
    bs->bytecnt += 4;
  }
}

/* return total number of generated bits */
int bitcount(bitstream_t *bs)
{
  return 8 * bs->bytecnt + bs->bitcnt;
}
#endif

static int putbits_flush(bitstream_t *bs)
{
  if (bs->bitcnt> 0) {
    register unsigned int v;
    v = bs->bitbfr;
    BSWAP32(v);
    *((unsigned int*)bs->stream) = v;
    bs->bitbfr = 0;
    v = ((bs->bitcnt+7)&~7) >> 3;
    bs->stream += v;
    bs->bytecnt += v;
    bs->bitcnt = 0;
  }
  return 8 * bs->bytecnt;
}

#define putbits(bs,val,n) \
{ register unsigned int v; \
  BITMASK((val), (n)); \
  bs.bitcnt += n; \
  if (bs.bitcnt > 32) { \
    bs.bitcnt -= 32; \
    v = bs.bitbfr; \
    v |= (val >> bs.bitcnt); \
    BSWAP32(v); \
    *((unsigned int*)bs.stream) = v; \
    bs.stream += 4; \
    bs.bytecnt += 4; \
    bs.bitbfr = val << (32-bs.bitcnt); \
  } \
  else if (bs.bitcnt == 32) { \
    v = bs.bitbfr; \
    v |= val; \
    BSWAP32(v); \
    *((unsigned int*)bs.stream) = v; \
    bs.stream += 4; \
    bs.bytecnt += 4; \
    bs.bitbfr = 0; \
    bs.bitcnt = 0; \
  } \
  else { \
    bs.bitbfr |= val << (32-bs.bitcnt); \
  } \
}

/* getbits */

#define showbits(bs,n) ((bs).bitbfr>>(32-(n)))

#define getbits_flush(bs, n) \
  bs.bitbfr <<= (n); \
  bs.bitcnt += (n); \
  if (bs.bitcnt >= 16) { \
    bs.bitcnt -= 16; \
    bs.bitbfr |= ((*(bs.stream)<<8) | (*(bs.stream+1))) << bs.bitcnt; \
    bs.stream+=2; \
    bs.bytecnt+=2; \
  }

#define getbits(bs, n, val) \
  (val) = (bs.bitbfr>>(32-(n))); \
  bs.bitbfr <<= (n); \
  bs.bitcnt += (n); \
  if (bs.bitcnt >= 16) { \
    bs.bitcnt -= 16; \
    bs.bitbfr |= ((*(bs.stream)<<8) | (*(bs.stream+1))) << bs.bitcnt; \
    bs.stream+=2; \
    bs.bytecnt+=2; \
  }


/* rtj1 stuff */

void mj_fdct_mmx(short *block);
void mj_idct_mmx(short *block);

#define Y_PLANE 0
#define U_PLANE 1
#define V_PLANE 2

static void
init_quant_mat(rtj1_t *rj, int mode)
{
  int i, quant;
  int v, add;

  quant = rj->quant;

  if (mode == RTJ1_MODE_ENCODE) {
    rj->quant_mat[0] = 0;
    for (i = 1; i < 64; i++) {
      v = quantizer_matrix[i];
      v = ((65536*4) + (v>>1)) / (v);
      if (v > 32767)
	  v = 32767;
      rj->quant_mat[i] = (int16_t)v;
    }

    v = ((65536/2) + (quant>>1)) / (quant);
    if (v > 32767)
      v = 32767;
    add = ((v) * (3*quant+2) / 4 + (v>>1)) / v;
    for (i = 0; i < 4; i++) {
      rj->quant_mul[i] = v;
      rj->quant_add[i] = add;
    }
//    printf("quant_mul %d\n", v);
//    printf("quant_add %d\n", add);
  }
  else {
    for (i = 0; i < 64; i++)
      rj->inv_quant_mat[i] = (int16_t)((int)quantizer_matrix[i] * quant);
  }
}

static int
init_csp_comp(rtj1_t *rj, int image_csp)
{
  int i, interlace_flag = 0;
  comp_block_t *comp_block;
  int y_block_num, u_block_num, v_block_num;
  int chroma_width;
  int stream_csp;

  if (image_csp == rj->image_csp)
    return RTJ1_OK;

  rj->image_csp = image_csp;

  stream_csp = 0;

  if (image_csp >= RTJ1_CSP_INTERLACED)
    interlace_flag = 1;
  if (image_csp >= RTJ1_CSP_INTERLACED_Y)
    interlace_flag = 2;

  image_csp &= RTJ1_CSP_MASK;
  if (image_csp == RTJ1_CSP_YUV444) {
    stream_csp = image_csp;
    rj->num_comp = 12;
    rj->chroma_horz_period = 1;
    rj->chroma_vert_period = 1;
  }
  else if (image_csp == RTJ1_CSP_YUV422) {
    stream_csp = image_csp;
    rj->num_comp = 8;
    rj->chroma_horz_period = 2;
    rj->chroma_vert_period = 1;
  }
  else if (image_csp == RTJ1_CSP_YUV420 || image_csp == RTJ1_CSP_YV12) {
    stream_csp = RTJ1_CSP_YUV420;
    rj->num_comp = 6;
    rj->chroma_horz_period = 2;
    rj->chroma_vert_period = 2;
  }
  else {
    rj->csp = RTJ1_CSP_UNKNOWN;
    fprintf(stderr, "rtj1: unknown csp.\n");
    return RTJ1_FAIL;
  }

  if (interlace_flag == 1)
    stream_csp |= RTJ1_CSP_INTERLACED;
  else if (interlace_flag == 2)
    stream_csp |= RTJ1_CSP_INTERLACED_Y;

  rj->csp = stream_csp;

  rj->u_offset = rj->width * rj->height;
  rj->v_offset = rj->u_offset +
    (rj->width/rj->chroma_horz_period * rj->height/rj->chroma_vert_period);

  if (image_csp == RTJ1_CSP_YV12) {
    int tmp = rj->u_offset;
    rj->u_offset = rj->v_offset;
    rj->v_offset = tmp;
  }

  comp_block = rj->comp_block;

  y_block_num = u_block_num = v_block_num = 0;
  chroma_width = rj->width / rj->chroma_horz_period;
  for (i = 0; i < rj->num_comp; i++) {
    int offset;
    if (i < 4) {
      comp_block[i].cc = Y_PLANE;
      if (interlace_flag) comp_block[i].stride = rj->width*2;
      else                comp_block[i].stride = rj->width;

      if (y_block_num < 2)
	offset = 0;
      else
        if (interlace_flag) offset = rj->width;
	else                offset = rj->width * 8;
      if (y_block_num & 1)
	offset += 8;
      comp_block[i].offset = offset;

      y_block_num++;
    } else if (!(i&1)) {
      comp_block[i].cc = U_PLANE;
      if (interlace_flag == 1) comp_block[i].stride = chroma_width*2;
      else                     comp_block[i].stride = chroma_width;

      if (!(u_block_num&1))
	offset = 0;
      else
        if (interlace_flag == 1) offset = chroma_width;
	else                     offset = chroma_width * 8;
      if (u_block_num >= 2)
	offset += 8;
      comp_block[i].offset = offset;

      u_block_num++;
    } else {
      comp_block[i].cc = V_PLANE;
      if (interlace_flag == 1) comp_block[i].stride = chroma_width*2;
      else                     comp_block[i].stride = chroma_width;

      if (!(v_block_num&1))
	offset = 0;
      else
        if (interlace_flag == 1) offset = chroma_width;
	else                     offset = chroma_width * 8;
      if (v_block_num >= 2)
	offset += 8;
      comp_block[i].offset = offset;

      v_block_num++;
    }
  }

  return RTJ1_OK;
}

rtj1_t *
rtj1_quit(rtj1_t *rj)
{
  if (rj->quant_mat)     free(rj->quant_mat);
  if (rj->inv_quant_mat) free(rj->inv_quant_mat);
  if (rj->quant_mul)     free(rj->quant_mul);
  if (rj->comp_block)    free(rj->comp_block);
  free(rj);
  return NULL;
}

static rtj1_t *
rtj1_init(int width, int height, int image_csp, int quant, int dc_prec,int mode)
{
  rtj1_t *rj;
  int16_t *mat_ptr;

  rj = (rtj1_t*)malloc(sizeof(rtj1_t));
  if (!rj) {
    fprintf(stderr, "rtj1_init: memory error.\n");
    return NULL;
  }
  rj->width = width;
  rj->height = height;
  rj->image_csp = RTJ1_CSP_UNKNOWN;
  rj->csp = RTJ1_CSP_UNKNOWN;

  if (dc_prec < 0)
    dc_prec = 0;
  else if (dc_prec > 3)
    dc_prec = 3;
  rj->dc_prec = dc_prec;

  if (quant <= 0)
    quant = 1;
  else if (quant > 31)
    quant = 31;
  rj->quant = quant;

  rj->quant_mat = rj->inv_quant_mat = NULL;
  rj->quant_mul = rj->quant_add = NULL;

  rj->num_comp = 0;
  rj->comp_block = (comp_block_t*)malloc(sizeof(comp_block_t)*12);
  if (!rj->comp_block) {
    fprintf(stderr, "rtj1: memory allocate_error.\n");
    rtj1_quit(rj);
    return NULL;
  }

  mat_ptr = (int16_t*)malloc(sizeof(int16_t)*64);
  if (!mat_ptr) {
    fprintf(stderr, "rtj1_init: memory error.\n");
    rtj1_quit(rj);
    return NULL;
  }

  if (mode == RTJ1_MODE_ENCODE) {
    rj->quant_mat = mat_ptr;
    rj->quant_mul = (int16_t*)malloc(sizeof(int16_t)*4*2);
    if (!rj->quant_mat || !rj->quant_mul) {
      fprintf(stderr, "rtj1_init: memory error.\n");
      rtj1_quit(rj);
      return NULL;
    }
    rj->quant_add = rj->quant_mul+4;
    init_quant_mat(rj, mode);

    if (init_csp_comp(rj, image_csp) != RTJ1_OK) {
      fprintf(stderr, "rtj1_init: init_csp_comp failed.\n");
      rtj1_quit(rj);
      return NULL;
    }
  }
  else {
    rj->inv_quant_mat = mat_ptr;
    init_quant_mat(rj, mode);

    if (image_csp != RTJ1_CSP_UNKNOWN) {
      if (init_csp_comp(rj, image_csp) != RTJ1_OK) {
        fprintf(stderr, "rtj1_init: init_csp_comp failed.\n");
        rtj1_quit(rj);
        return NULL;
      }
    }
  }

  return rj;
}

rtj1_t*
rtj1_encode_init (int width, int height, int image_csp, int quant, int dc_prec)
{
  return rtj1_init(width, height, image_csp, quant, dc_prec, RTJ1_MODE_ENCODE);
}

rtj1_t*
rtj1_decode_init (int width, int height, int image_csp)
{
  return rtj1_init(width, height, image_csp, 2, 3, RTJ1_MODE_DECODE);
}

static inline void char_to_short(uint8_t *src, int16_t *dst, int stride)
{
  int i;
  pxor_r2r(mm2, mm2);
  for (i = 0; i < 8; i++) {
    movq_m2r(*src, mm0);
    movq_r2r(mm0, mm1);
    punpcklbw_r2r(mm2, mm0);
    punpckhbw_r2r(mm2, mm1);
    movq_r2m(mm0, *dst);
    movq_r2m(mm1, *(dst+4));
    dst += 8;
    src += stride;
  }
}

static inline void quantize(rtj1_t *rj, int16_t *dst, int dc_prec)
{
  int i;
  int dc;
  int16_t *dcp;
  int16_t *qmat = rj->quant_mat;

  dcp = dst;
  dc = *dcp;
  if (dc < 0) {
    dc = -dc;
    dc >>= (3-dc_prec);
    dc = -dc;
  } else {
    dc >>= (3-dc_prec);
  }

#if 0
  for (i = 1; i < 64; i++) {
    int x = dst[i];
    int d = default_intra_quantizer_matrix[i];
    int y = (32*(x>=0 ? x : -x) + (d>>1))/d;
    d = (3*mquant+2)>>2;
    y = (y+d)/(2*mquant);
    if (y > 2047) {
      fprintf(stderr, "rtj1: quantize over 2047 %d\n", y);
      y = 2047;
    }
    dst[i] = (x>=0) ? y : -y;
    if (x > max_fdct)
      max_fdct = x;
    if (x < min_fdct)
      min_fdct = x;
    if (dst[i] > max_quat)
      max_quat = dst[i];
    if (dst[i] < min_quat)
      min_quat = dst[i];
  }
#else
  movq_m2r(*(rj->quant_mul), mm2);
//  movq_m2r(*(rj->quant_add), mm3);
  for (i = (64/4); i > 0; i--) {
    movq_m2r(*dst, mm0);
    movq_m2r(*qmat, mm4);
    pxor_r2r(mm1, mm1);
    pcmpgtw_r2r(mm0, mm1);
    pxor_r2r(mm1, mm0);
    psubw_r2r(mm1, mm0);
    pmulhw_r2r(mm4, mm0);
    psllw_i2r(3, mm0);
//    paddw_r2r(mm3, mm0);
    pmulhw_r2r(mm2, mm0);
    pxor_r2r(mm1, mm0);
    psubw_r2r(mm1, mm0);
    movq_r2m(mm0, *dst);
    dst += 4;
    qmat += 4;
  }
#endif
  *dcp = dc;
}

static void putvlc(short *block, bitstream_t *bs, int cc)
{
  int i, run, signed_level, level, len, size, code;
//  int dct_diff;
  short *bp = block;
  sVLCtable *stab = NULL;
  VLCtable *ptab = NULL;
  bitstream_t pbs = *bs;

  if (cc == Y_PLANE) {
    stab = DClumtab;
  } else {
    stab = DCchromtab;
  }
  signed_level = block[0];
  level = (signed_level<0) ? -signed_level : signed_level;
#ifdef DEBUG
  if (level>2047) {
    fprintf(stderr, "DC value out of range (signed_level=%d)\n", signed_level);
    exit(1);
  }
#endif
  size = 0;
  if (level) {
    BSR(level, size);
    size++;
  } else {
    size = 0;
  }

  len = stab[size].len;
  code = stab[size].code;
  putbits(pbs, code, len);
  if (size!=0) {
    if (signed_level < 0)
      level = signed_level + (1<<size) - 1;
    putbits(pbs, level, size);
  }

  run = 0;
  for (i = 1; i < 64; i++) {
    signed_level = bp[SCAN[i]];
    if (signed_level == 0) {
      run++;
      continue;
    }

    level = (signed_level<0) ? -signed_level : signed_level;
#ifdef DEBUG
    if (run<0 || run>63 || level==0 || level>2047) {
      fprintf(stderr, "AC value out of range (run=%d, signed_level=%d)\n",
      run,signed_level);
      exit(1);
    }
#endif
    len = 0;
    if (run < 2 && level < 41) {
      ptab = &dct_code_tab1a[run][level-1];
      len = ptab->len;
    } else if (run < 32 && level < 6) {
      ptab = &dct_code_tab2a[run-2][level-1];
      len = ptab->len;
    }

    if (len != 0) {
      code = ptab->code;
      putbits(pbs, code, len);
      if (signed_level<0)
        signed_level = 1;
      else
        signed_level = 0;
      putbits(pbs, signed_level,1);
    } else {
      len = 1l;
      putbits(pbs, len,6);
      putbits(pbs, run,6);
      putbits(pbs, signed_level,12);
    }
    run = 0;
  }
  level = 6;
  putbits(pbs, level,4);

  *bs = pbs;
}

int rtj1_encode(rtj1_t *rj, uint8_t *src, uint8_t *dst)
{
  int width, height;
  int chroma_width, chroma_block_width, chroma_block_height;
  int num_comp;
  comp_block_t *comp_block;
  int x, y, c;
  uint8_t *src_planes[3], *planes[3];
  uint8_t *sp;
  int dc_dct_pred[3];
  int dct_diff;
  short work[64];
  bitstream_t bs;
  int quant;
  int dc_prec;
  rtj1_head_t *head;
  int size;

  head = (rtj1_head_t*)dst;
  head->magic = RTJ1_MAGIC;
  head->width = rj->width;
  BSWAP16(head->width);
  head->height = rj->height;
  BSWAP16(head->height);
  head->csp = rj->csp;
  quant = rj->quant | (rj->dc_prec << 5);
  head->quant = quant;
  head->method = 0;
  dst += sizeof(rtj1_head_t);

  bitstream_init(&bs, dst, MODE_PUT);

  width = rj->width;
  height = rj->height;
  chroma_width = width / rj->chroma_horz_period;
  chroma_block_width = 16 / rj->chroma_horz_period;
  chroma_block_height = 16 / rj->chroma_vert_period;

  src_planes[Y_PLANE] = src;
  src_planes[U_PLANE] = src + rj->u_offset;
  src_planes[V_PLANE] = src + rj->v_offset;

  num_comp = rj->num_comp;
  comp_block = rj->comp_block;

  quant = rj->quant;
  dc_prec = rj->dc_prec;

  dc_dct_pred[Y_PLANE] =
  dc_dct_pred[U_PLANE] =
  dc_dct_pred[V_PLANE] = 0;

  for (y = 0; y < height; y+=16) {
    planes[Y_PLANE] = src_planes[Y_PLANE];
    planes[U_PLANE] = src_planes[U_PLANE];
    planes[V_PLANE] = src_planes[V_PLANE];
    for (x = 0; x < width; x+=16) {
      for (c = 0; c < num_comp; c++) {
	sp = planes[comp_block[c].cc];
        sp += comp_block[c].offset;

        char_to_short(sp, work, comp_block[c].stride);
        mj_fdct_mmx(work);
        quantize(rj, work, dc_prec);

	dct_diff = work[0] - dc_dct_pred[comp_block[c].cc];
	dc_dct_pred[comp_block[c].cc] = work[0];
	work[0] = dct_diff;

        putvlc(work, &bs, comp_block[c].cc);

      }
      planes[Y_PLANE] += 16;
      planes[U_PLANE] += chroma_block_width;
      planes[V_PLANE] += chroma_block_width;
    }
    src_planes[Y_PLANE] += width * 16;
    if (num_comp != 6 || !(rj->csp & RTJ1_CSP_INTERLACED)) {
      src_planes[U_PLANE] += chroma_width * chroma_block_height;
      src_planes[V_PLANE] += chroma_width * chroma_block_height;
    } else {
      if (y & 16) {
        src_planes[U_PLANE] += chroma_width * (16-1);
        src_planes[V_PLANE] += chroma_width * (16-1);
      } else {
        src_planes[U_PLANE] += chroma_width;
        src_planes[V_PLANE] += chroma_width;
      }
    }
  }
  emms();

  putbits_flush(&bs);

  size = bs.bytecnt;
  BSWAP32(size);
  head->data_size = size;

  return bs.bytecnt + sizeof(rtj1_head_t);
}


/* rtj1 decode stuff */

int rtj1_get_width(rtj1_t *rj)
{
  return rj->width;
}

int rtj1_get_height(rtj1_t *rj)
{
  return rj->height;
}

int rtj1_get_csp(rtj1_t *rj)
{
  return rj->csp;
}

int rtj1_get_decode_csp(rtj1_t *rj)
{
  return rj->image_csp;
}

int rtj1_set_decode_csp(rtj1_t *rj, int csp)
{
  rj->image_csp = csp;
  return init_csp_comp(rj, csp);
}

int rtj1_decode_analyze(rtj1_t *rj, uint8_t *src, int size)
{
  rtj1_head_t *head;
  int width, height;
  int quant, dc_prec;

  if (size < (int)sizeof(rtj1_head_t)) {
    return RTJ1_NEED_MORE_DATA;
  }

  head = (rtj1_head_t*)src;
  if (head->magic != RTJ1_MAGIC) {
    fprintf(stderr, "rtj1_decode_analyze: magic doesn't suit.\n");
    return RTJ1_FAIL;
  }
  width = head->width;
  BSWAP16(width);
  if (width != rj->width) {
    if (rj->width != 0)
      fprintf(stderr, "rtj1_decode_analyze: width doesn't suit.\n");
    rj->width = width;
  }
  height = head->height;
  BSWAP16(height);
  if (height != rj->height) {
    if (rj->height != 0)
      fprintf(stderr, "rtj1_decode_analyze: height doesn't suit.\n");
    rj->height = height;
  }

  if (rj->image_csp == RTJ1_CSP_UNKNOWN) {
    if (init_csp_comp(rj, head->csp) != RTJ1_OK) {
      fprintf(stderr, "rtj1_decode_analyze: invalid csp.\n");
      exit(1);
    }
  } else {
    switch (head->csp & RTJ1_CSP_MASK) {
      case RTJ1_CSP_YUV444:
	if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV444) {
          fprintf(stderr, "rtj1_decode_analyze: output csp doesn't suit.\n");
	  return RTJ1_FAIL;
	}
	break;
      case RTJ1_CSP_YUV422:
	if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV422) {
          fprintf(stderr, "rtj1_decode_analyze: output csp doesn't suit.\n");
	  return RTJ1_FAIL;
	}
	break;
      case RTJ1_CSP_YUV420:
	if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV420 &&
	    (rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YV12){
          fprintf(stderr, "rtj1_decode_analyze: output csp doesn't suit.\n");
	  return RTJ1_FAIL;
	}
	break;
    }

    if (init_csp_comp(rj, (head->csp&RTJ1_CSP_INTERLACED_MASK)|(rj->image_csp&RTJ1_CSP_MASK)) != RTJ1_OK) {
      fprintf(stderr, "rtj1_decode_analyze: invalid csp.\n");
      exit(1);
    }
  }

  quant = head->quant;
  dc_prec = (quant & (128-1)) >> 5;
  quant = quant & (32-1);
  rj->dc_prec = dc_prec;
  if (quant != rj->quant) {
    rj->quant = quant;
    init_quant_mat(rj, RTJ1_MODE_DECODE);
  }

  return RTJ1_OK;
}

static inline void short_to_char(int16_t *src, uint8_t *dst, int stride)
{
  int i;
  for (i = 0; i < 8; i++) {
    movq_m2r(*src, mm0);
    movq_m2r(*(src+4), mm1);
    packuswb_r2r(mm1, mm0);
    movntq_r2m(mm0, *dst);
    dst += stride;
    src += 8;
  }
}

static inline void dequantize(rtj1_t *rj, int16_t *dst, int dc_prec)
{
  int i;
  int dc;
  int16_t *dcp;
  int16_t *qmat = rj->inv_quant_mat;
//  int sum;

  dcp = dst;
  dc = *dcp << (3-dc_prec);

  for (i = (64/4); i > 0; i--) {
    movq_m2r(*dst, mm0);
    pmullw_m2r(*qmat, mm0);
    psraw_i2r(4, mm0);
    movq_r2m(mm0, *dst);
    dst += 4;
    qmat += 4;
  }

  *dcp = dc;
#if 0
  for (i = sum = 0; i < 64; i++)
    sum += *dcp++;
  if ((sum&1)==0)
    *--dcp ^= 1;
#endif
}


static int getvlc(short *block, bitstream_t *bs, int cc)
{
  int val, i, j, sign, run;
  unsigned int code;
  int size, dct_diff;
  DCTtab *tab;
  short *bp;
  bitstream_t gbs = *bs;

  bp = block;

  code = showbits(gbs, 5);

  if (cc == Y_PLANE) {
    if (code < 31) {
      size = DClumtab0[code].val;
      getbits_flush(gbs, DClumtab0[code].len);
    } else {
      code = showbits(gbs, 9) - 0x1f0;
      size = DClumtab1[code].val;
      getbits_flush(gbs, DClumtab1[code].len);
    }
  } else {
    if (code < 31) {
      size = DCchromtab0[code].val;
      getbits_flush(gbs, DCchromtab0[code].len);
    } else {
      code = showbits(gbs, 10) - 0x3e0;
      size = DCchromtab1[code].val;
      getbits_flush(gbs, DCchromtab1[code].len);
    }
  }

  if (size == 0)
    dct_diff = 0;
  else {
    getbits(gbs, size, dct_diff);
    if ((dct_diff & (1<<(size-1))) == 0)
      dct_diff -= (1<<size) - 1;
  }

  bp[0] = dct_diff;

//  for (i=0; i < 64; i++)
  for (i=1; ; i++)
  {
    code = showbits(gbs, 16);
    if (code>=1024)
    {
      tab = &DCTtab0a[(code>>8)-4];
    }
    else if (code>=512)
    {
      tab = &DCTtab1a[(code>>6)-8];
    }
    else if (code>=256)
      tab = &DCTtab2[(code>>4)-16];
    else if (code>=128)
      tab = &DCTtab3[(code>>3)-16];
    else if (code>=64)
      tab = &DCTtab4[(code>>2)-16];
    else if (code>=32)
      tab = &DCTtab5[(code>>1)-16];
    else if (code>=16)
      tab = &DCTtab6[code-16];
    else
    {
      fprintf(stderr, "invalid Huffman code in Decode_MPEG2_Intra_Block()\n");
      exit(1);
//      Fault_Flag = 1;
//      return 1;
    }

    getbits_flush(gbs, tab->len);

    if (tab->run==64) /* end_of_block */
    {
#if 0
      for (; i < 64; i++) {
        j = SCAN[i];
        bp[j] = 0;
      }
#endif
      *bs = gbs;
      return 0;
    }

    if (tab->run==65) /* escape */
    {
      getbits(gbs, 6, run);
      i += run;

      getbits(gbs, 12, val);
      if ((val&2047)==0)
      {
        fprintf(stderr, "invalid escape in Decode_MPEG2_Intra_Block()\n");
	exit(1);
//        Fault_Flag = 1;
//        return 1;
      }
      if((sign = (val>=2048)))
        val = 4096 - val;
    }
    else
    {
      i+= run = tab->run;
      val = tab->level;
      getbits(gbs, 1, sign);
    }

    if (i>=64)
    {
      fprintf(stderr,"DCT coeff index (i) out of bounds (intra2)\n");
      fprintf(stderr, "tab.len %d, tab.run %d, tab.level %d\n", tab->len, tab->run, tab->level);
      exit(1);
//      Fault_Flag = 1;
      return 1;
    }

    j = SCAN[i];
    bp[j] = sign ? -val : val;
  }

  *bs = gbs;
  return 0;
}

#define clear_wp(wp) { int i; for (i = 0; i<64; i++) {(wp)[i] = 0;} }

int rtj1_decode(rtj1_t *rj, uint8_t *src, uint8_t *dst, int size)
{
  int width, height;
  int chroma_width, chroma_block_width, chroma_block_height;
  int num_comp;
  comp_block_t *comp_block;
  int x, y, c;
  uint8_t *dst_planes[3], *planes[3];
  uint8_t *dp;
  int dc_dct_pred[3];
//  int dct_diff;
  short work[64];
  bitstream_t bs;
  int quant;
  int dc_prec;
  rtj1_head_t *head;

  head = (rtj1_head_t*)src;
  if (head->magic != RTJ1_MAGIC) {
    fprintf(stderr, "rtj1_decode: magic doesn't suit.\n");
    exit(1);
  }
  width = head->width;
  BSWAP16(width);
  if (width != rj->width) {
    fprintf(stderr, "rtj1_decode: width doesn't suit.\n");
    exit(1);
  }
  height = head->height;
  BSWAP16(height);
  if (height != rj->height) {
    fprintf(stderr, "rtj1_decode: height doesn't suit.\n");
    exit(1);
  }
  if (head->csp != rj->csp) {
    if (rj->image_csp == RTJ1_CSP_UNKNOWN) {
      if (init_csp_comp(rj, head->csp) != RTJ1_OK) {
        fprintf(stderr, "rtj1_decode: invalid csp.\n");
        exit(1);
      }
    } else {
      switch (head->csp & RTJ1_CSP_MASK) {
        case RTJ1_CSP_YUV444:
	  if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV444) {
            fprintf(stderr, "rtj1_decode: output csp doesn't suit.\n");
  fprintf(stderr, "head->csp %d, rj->csp %d, rj->image_csp %d\n", head->csp, rj->csp, rj->image_csp);
	    exit(1);
	  }
	  break;
        case RTJ1_CSP_YUV422:
	  if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV422) {
            fprintf(stderr, "rtj1_decode: output csp doesn't suit.\n");
  fprintf(stderr, "head->csp %d, rj->csp %d, rj->image_csp %d\n", head->csp, rj->csp, rj->image_csp);
	    exit(1);
	  }
	  break;
        case RTJ1_CSP_YUV420:
	  if ((rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YUV420 &&
	      (rj->image_csp&RTJ1_CSP_MASK) != RTJ1_CSP_YV12){
            fprintf(stderr, "rtj1_decode: output csp doesn't suit.\n");
  fprintf(stderr, "head->csp %d, rj->csp %d, rj->image_csp %d\n", head->csp, rj->csp, rj->image_csp);
	    exit(1);
	  }
	  break;
      }
      if (init_csp_comp(rj, (head->csp&RTJ1_CSP_INTERLACED_MASK)|(rj->image_csp&RTJ1_CSP_MASK)) != RTJ1_OK) {
        fprintf(stderr, "rtj1_decode: invalid csp.\n");
        exit(1);
      }
    }
  }

  quant = head->quant;
  dc_prec = (quant & (128-1)) >> 5;
  quant = quant & (32-1);
  if (quant != rj->quant) {
    rj->quant = quant;
    init_quant_mat(rj, RTJ1_MODE_DECODE);
  }

  BSWAP32(head->data_size);
  if (size != (int)(head->data_size + sizeof(rtj1_head_t))) {
    fprintf(stderr, "rtj1_decode: data_size doesn't size.\n");
  }
  src += sizeof(rtj1_head_t);

  bitstream_init(&bs, src, MODE_GET);

  chroma_width = width / rj->chroma_horz_period;
  chroma_block_width = 16 / rj->chroma_horz_period;
  chroma_block_height = 16 / rj->chroma_vert_period;

  dst_planes[Y_PLANE] = dst;
  dst_planes[U_PLANE] = dst + rj->u_offset;
  dst_planes[V_PLANE] = dst + rj->v_offset;

  num_comp = rj->num_comp;
  comp_block = rj->comp_block;

  quant = rj->quant;

  dc_dct_pred[Y_PLANE] =
  dc_dct_pred[U_PLANE] =
  dc_dct_pred[V_PLANE] = 0;

  for (y = 0; y < height; y+=16) {
    planes[Y_PLANE] = dst_planes[Y_PLANE];
    planes[U_PLANE] = dst_planes[U_PLANE];
    planes[V_PLANE] = dst_planes[V_PLANE];
    for (x = 0; x < width; x+=16) {
      for (c = 0; c < num_comp; c++) {
        clear_wp(work);
        getvlc(work, &bs, comp_block[c].cc);

	dc_dct_pred[comp_block[c].cc] += work[0];
	work[0] = dc_dct_pred[comp_block[c].cc];

        dequantize(rj, work, dc_prec);
        mj_idct_mmx(work);
	dp = planes[comp_block[c].cc];
        dp += comp_block[c].offset;
        short_to_char(work, dp, comp_block[c].stride);
      }
      planes[Y_PLANE] += 16;
      planes[U_PLANE] += chroma_block_width;
      planes[V_PLANE] += chroma_block_width;
    }
    dst_planes[Y_PLANE] += width * 16;
    if (num_comp != 6 || !(rj->csp & RTJ1_CSP_INTERLACED)) {
      dst_planes[U_PLANE] += chroma_width * chroma_block_height;
      dst_planes[V_PLANE] += chroma_width * chroma_block_height;
    } else {
      if (y & 16) {
        dst_planes[U_PLANE] += chroma_width * (16-1);
        dst_planes[V_PLANE] += chroma_width * (16-1);
      } else {
        dst_planes[U_PLANE] += chroma_width;
        dst_planes[V_PLANE] += chroma_width;
      }
    }
  }
  emms();

  return RTJ1_OK;
}

