/* vc_rtj1.c */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef ARCH_X86

#include <stdio.h>
#include <string.h>

#include "csp.h"
#include "parseopt.h"
#include "vcodec.h"
#include "vc_rtj1.h"
#include "rtj1.h"

static int vc_rtj1_quant = 2;
static int vc_rtj1_dc_precision = 3;
static int vc_rtj1_interlaced = 2;

OptionDef rtj1_param[] = {
  {"rtj1-quant", "RTJ1_QUANT", HAS_ARG, {(void*)&vc_rtj1_quant}, {2}, {1}, 31, "RTJ1 quant", "d"},
  {"rtj1-dc-precision", "RTJ1_DC_PRECISION", HAS_ARG, {(void*)&vc_rtj1_dc_precision}, {3}, {0}, 3, "RTJ1 DC precision (0: 8bit, 1: 9 bit, 2: 10 bit, 3: 11 bit", "d"},
};
int rtj1_param_num = sizeof(rtj1_param) / sizeof(OptionDef);

typedef struct {
  unsigned long codec;
  int width;
  int height;
  int in_csp;
  int out_csp;
  int quant;
  int data_size;
  rtj1_t *rj;
} RTJ1_PARAM;


static RTJ1_PARAM vc_rjenc;
static RTJ1_PARAM vc_rjdec;

static int enc_initialized = 0;
static int dec_initialized = 0;


static struct CODEC_FOURCC {
  uint32_t codec;
  const char *fourcc_str;
  uint32_t codec_cap;
} codec_fourcc[] = {
  { VCODEC_RTJ1, "RTJ1", VCODEC_CAP_ENCODE|VCODEC_CAP_DECODE },
};

static int codec_fourcc_num = sizeof(codec_fourcc)/sizeof(struct CODEC_FOURCC);

uint32_t
vc_rtj1_codec_cap(uint32_t codec, uint32_t cap_flag)
{
  int i;

  for (i = 0; i < codec_fourcc_num; i++) {
    if (codec == codec_fourcc[i].codec)
      return (cap_flag & codec_fourcc[i].codec_cap);
  }
  return VCODEC_CAP_NONE;
}

const char*
vc_rtj1_codec_to_fourcc_str(uint32_t codec)
{
  int i;

  for (i = 0; i < codec_fourcc_num; i++) {
    if (codec == codec_fourcc[i].codec)
      return codec_fourcc[i].fourcc_str;
  }
  return NULL;
}

uint32_t
vc_rtj1_fourcc_str_to_codec(const char *fourcc_str)
{
  int i;

  if (!fourcc_str)
    return VCODEC_UNKNOWN;

  for (i = 0; i < codec_fourcc_num; i++) {
    if (!strcmp(fourcc_str, codec_fourcc[i].fourcc_str))
      return codec_fourcc[i].codec;
  }
  return VCODEC_UNKNOWN;
}

/* encode functions */

int
vc_rtj1_encode_csp_cap(int csp)
{
  int ret;
  switch (csp) {
    default:
      ret = CSP_UNKNOWN;
      break;
    case CSP_UNKNOWN:
      ret = CSP_YUV420;
      break;
    case CSP_YV12:
//    case CSP_I420:
    case CSP_YUV420:
      ret = csp;
      break;
  }
  return ret;
}

int
vc_rtj1_encode (unsigned char *pic_data, unsigned char *stream_buf,
    int stream_buf_size, unsigned int *flag)
{
  int data_size;

  if (!enc_initialized) {
    fprintf(stderr, "vc_rtj1_encode: uninitialized.\n");
    return VCODEC_FAIL;
  }

  data_size = rtj1_encode(vc_rjenc.rj, pic_data, stream_buf);
  if (data_size > stream_buf_size) {
    fprintf(stderr, "vc_rtj1_encode: buffer overflow.\n");
    return VCODEC_FAIL;
  }

  *flag = VCODEC_IS_INTRA;
  return data_size;
}

int
vc_rtj1_encode_quit (void)
{
  vc_rjenc.codec = 0;

  vc_rjenc.rj = rtj1_quit(vc_rjenc.rj);
  enc_initialized = 0;
  return VCODEC_OK;
}

void
vc_rtj1_encode_print_param(void)
{
  if (!enc_initialized) {
    fprintf (stdout, "RTJ1 ENCODE uninitialized.\n");
    return;
  }
  fprintf (stdout, "RTJ1 ENCODE PARAMETER\n");
  fprintf (stdout, "dimmention:    %dx%d\n",vc_rjenc.width,vc_rjenc.height);
  fprintf (stdout, "dc precision   %d\n", vc_rjenc.rj->dc_prec);
  fprintf (stdout, "quant          %d\n", vc_rjenc.rj->quant);
  fflush (stdout);
}

int
vc_rtj1_encode_init (VCODEC *vcodec)
{
  static int first = 1;
  int csp, rtj1_csp;

  if (vcodec->codec != VCODEC_RTJ1) {
    fprintf(stderr, "vc_rtj1_encode_init: invalid codec.\n");
    return VCODEC_FAIL;
  }

  if (first) {
    vc_rjenc.codec = 0;
    vc_rjenc.width = 0;
    vc_rjenc.height = 0;
    vc_rjenc.in_csp = CSP_UNKNOWN;
    vc_rjenc.out_csp = CSP_UNKNOWN;

    vc_rjenc.rj = NULL;
    vc_rjenc.quant = 2;
    first = 0;
  }

  vc_rjenc.codec = vcodec->codec;

  csp = CSP_UNKNOWN;
  csp = vc_rtj1_encode_csp_cap(vcodec->in_csp);
  if (csp != vcodec->in_csp) {
    fprintf(stderr,"vc_rtj1_encode_init: invalid in csp %s.\n",csp_to_name(vcodec->in_csp));
    return VCODEC_FAIL;
  }
  vc_rjenc.in_csp = csp;

  vcodec->out_csp = CSP_COMPRESSED;
  vc_rjenc.out_csp = vcodec->out_csp;

  vc_rjenc.width = vcodec->width;
  vc_rjenc.height = vcodec->height;

  rtj1_csp = RTJ1_CSP_UNKNOWN;
  switch (csp) {
    case CSP_YUV420:
      rtj1_csp = RTJ1_CSP_YUV420;
      if (vc_rtj1_interlaced) {
        if (vc_rtj1_interlaced == 1)
          rtj1_csp = RTJ1_CSP_YUV420 | RTJ1_CSP_INTERLACED;
        else
          rtj1_csp = RTJ1_CSP_YUV420 | RTJ1_CSP_INTERLACED_Y;
      }
      break;
    case CSP_YV12:
      rtj1_csp = RTJ1_CSP_YV12;
      if (vc_rtj1_interlaced) {
        if (vc_rtj1_interlaced == 1)
          rtj1_csp = RTJ1_CSP_YV12 | RTJ1_CSP_INTERLACED;
	else
          rtj1_csp = RTJ1_CSP_YV12 | RTJ1_CSP_INTERLACED_Y;
      }
      break;
  }
  vc_rjenc.rj = rtj1_encode_init(vc_rjenc.width, vc_rjenc.height, rtj1_csp, vc_rtj1_quant, vc_rtj1_dc_precision);
  if (!vc_rjenc.rj) {
    fprintf(stderr,"vc_rtj1_encode_init: codec initialize failed.\n");
  }

  enc_initialized = 1;

  return VCODEC_OK;
}

/* decode functions */

int
vc_rtj1_decode_csp_cap(int csp)
{
  int ret;
  switch (csp) {
    default:
      ret = CSP_UNKNOWN;
      break;
    case CSP_UNKNOWN:
      ret = CSP_YUV420;
      break;
//    case CSP_I420:
    case CSP_YUV420:
    case CSP_YV12:
      ret = csp;
      break;
  }
  return ret;
}

int
vc_rtj1_decode (unsigned char *pic_data, unsigned char *stream,
    int stream_length, unsigned int *flag)
{
  if (!dec_initialized) {
    fprintf(stderr, "vc_rtj1_decode: uninitialized.\n");
    return VCODEC_FAIL;
  }

  if (flag && *flag & VCODEC_NULL_DECODE)
    return 0;

  rtj1_decode(vc_rjdec.rj, stream, pic_data, stream_length);

  *flag = VCODEC_IS_INTRA;
  return vc_rjdec.data_size;
}

int
vc_rtj1_decode_quit (void)
{
  vc_rjdec.codec = 0;

  vc_rjdec.rj = rtj1_quit(vc_rjdec.rj);
  vc_rjdec.rj = NULL;
  dec_initialized = 0;

  return VCODEC_OK;
}

void
vc_rtj1_decode_print_param(void)
{
  if (!dec_initialized) {
    fprintf (stdout, "RTJ1 DECODE uninitialized.\n");
    return;
  }
  fprintf (stdout, "RTJ1 DECODE PARAMETER\n");
  fprintf (stdout, "dimmention: %dx%d\n",vc_rjdec.width,vc_rjdec.height);
  fflush (stdout);
}

int
vc_rtj1_decode_init (VCODEC *vcodec)
{
  static int first = 1;
  int csp, depth;
  int rtj1_csp;

  if (vcodec->codec != VCODEC_RTJ1) {
    fprintf(stderr, "vc_rtj1_decode_init: invalid codec.\n");
    return VCODEC_FAIL;
  }

  if (first) {
    vc_rjdec.codec = 0;
    vc_rjdec.width = 0;
    vc_rjdec.height = 0;
    vc_rjdec.in_csp = CSP_UNKNOWN;
    vc_rjdec.out_csp = CSP_UNKNOWN;

    vc_rjdec.data_size = 0;
    vc_rjdec.rj = NULL;
    vc_rjdec.quant = 2;
    first = 0;
  }

  vc_rjdec.codec = vcodec->codec;

  csp = CSP_UNKNOWN;
  csp = vc_rtj1_decode_csp_cap(vcodec->out_csp);
  if (csp != vcodec->out_csp) {
    fprintf(stderr,"vc_rtj1_decode_init: invalid in csp %s.\n",csp_to_name(vcodec->out_csp));
    return VCODEC_FAIL;
  }
  vc_rjdec.out_csp = csp;

  vcodec->in_csp = CSP_COMPRESSED;
  vc_rjdec.in_csp = vcodec->in_csp;

  vc_rjdec.width = vcodec->width;
  vc_rjdec.height = vcodec->height;

  depth = csp_to_bits_per_pixel(vc_rjdec.out_csp);
  vc_rjdec.data_size = vcodec->width * vcodec->height * depth / 8;

  switch (vc_rjdec.out_csp) {
    case CSP_YUV420:
      rtj1_csp = RTJ1_CSP_YUV420;
      break;
    case CSP_YV12:
      rtj1_csp = RTJ1_CSP_YV12;
      break;
    default:
      rtj1_csp = RTJ1_CSP_UNKNOWN;
      break;
  }
  vc_rjdec.rj = rtj1_decode_init(vc_rjdec.width, vc_rjdec.height, rtj1_csp);

  dec_initialized = 1;

  return VCODEC_OK;
}

/* vcodec functions struct */

VCODEC_FUNCS vc_rtj1_funcs = {
  vc_rtj1_encode_init,
  vc_rtj1_encode_quit,
  vc_rtj1_encode_csp_cap,
  vc_rtj1_encode,
  vc_rtj1_encode_print_param,

  vc_rtj1_decode_init,
  vc_rtj1_decode_quit,
  vc_rtj1_decode_csp_cap,
  vc_rtj1_decode,
  vc_rtj1_decode_print_param,

  vc_rtj1_codec_to_fourcc_str,
  vc_rtj1_fourcc_str_to_codec,
  vc_rtj1_codec_cap,
};

#endif /* ARCH_X86 */

