/* venc.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <pthread.h>
#include <errno.h>

#include "vcodec.h"
#include "venc.h"
#include "vframe.h"
#include "vframe_mgr.h"
#include "vgrab.h"


/* Thread functions */

static void venc_error(venc_t *venc, const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "venc: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);

  if (!venc) return;

  pthread_mutex_lock(&venc->mutex);
  venc->state = VENC_STATE_ERROR;
  pthread_mutex_unlock(&venc->mutex);
  if (venc->error_cb)
    venc->error_cb(venc->error_cb_data);
//  exit(1);
}

static inline int
get_state(venc_t *venc)
{
  int state;
  pthread_mutex_lock(&venc->mutex);
  state = venc->state;
  pthread_mutex_unlock(&venc->mutex);
  return state;
}

static inline void
state_change_done(venc_t *venc)
{
  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_INTERFACE_WAIT) {
    venc->state &= ~VENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&venc->interface_wait_cond);
  }
  pthread_mutex_unlock(&venc->mutex);
}

static inline void
set_state(venc_t *venc, uint32_t state)
{
  pthread_mutex_lock(&venc->mutex);
  venc->state &= ~(VENC_STATE_START|VENC_STATE_PAUSE|VENC_STATE_STOP|VENC_STATE_QUIT);
  venc->state |= state;
  if (venc->state & VENC_STATE_INTERFACE_WAIT) {
    venc->state &= ~VENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&venc->interface_wait_cond);
  }
  pthread_mutex_unlock(&venc->mutex);
}

static int
add_vframe_chain(venc_t *venc, vframe_t *vfrm)
{
  int ret;
  pthread_mutex_lock(&venc->mutex);
  ret = vframe_chain_add(&venc->vframe_chain, vfrm);
  pthread_mutex_unlock(&venc->mutex);
  if (venc->vframe_ready_cb)
    venc->vframe_ready_cb(venc->vframe_ready_cb_data);
  return ret;
}

static int
put_enc_vframe(venc_t *venc)
{
//  vframe_t *dvfrm;
  int data_length;
  unsigned int ext_flag;
  int ret = VENC_OK;

#ifndef NDEBUG
  if (!venc->src_vframe->bufp) {
    venc_error(venc, "put_enc_vframe: error EXT_VFRAME, flag %u.\n",venc->src_vframe->flag);
    return VENC_FAIL;
  }
#endif /* NDEBUG */

  if (!venc->dst_vframe) {
    while((venc->dst_vframe=vframe_mgr_get_vframe(venc->vframe_mgr)) == NULL) {
#ifndef NDEBUG
      fprintf(stderr, "put_enc_vframe: dvfrm == NULL\n");
#endif /* NDEBUG */
      pthread_mutex_lock(&venc->mutex);
      venc->state |= VENC_STATE_THREAD_WAIT;
      pthread_cond_wait(&venc->thread_wait_cond, &venc->mutex);
      if (venc->state & VENC_STATE_ERROR) {
        pthread_mutex_unlock(&venc->mutex);
        return VENC_FAIL;
      }
      pthread_mutex_unlock(&venc->mutex);
    }
  }

  ext_flag = 0;
  data_length = vcodec_encode(venc->src_vframe->data, venc->dst_vframe->data,
      venc->dst_vframe->data_buf_size, &ext_flag);

  if (data_length == VCODEC_NEED_MORE_DATA) {
    data_length = 0;
  }
  else if (data_length < 0) {
    venc_error(venc, "put_enc_vframe: vcodec_encode() error %d.\n",data_length);
    ret = VENC_FAIL;
  }

  if (ext_flag & VCODEC_IS_INTRA)
    venc->dst_vframe->flag = venc->src_vframe->flag | VFRAME_FLAG_INTRA;
  else
    venc->dst_vframe->flag = venc->src_vframe->flag | VFRAME_FLAG_INTER;
  venc->dst_vframe->data_length = data_length;
  venc->dst_vframe->duration = venc->src_vframe->duration;

  if (data_length > 0)
    add_vframe_chain(venc, venc->dst_vframe);
  else
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->dst_vframe);

  venc->dst_vframe = venc->src_vframe;
  venc->src_vframe = NULL;

  return ret;
}

/* Thread */

static void*
venc_thread (void *data)
{
  uint32_t state;
  int rec_start_flag = 0;
  int quit_flag = 0;
  int drop_frame_num = 0;
  venc_t *venc = (venc_t*)data;

  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_INTERFACE_WAIT) {
    pthread_cond_signal(&venc->interface_wait_cond);
  }
  venc->state = VENC_STATE_STOP;
  pthread_mutex_unlock(&venc->mutex);

  while (1) {
    state = get_state(venc);

    if (quit_flag || state & VENC_STATE_ERROR) {
      break;
    }

    pthread_mutex_lock(&venc->mutex);
    while ((venc->src_vframe = vgrab_get_vframe(venc->vgrab)) == NULL) {
      venc->state |= VENC_STATE_THREAD_WAIT;
      pthread_cond_wait(&venc->thread_wait_cond, &venc->mutex);
      state = venc->state;
      if (state & VENC_STATE_ERROR)
	break;
    }
    pthread_mutex_unlock(&venc->mutex);

    if (state & VENC_STATE_ERROR) {
#ifndef NDEBUG
      fprintf(stderr,"venc_thread: state VENC_STATE_ERROR.\n");
#endif /* NDEBUG */
      break;
    }

    if (venc->src_vframe->flag & VFRAME_FLAG_NORMAL) {
      if (put_enc_vframe(venc) == VENC_FAIL) {
        venc_error(venc, "venc_thread: put_enc_vframe() failed.\n");
	break;
      }
    } else {
      if (venc->src_vframe->flag & VFRAME_FLAG_START) {
        rec_start_flag = 1;
        set_state(venc, VENC_STATE_START);
        drop_frame_num = 0;
	state_change_done(venc);
      }
      else if (venc->src_vframe->flag & VFRAME_FLAG_END) {
        if (rec_start_flag) {
          if (vcodec_encode_reset() != VCODEC_OK) {
            venc_error(venc, "venc_thread: vcodec_reset() error.\n");
            break;
          }
        }
        rec_start_flag = 0;
        set_state(venc, VENC_STATE_STOP);
	state_change_done(venc);
      }
      else if (venc->src_vframe->flag & VFRAME_FLAG_QUIT) {
        rec_start_flag = 0;
        set_state(venc, VENC_STATE_QUIT);
        quit_flag = 1;
      }
      else if (venc->src_vframe->flag & VFRAME_FLAG_ERROR) {
        rec_start_flag = 0;
        set_state(venc, VENC_STATE_ERROR);
        quit_flag = 1;
      }
      else if (venc->src_vframe->flag & VFRAME_FLAG_DROP) {
        drop_frame_num++;
      }
      add_vframe_chain(venc, venc->src_vframe);
      venc->src_vframe = NULL;
    }
  }

  if (venc->dst_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->dst_vframe);
    venc->dst_vframe = NULL;
  }
  if (venc->src_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->src_vframe);
    venc->src_vframe = NULL;
    venc_error(venc, "src_vframe remains.\n");
  }

  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_ERROR)
    fprintf(stderr, "venc_thread: STATE_ERROR break\n");
  if (venc->state & VENC_STATE_INTERFACE_WAIT) {
    venc->state &= ~VENC_STATE_INTERFACE_WAIT;
    pthread_cond_signal(&venc->interface_wait_cond);
  }
  venc->state &= (VENC_STATE_QUIT|VENC_STATE_ERROR);
  venc->state |= VENC_STATE_THREAD_FINISH;
  pthread_mutex_unlock(&venc->mutex);

#ifndef NDEBUG
  fprintf(stderr, "venc_thread: done\n");
#endif /* NDEBUG */
  return venc;
}

/* interface functions */

static int
venc_src_vframe_ready_cb(void *cb_data)
{
  venc_t *venc = (venc_t*)cb_data;

  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_THREAD_WAIT) {
    venc->state &= ~VENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&venc->thread_wait_cond);
  }
  pthread_mutex_unlock(&venc->mutex);
  return VENC_OK;
}

int
venc_set_vframe_ready_cb(venc_t *venc, vframe_ready_cb_t vframe_ready_cb, void *vframe_ready_cb_data)
{
  pthread_mutex_lock(&venc->mutex);
  venc->vframe_ready_cb = vframe_ready_cb;
  venc->vframe_ready_cb_data = vframe_ready_cb_data;
  pthread_mutex_unlock(&venc->mutex);
  return VENC_OK;
}

int
venc_set_vframe_mgr(venc_t *venc, vframe_mgr_t *vfrm_mgr)
{
  pthread_mutex_lock(&venc->mutex);
  venc->vframe_mgr = vfrm_mgr;
  pthread_mutex_unlock(&venc->mutex);
  venc->dst_vframe = vframe_mgr_get_vframe(vfrm_mgr);
  return VENC_OK;
}

int
venc_set_error_cb(venc_t *venc, error_cb_t error_cb, void *error_cb_data)
{
  pthread_mutex_lock(&venc->mutex);
  venc->error_cb = error_cb;
  venc->error_cb_data = error_cb_data;
  pthread_mutex_unlock(&venc->mutex);
  return VENC_OK;
}

int
venc_get_out_csp(venc_t *venc)
{
  int out_csp;
  pthread_mutex_lock(&venc->mutex);
  out_csp = venc->out_csp;
  pthread_mutex_unlock(&venc->mutex);
  return out_csp;
}

vframe_t*
venc_get_vframe(venc_t *venc)
{
  vframe_t *vfrm;
  pthread_mutex_lock(&venc->mutex);
  vfrm = vframe_chain_remove(&venc->vframe_chain);
  pthread_mutex_unlock(&venc->mutex);
  return vfrm;
}

int
venc_state_error(venc_t *venc)
{
  int ret;
  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_THREAD_WAIT) {
    venc->state &= ~VENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&venc->thread_wait_cond);
  }
  if (venc->state & (VENC_STATE_ERROR|VENC_STATE_THREAD_FINISH)) {
    ret = VENC_OK;
  } else {
    venc->state &= ~(VENC_STATE_START|VENC_STATE_PAUSE|VENC_STATE_STOP);
    venc->state |= VENC_STATE_ERROR;
#if 0
    venc->state |= VENC_STATE_INTERFACE_WAIT;
    while (venc->state & VENC_STATE_INTERFACE_WAIT)
      pthread_cond_wait(&venc->interface_wait_cond, &venc->mutex);
#endif
    ret = VENC_OK;
  }
  pthread_mutex_unlock(&venc->mutex);
  return ret;
}

static int
venc_state_change_wait(venc_t *venc, uint32_t state)
{
  int ret;
  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_THREAD_WAIT) {
    venc->state &= ~VENC_STATE_THREAD_WAIT;
    pthread_cond_signal(&venc->thread_wait_cond);
  }
  if (venc->state & (state|VENC_STATE_QUIT|VENC_STATE_ERROR|VENC_STATE_THREAD_FINISH)) {
    ret = VENC_OK;
  } else {
    while (!(venc->state & state)) {
      venc->state |= VENC_STATE_INTERFACE_WAIT;
      pthread_cond_wait(&venc->interface_wait_cond, &venc->mutex);
    }
    ret = VENC_OK;
  }
  pthread_mutex_unlock(&venc->mutex);
  return ret;
}

int
venc_state_start(venc_t *venc)
{
  return venc_state_change_wait(venc, VENC_STATE_START);
}

int
venc_state_stop(venc_t *venc)
{
  return venc_state_change_wait(venc, VENC_STATE_STOP);
}

int
venc_state_quit(venc_t *venc)
{
  return venc_state_change_wait(venc, VENC_STATE_QUIT);
}

int
venc_state_toggle_pause(venc_t *venc)
{
  if (venc) return VENC_FAIL;
  return VENC_OK;
}

int
venc_error_quit (venc_t *venc)
{
  int retcode, state;
  void* retval;
  vframe_t *vfrm;

  venc_state_error(venc);

  state = get_state(venc);
  if (!(state & VENC_STATE_THREAD_FINISH) &&
      !(state & VENC_STATE_NOT_READY)) {
    retcode = pthread_join(venc->venc_th, &retval);
    if (retcode != 0)
       fprintf (stderr, "venc_error_quit: thread join error %d\n", retcode);
  }

  if (venc->dst_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->dst_vframe);
    venc->dst_vframe = NULL;
  }
  if (venc->src_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->src_vframe);
    venc->src_vframe = NULL;
  }

  pthread_mutex_lock(&venc->mutex);
  while ((vfrm = vframe_chain_remove(&venc->vframe_chain)) != NULL)
    vframe_mgr_add_free_frame(venc->vframe_mgr, vfrm);
  pthread_mutex_unlock(&venc->mutex);

  free(venc);

  return VENC_OK;
}

int
venc_quit (venc_t *venc)
{
  int retcode, state;
  void* retval;

  venc_state_quit(venc);

  state = get_state(venc);
  if (!(state & VENC_STATE_THREAD_FINISH) &&
      !(state & VENC_STATE_NOT_READY)) {
    retcode = pthread_join(venc->venc_th, &retval);
    if (retcode != 0)
       fprintf (stderr, "venc_quit: thread join error %d\n", retcode);
  }

  if (venc->dst_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->dst_vframe);
    venc->dst_vframe = NULL;
  }
  if (venc->src_vframe) {
    vframe_mgr_add_free_frame(venc->vframe_mgr, venc->src_vframe);
    venc->src_vframe = NULL;
    fprintf(stderr, "venc_quit src_vframe remains.\n");
  }
  free(venc);
//  pthread_mutex_destroy(&venc->mutex);
  return VENC_OK;
}

venc_t*
venc_init (int width, int height, double fps, uint32_t vcodec, int in_csp, vgrab_t *vgrab)
{
  int retcode;
  venc_t *venc = NULL;

  if (vcodec_encode_init(vcodec, width, height, fps, in_csp) != VCODEC_OK) {
    venc_error(venc, "vcodec_encode_init failed.\n");
    return NULL;
  }

  venc = (venc_t*)malloc(sizeof(venc_t));
  if (!venc) {
    venc_error(venc, "venc_init failed.\n");
    return NULL;
  }

  venc->state = VENC_STATE_NOT_READY;
  venc->src_vframe = NULL;
  venc->dst_vframe = NULL;
  venc->vframe_mgr = NULL;
  venc->vframe_ready_cb = NULL;
  venc->vframe_ready_cb_data = NULL;
  vframe_chain_init(&venc->vframe_chain);

  venc->error_cb = NULL;
  venc->error_cb_data = NULL;

  venc->width = width;
  venc->height = height; 
  venc->fps = fps;
  venc->vcodec = vcodec;
  venc->in_csp = in_csp;
  venc->out_csp = vcodec_encode_get_out_csp();

  venc->vgrab = vgrab;
  vgrab_set_vframe_ready_cb(vgrab, venc_src_vframe_ready_cb, venc);

  pthread_mutex_init (&venc->mutex, NULL);
  pthread_cond_init (&venc->interface_wait_cond, NULL);
  pthread_cond_init (&venc->thread_wait_cond, NULL);

  retcode = pthread_create (&venc->venc_th, NULL, venc_thread, venc);
  if (retcode != 0) {
    venc_error(venc, "venc_init() thread create failed %d\n", retcode);
    venc_error_quit(venc);
    return NULL;
  }

  pthread_mutex_lock(&venc->mutex);
  if (venc->state & VENC_STATE_NOT_READY) {
    venc->state |= VENC_STATE_INTERFACE_WAIT;
    pthread_cond_wait(&venc->interface_wait_cond, &venc->mutex);
  }
  pthread_mutex_unlock(&venc->mutex);

  return venc;
}

#if 0
void
venc_print_param(void)
{
  VENC_CONFIG *conf = &venc_conf;
//  SEL_COMPO *sel;
//  int i;

  fprintf (stderr, "VIDEO PARAM\n");
#if 0
  sel = codecs;
  for (i=0; i < sizeof(codecs)/sizeof(SEL_COMPO); i++, sel++) {
    if (sel->id == conf->fourcc) {
      fprintf (stderr, "video codec:            %s\n", sel->name);
      break;
    }
  }
#endif
  fprintf (stderr, "fps:                    %f\n", conf->fps);
  fprintf (stderr, "dimmention:             %dx%d\n",conf->width,conf->height);
}
#endif

