/* vframe_mgr.c */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>

#include "csp.h"
#include "vframe.h"
#include "vframe_mgr.h"

#define VFRAME_NUM_MAX 64
#define VFRAME_NUM_MIN 5
#define VFRAME_MAX_LAST_FRAME_ALLOC_COUNT (64)
#define VFRAME_MAX_EXT_FRAME_NUM 256

static void error(const char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "vframe_mgr: ");
  vfprintf(stderr, format, ap);
  va_end(ap);
  putc('\n',stderr);
  exit(1);
}

static vframe_t*
alloc_vframe(vframe_mgr_t *vfrm_mgr, int buf_alloc_flag)
{
  int offset;
  vframe_t *vfrm;

  if (buf_alloc_flag) {
    if (vfrm_mgr->alloc_frame_num >= vfrm_mgr->max_frame_num)
      return NULL;
  } else if (vfrm_mgr->alloc_ext_frame_num >= vfrm_mgr->max_ext_frame_num)
    return NULL;

  vfrm = (vframe_t*)malloc(sizeof(vframe_t));
  if (!vfrm) {
    pthread_mutex_unlock(&vfrm_mgr->mutex);
    error("memory allocate error.\n");
  }

  if (buf_alloc_flag) {
#ifdef HAVE_MEMALIGN
    vfrm->bufp = (uint8_t*)memalign(16, vfrm_mgr->buf_size);
    vfrm->data = vfrm->bufp;
#else
    vfrm->bufp = (uint8_t*)malloc(vfrm_mgr->buf_size+15);
    vfrm->data = (uint8_t*)((((uint32_t)(vfrm->bufp)) + 15) & ~0xf);
#endif
    if (!vfrm->bufp) {
      free(vfrm);
      pthread_mutex_unlock(&vfrm_mgr->mutex);
      error("memory allocate error.\n");
    }

    vfrm->csp = vfrm_mgr->csp;
    vfrm->data_buf_size = vfrm_mgr->buf_size;
    vfrm->width = vfrm_mgr->width;
    vfrm->height = vfrm_mgr->height;
    vfrm->horz_border = vfrm_mgr->horz_border;
    vfrm->vert_border = vfrm_mgr->vert_border;
    vfrm->stride[0] = vfrm_mgr->stride[0];
    vfrm->stride[1] = vfrm_mgr->stride[1];
    vfrm->stride[2] = vfrm_mgr->stride[2];

    offset = 0;
    vfrm->planes[0] = vfrm->data + offset + vfrm_mgr->plane_offset[0];
    offset += vfrm_mgr->plane_size[0];
    vfrm->planes[1] = vfrm->data + offset + vfrm_mgr->plane_offset[1];
    offset += vfrm_mgr->plane_size[1];
    vfrm->planes[2] = vfrm->data + offset + vfrm_mgr->plane_offset[2];

    vfrm->flag = VFRAME_FLAG_NORMAL;

    vfrm_mgr->alloc_frame_num++;
    vfrm_mgr->last_frame_alloc_count = 0;
  } else {
    vfrm->csp = CSP_UNKNOWN;
    vfrm->data_buf_size = 0;
    vfrm->width = vfrm->height =
      vfrm->horz_border = vfrm->vert_border = 
      vfrm->stride[0] = vfrm->stride[1] = vfrm->stride[2] = 0;
    vfrm->bufp = vfrm->data =
      vfrm->planes[0] = vfrm->planes[1] = vfrm->planes[2] = NULL;

    vfrm->flag = VFRAME_FLAG_EXT;

    vfrm_mgr->alloc_ext_frame_num++;
//    vfrm_mgr->last_frame_alloc_count = 0;
  }

  vfrm->data_length = 0;
//  vfrm->exflag =
  vfrm->duration = vfrm->count = 0;
  vfrm->prev = vfrm->next = NULL;

  return vfrm;
}

static void
free_vframe(vframe_mgr_t *vfrm_mgr, vframe_t *vfrm)
{
#ifndef NDEBUG
  if (vfrm->prev || vfrm->next) {
    error("free_vframe: frame doesn't removed chain.\n");
  }
#endif /* NDEBUG */
  if (vfrm->bufp) {
    free (vfrm->bufp);
    vfrm_mgr->alloc_frame_num--;
    vfrm_mgr->last_frame_alloc_count = 0;
#ifndef NDEBUG
    if (vfrm_mgr->alloc_frame_num < 0)
      error("free_vframe: alloc_frame_num %d\n", vfrm_mgr->alloc_frame_num);
#endif /* NDEBUG */
  } else {
    vfrm_mgr->alloc_ext_frame_num--;
//    vfrm_mgr->last_frame_alloc_count = 0;
#ifndef NDEBUG
    if (vfrm_mgr->alloc_ext_frame_num < 0)
      error("free_vframe: alloc_ext_frame_num %d\n", vfrm_mgr->alloc_ext_frame_num);
#endif /* NDEBUG */
  }
  free(vfrm);
}

int
vframe_mgr_add_free_frame(vframe_mgr_t *vfrm_mgr, vframe_t *vfrm)
{
  int ret;
  pthread_mutex_lock(&vfrm_mgr->mutex);
  if (!vfrm->bufp) {
    free_vframe(vfrm_mgr, vfrm);
    pthread_mutex_unlock(&vfrm_mgr->mutex);
    return 0;
  }

  if (vfrm_mgr->alloc_frame_num > VFRAME_NUM_MIN &&
      vfrm_mgr->last_frame_alloc_count > vfrm_mgr->max_last_frame_alloc_count &&
      vfrm_mgr->free_frame_num > 0) {
    free_vframe(vfrm_mgr, vfrm);
    vfrm_mgr->last_frame_alloc_count = 0;
    ret = 1;
  } else {
    vfrm->flag = VFRAME_FLAG_NORMAL;
    vfrm->data_length = 0;
//    vfrm->exflag = 0;
    vframe_chain_add(&vfrm_mgr->free_chain, vfrm);
    vfrm_mgr->free_frame_num++;
    vfrm_mgr->last_frame_alloc_count++;
    ret = 0;
  }
  vfrm_mgr->used_frame_num--;
  pthread_mutex_unlock(&vfrm_mgr->mutex);
  return ret;
}

vframe_t*
vframe_mgr_get_ext_vframe(vframe_mgr_t *vfrm_mgr)
{
  vframe_t *vfrm = NULL;
  pthread_mutex_lock(&vfrm_mgr->mutex);
  vfrm = alloc_vframe(vfrm_mgr, 0);
  pthread_mutex_unlock(&vfrm_mgr->mutex);
  return vfrm;
}

vframe_t*
vframe_mgr_get_vframe(vframe_mgr_t *vfrm_mgr)
{
  vframe_t *vfrm = NULL;

  pthread_mutex_lock(&vfrm_mgr->mutex);
//  if (vfrm_mgr->used_frame_num < vfrm_mgr->max_frame_num-1) {
    if (vfrm_mgr->free_chain.vframes) {
      vfrm = vframe_chain_remove(&(vfrm_mgr->free_chain));
      vfrm_mgr->free_frame_num--;
      vfrm_mgr->used_frame_num++;
    } else {
      vfrm = alloc_vframe(vfrm_mgr, 1);
      if (vfrm)
        vfrm_mgr->used_frame_num++;
    }
//  }
#ifndef NDEBUG
  if (!vfrm) {
    fprintf(stderr, "vframe_mgr_get_vframe: frame_num %d, free_frame_num %d\n", vfrm_mgr->alloc_frame_num, vfrm_mgr->free_frame_num);
  }
#endif /* NDEBUG */
  pthread_mutex_unlock(&vfrm_mgr->mutex);

  return vfrm;
}

int
vframe_mgr_quit(vframe_mgr_t *vfrm_mgr)
{
  vframe_t *vfrm;

  pthread_mutex_lock(&vfrm_mgr->mutex);
  while (vfrm_mgr->free_chain.vframes) {
    vfrm = vframe_chain_remove(&vfrm_mgr->free_chain);
    free_vframe(vfrm_mgr, vfrm);
  }
  pthread_mutex_unlock(&vfrm_mgr->mutex);
//  pthread_mutex_destroy(&vfrm_mgr->mutex);
  free(vfrm_mgr);
  return VFRAME_MGR_OK;
}

vframe_mgr_t*
vframe_mgr_init(int width, int height, int horz_border, int vert_border, int csp, int max_frame_num)
{
  int i;
  vframe_t *vfrm;
  vframe_mgr_t *vfrm_mgr;
  CSP_INFO csp_info;

  vfrm_mgr = (vframe_mgr_t*)malloc(sizeof(vframe_mgr_t));
  if (!vfrm_mgr)
    error("memory allocate error.\n");

  if (max_frame_num < VFRAME_NUM_MIN)
    max_frame_num = VFRAME_NUM_MIN;
  else if (max_frame_num > VFRAME_NUM_MAX)
    max_frame_num = VFRAME_NUM_MAX;

  vfrm_mgr->width = width;
  vfrm_mgr->height = height;
  vfrm_mgr->horz_border = horz_border;
  vfrm_mgr->vert_border = vert_border;
  vfrm_mgr->csp = csp;
  vfrm_mgr->max_frame_num = max_frame_num;
  vfrm_mgr->max_last_frame_alloc_count = VFRAME_MAX_LAST_FRAME_ALLOC_COUNT;
  vfrm_mgr->max_ext_frame_num = VFRAME_MAX_EXT_FRAME_NUM;
  vfrm_mgr->alloc_frame_num = 0;
  vfrm_mgr->alloc_ext_frame_num = 0;
  vfrm_mgr->free_frame_num = 0;
  vfrm_mgr->used_frame_num = 0;

  vframe_chain_init(&vfrm_mgr->free_chain);

  pthread_mutex_init(&vfrm_mgr->mutex, NULL);

  if (csp_get_info(csp, &csp_info) == CSP_FAIL)
    error("csp_get_info failed, csp = %d\n", csp);

  vfrm_mgr->buf_size = csp_calc_buf_size(csp,
      width, height, horz_border, vert_border,
      vfrm_mgr->plane_size, vfrm_mgr->plane_offset, vfrm_mgr->stride);

  for (i = 0; i < VFRAME_NUM_MIN; i++) {
    vfrm = alloc_vframe(vfrm_mgr, 1);
    vframe_chain_add(&vfrm_mgr->free_chain, vfrm);
    vfrm_mgr->free_frame_num++;
  }

  return vfrm_mgr;
}

